////////////////////////////////////////////////////////
//	HEPEvtLite.h
//	Author: V.Makarenko (Vladimir.Makarenko@cern.ch)
////////////////////////////////////////////////////////

#if !defined(HEPEvtLite_H_INCLUDED_)
#define HEPEvtLite_H_INCLUDED_
/*
*	Pre-includes:
*	#include <iostream>
*	#include <fstream>
*	#include "VMUtil.h"
*	#include "VMOutput.h"
*	#include "VMHistLite.h"
*	#include "VMGen.h"
*
*	#include "vector"
*/

//[version:] ver, sub_ver(e.g. new feature or interface changed), bug_fix
// 1.2. CVMPhaseSpaceGen interface changed
#define CAlHepLib_VER		0x010201

///////////////////////////////////////////// Constants
// Note: PDG_KF_XXX are temporary definitions, don't use it
#define PDG_KF_D			1
#define PDG_KF_U			2
#define PDG_KF_S			3
#define PDG_KF_C			4
#define PDG_KF_B			5
#define PDG_KF_T			6
#define PDG_KF_ELECTRON		11
#define PDG_KF_NEUTRINO_E	12
#define PDG_KF_MUON			13
#define PDG_KF_NEUTRINO_MU	14
#define PDG_KF_TAU			15
#define PDG_KF_NEUTRINO_TAU	16
#define PDG_KF_PROTON		2212
#define PDG_KF_NEUTRON		2112
#define PDG_KF_GLUON		21
#define PDG_KF_PHOTON		22
#define PDG_KF_ZBOSON		23
#define PDG_KF_WBOSON		24
#define PDG_KF_HIGGS0		25

class CVMConst {
	static tReal m_Pi;
	static bool m_bRecalcPi;
public:
	static tReal Pi();
	static tReal Epsilon() { return epsReal; }
	//
	static const tReal Euler;
	static const tReal Alpha;
	static const tReal M_e;
	static const tReal M_mu;
	static const tReal M_p;
	static const tReal ReClassic;
	static const tReal GeV2pb;
	static const tReal El;			// electric charge
	static const tReal MMom_p;		// magnetic moment of proton
};

///////////////////////////////////////////// Math utilites
class CVMMathUtil {
public:
	static tReal Lambda(tReal x, tReal y, tReal z) { return (x*x+y*y+z*z-2.*x*(y+z)-2.*y*z); }
};

///////////////////////////////////////////// 4-vector
class CVM4Vec {
public:
	tReal m_E; // or t
	tReal m_Px; // or x
	tReal m_Py; // or y
	tReal m_Pz; // or z
public:
	CVM4Vec();
	CVM4Vec(tReal _E, tReal _Px, tReal _Py, tReal _Pz);
	CVM4Vec(const CVM4Vec& vec);
	//TODO: declare operators as inline
	CVM4Vec& operator=(const CVM4Vec& vec);
	CVM4Vec& operator+=(const CVM4Vec& vec);
	CVM4Vec& operator-=(const CVM4Vec& vec);
	CVM4Vec& operator*=(const tReal& f);
	CVM4Vec operator+(const CVM4Vec& vec) const; // Note: "+=" is faster
	CVM4Vec operator-(const CVM4Vec& vec) const; // Note: "-=" is faster
	bool operator==(const CVM4Vec& vec) const;
	void Set(tReal _E, tReal _Px, tReal _Py, tReal _Pz);
	bool Load(std::ifstream* ifs, bool _binary);
	bool Save(std::ostream* ofs, bool _binary) const;
	friend std::ostream& operator<<(std::ostream& os, const CVM4Vec& vec);
	tReal GetMassSigned() const; // sign(SQR)*sqrt(abs(SQR))
	tReal GetSqr() const;
	tReal GetCpl(const CVM4Vec& vec) const;
	tReal GetPAbs() const;
	tReal GetPSqr() const;
	tReal GetPt() const;
	tReal GetPtSqr() const;			//much faster than 'GetPt'
	tReal GetPhi() const;			//axial angle
	tReal GetThetaCos() const;		//polar angle cosine
	tReal GetTheta() const;			//polar angle (uses sin() or cos() depending on angle value)
	tReal GetEta() const;
	tReal GetAngleCos3Vec(const CVM4Vec& vec) const; // Cosine of angle between 3-vectors
	tReal GetDRSqr(const CVM4Vec& vec) const; // delta R squared
	void LorMToK(tReal m, const CVM4Vec& k); //Lorentz transform with system transform: [m,0,0,0]->k
	void LorKToM(const CVM4Vec& k, tReal m); //Lorentz transform with system transform: k->[m,0,0,0]
	void RotateXY(tReal phi); //Rotate vector in XY-plane (preserving pT)
	void RotateXY(tReal cosPhi, tReal sinPhi); //Rotate vector in XY-plane (preserving pT)
	void RotateXZ(tReal phi); //Rotate vector in XZ-plane
	void RotateXZ(tReal cosPhi, tReal sinPhi); //Rotate vector in XZ-plane
	void RotateYZ(tReal phi); //Rotate vector in YZ-plane
	void RotateYZ(tReal cosPhi, tReal sinPhi); //Rotate vector in YZ-plan
	/*
	*	RotateZYYX: rotate vector in ZY+YX planes accroding to {0,0,zPAbs}->{zRes} transformation.
	*	Nt. {0,0,zPAbs}->{0,0,-zPAbs} transformation is treated as pi-rotation in YZ-plane
	*	'zPAbs' and 'zResPt' are used to avoid extra 'sqrt()' operations.
	*/
	void RotateZYYX(const CVM4Vec& zRes, tReal zPAbs, tReal zResPt=-1.);
};
//CVM4Vec operator+(const CVM4Vec& vec1, const CVM4Vec& vec2);
inline std::ostream& operator<<(std::ostream& os, const CVM4Vec& vec) { os<<vec.m_E<<", "<<vec.m_Px<<", "<<vec.m_Py<<", "<<vec.m_Pz; return os; }

///////////////////////////////////////////// HEP library
class CAlHepLib {
public:
	static const int NUM_RND_BANKS;						//may be safely changed in 'HEPEvtLite.cpp' (but must be uniform for all runs!)
	static bool SelectRndBank(const char* bank_no);		//to be called first (before any RND-use), BEFORE 'CAlHepLib::Init()'!
	static bool Init(); //Inits LoopTools, QD, Pythia etc.
	static void Finalize(); //Finalizes LoopTools, QD, Pythia etc.
	static unsigned long GetVersion() { return CAlHepLib_VER; }
private:
	static bool m_Inited; //to avoid repetion of 'Init()' call
	static unsigned int m_QD_oldcw; //used by QD package initialization procedure
public:
	static CVMFileIn* m_DebugIn;
	static CVMFileOut* m_DebugOut;

public:
	//Self-test
	static bool TestLibrary(bool bSilent=true); //calls all tests: TestReal, LTTest, TestLi2

	//Sub-tests
	static bool TestLi2(bool bSilent=true);	//saves 'TestLi2.eps'
	static bool LTTest(bool bSilent=true);
	static bool TestReal(bool bLog=false);	//Test if precision of tReal is reduced by compiler setting
	static bool TestRnd(bool bLog=false);	//Test default RND

	//Other debug functions
	static char* LogConstantValues();		//Allocates memory using CVMUtil::strAlloc()

public:
	//proton structure functions
	static tReal F1(tReal Q2, tReal M2);
	static tReal F2(tReal Q2, tReal M2);
	static tReal F3(tReal Q2, tReal M2);
	static tReal F4(tReal Q2, tReal M2);
	static tReal Ge(tReal Q2);
	static tReal Gm(tReal Q2);

#ifdef VM_USE_PYTHIA8
	static Pythia8::Pythia* pythia;			//Pythia instance
#endif //#ifdef VM_USE_PYTHIA8

	//Soft bremsstrahlung integrals (see Denner'93 vol.7.2)
	static tReal m_SoftPhotonE;	//soft-hard photon delimiter, don't set to '0'
	static tReal SoftBremsInt1(const CVM4Vec& k1, tReal m1);
	static tReal SoftBremsInt2(const CVM4Vec& k1, const CVM4Vec& k2, tReal m1, tReal m2);

	//alternative 'pow(real,int)' function
	static tReal Pow(tReal val, int p);
	//static tReal Pow(tReal val, int p) { return pow(val,p); }

	//Algebraic functions
	static int m_iNormGenMethod; //default: 3;
	static tReal GenNormDistr(); //N(0,1), Note: N(mu,sigma)=mu+sigma*N(0,1)
	static tReal Li2(tReal x);
private:
	static tReal Li2_01(tReal x);
public:
	//static tReal Erf(tReal x); [TODO]

	//LoopTools settings and tests (empty calls if 'VM_USE_LOOPTOOLS' is not defined)
private:
	static bool LTInit(); //Used by CAlHepLib::Init()
public:
	static void LTClearCache();	// To be called regularly to avoid 'memory leak' effect in LoopTools.
public:
	//IR-delimiter
	static tReal GetIRLambdaSqr();
	static void SetIRLambdaSqr(tReal lambdaSqr);
	//UV-delimiters
	static tReal GetUVChi();
	static tReal GetDimMuSqr();
	static void SetUVChi(tReal chi);
	static void SetDimMuSqr(tReal mudim2);
private:
	static tReal m_IRLambda2;	// Used if LT is unavailable only.
	static tReal m_UVChi;		// Used if LT is unavailable only.
	static tReal m_DimMuSqr;	// Used if LT is unavailable only.
	static bool m_bLTInited;
public:
	// Calculate renorm. constants (currently in QED)
	static tReal LTCalcDFe_QED();
	static tReal LTCalcDMe_QED();
	static tReal LTCalcDZaa_QED();
	static tReal LTCalcDZe_QED();

	//VMMath library
public:
	// Find single solufion of func(x)==0, very slow implementation, for debug only
	static bool NSolve(tReal(vmdecl *func)(tReal), tReal xMin, tReal xMax, tReal& res); //Very slow!

public:
	//misc. operations
	static const char* GetParticleTitle(int idAbs);
	//static bool IsNaN(double val) { return (val!=val); }
	static bool IsNaN(double val);
	static double GetNaN();
	/*
	*	Note: unexpected problems spotted with inline 'IsInf' and -O2.
	*/
	//static bool IsInf(double val) { return ( (((unsigned long*)&val)[1] & 0x7ff80000) == 0x7ff00000); }
	static bool IsInf(double val);
	static double GetInf();
	/*
	*	Fill 'real(kind=16)' variable (as byte array)
	*/
	static void ConvertToReal16(double val, unsigned char* res16);
	static double ConvertFromReal16(const unsigned char* res16);
private:
	static long dbg_FirstStackPos; // Used by 'ViewStackUsed'
public:
	/*
	*	ViewStackUsed: Writes 'Stack used: xx bytes' (since first call) to 'std::cout'
	*	'pLastLocalVar' is used to trace stack position of specific variables.
	*	Note: ViewStackUsed(x,NULL) gives EQUAL result in any point of function ==>
	*		==> use 'last-var' to trace memory usage inside a function
	*/
	static void ViewStackUsed(const char* message=NULL, void* pLastLocalVar=NULL);

public:
	//Vector operations
	static tReal AsymmProd(const CVM4Vec& vec1, const CVM4Vec& vec2, const CVM4Vec& vec3, const CVM4Vec& vec4);
};

///////////////////////////////////////////// Particle
class CHEPParticleLite // same to TRootGenParticle
{
public:
	static const int STATUS_EMPTY;		//hepevt.isthep == 0
	static const int STATUS_FINAL;		//hepevt.isthep == 1
	static const int STATUS_INTERNAL;	//hepevt.isthep == 2
	static const int STATUS_INITIAL;	//hepevt.isthep == -1 (as in MG)
	static const int STATUS_DOC;		//hepevt.isthep == 21

public:
	int m_PID; // particle HEP ID number | hepevt.idhep[number]
	int m_Status; // particle status | hepevt.isthep[number]
	int m_Mother1_f0; // particle 1st mother | hepevt.jmohep[number][0] - 1
	int m_Mother2_f0; // particle 2nd mother | hepevt.jmohep[number][1] - 1
	int m_Daughter1_f0; // particle 1st daughter | hepevt.jdahep[number][0] - 1
	int m_Daughter2_f0; // particle 2nd daughter | hepevt.jdahep[number][1] - 1

	CVM4Vec m_Mom; // particle momentum (hepevt.phep[number][i])

	tReal m_Pt; // particle transverse momentum
	tReal m_Eta; // particle pseudorapidity
	tReal m_Phi; // particle azimuthal angle
	tReal m_Rapidity; // particle rapidity

	CVM4Vec m_Pos; // particle vertex position (hepevt.vhep[number][i])

	int m_Mark; // volatile user mark (no saved, no cloned)

public:
	CHEPParticleLite();
	CHEPParticleLite(const CHEPParticleLite& p);
	CHEPParticleLite &operator=(const CHEPParticleLite& p);
	bool Load(std::ifstream* ifs, bool _binary);
	bool Save(std::ofstream* ofs, bool _binary);
	bool IsParticleOK();
	tReal GetMass();
	void Log(bool bShortLog=true);
};

///////////////////////////////////////////// Particle order
//
// may be used instead of the simple int[] array
class CHEPParticleOrderLite {
private:
	int m_nParticles;
	int* m_order;			//original order: 0,1,2,3,...
public:
	CHEPParticleOrderLite(int nParticles);
	CHEPParticleOrderLite(const CHEPParticleOrderLite& order);
	CHEPParticleOrderLite(int nParticles, int* orderIToX);
	~CHEPParticleOrderLite();
	void ResetOrder();		//to straight order
	void SetOrder(int* orderIToX);
	void Swap(int i, int j);
	int Get(int i) const { return m_order[i]; }
	int GetOrigin(int i) const;
	bool CheckConsistency() const;
	CHEPParticleOrderLite* CreateBackwardOrder() const;
};

///////////////////////////////////////////// PDF
//
//	TODO: make single virtual function for every event (like in Pythia)
//		- e.g. 'GetDensity(int id[-6..+6],...)'
//
class CPDFLite
{
protected:
	CPDFLite();
	//
public:
	virtual ~CPDFLite();
	virtual bool IsReady()									= 0;
	virtual double GetDensity(int id, double x, double Q2)	= 0;
	virtual CPDFLite* Clone() const							= 0;
	//
	//virtual bool IsReady()		= 0;
	// Constructor.
	//[py]PDF(int idBeamIn = 2212) {idBeam = idBeamIn; idSav = 9; xSav = -1.;  Q2Sav = -1.; isSet = true; isInit = false;}  
	//[py]bool isSetup() {return isSet;} // Confirm that PDF has been set up (important for LHAPDF).
	//[py]virtual void setExtrapolate(bool) {} // Allow extrapolation beyond boundaries. This is optional.
	//[py]double xf(int id, double x, double Q2); // Read out parton density
	//[py]double xfVal(int id, double x, double Q2); // Read out valence and sea part of parton densities.
	//[py]double xfSea(int id, double x, double Q2);
	//[py]protected:
	//[py]// Store relevant quantities.
	//[py]int    idBeam, idSav;
	//[py]double xSav, Q2Sav;
	//[py]double xu, xd, xubar, xdbar, xs, xc, xb, xg, xlepton, xgamma, xuVal, xuSea, xdVal, xdSea;
	//[py]bool   isSet, isInit, hasLimits;
	//[py]virtual void xfUpdate(int id, double x, double Q2) = 0; // Update parton densities.

public:
	static CPDFLite* Create(long gup, long sup, int idBeam);
};

#ifdef VM_USE_PYTHIA8
class CPDFPyCTEQ5L : public CPDFLite {
	Pythia8::CTEQ5L pdf;	//Pythia-embeeded CTEQ5L set (PDFGUP(1)=4, PDFSUP(1)=46  <-- TODO: check!)
	int idBeam;
public:
	CPDFPyCTEQ5L(int idBeamIn=2212);
	virtual ~CPDFPyCTEQ5L();
public:
	virtual bool IsReady();
	virtual double GetDensity(int id, double x, double Q2);
	virtual CPDFLite* Clone() const;
};
#endif //#ifdef VM_USE_PYTHIA8

///////////////////////////////////////////// Event
#define EVTL_MAX_CPL_PART	10
#define EVTL_MAX_PART		2000
class CHEPEvtLite
{
public:
	//particle list
	int m_nParticles;				// number of records filled in 'm_Particles'
	CHEPParticleLite m_Particles[EVTL_MAX_PART];

	//event weight
	tReal wgt;						//XWGTUP
	tReal wgtErr;

	//process id, scale & constants
	int idProc;						//IDPRUP
	tReal scale;					//SCALUP
	tReal a_QED;					//AQEDUP
	tReal a_QCD;					//AQCDUP

private:
	class CCouplingsInfo {			//Cache: volatile numbers to cache event values
	public:
		tReal m_Q2abs;				// = -m_vecQ.GetSqr(). <-- Minus!
		tReal m_Cpls[EVTL_MAX_CPL_PART][EVTL_MAX_CPL_PART]; //used by GetCpl()
		bool m_CplsFilled;
		long m_iCplsFilledAfter;	// (used if 'm_CplsFilled') couplings are filled for particles n>X
									//		- used e.g. to avoid calculation of nuclon-vs.-final
									//		  couplings in parton-based processes
		tReal m_QCpls[EVTL_MAX_CPL_PART]; //used by GetQCpl()
		bool m_QCplsFilled;
		tReal m_UserCpls[EVTL_MAX_CPL_PART];
		CVM4Vec m_vecQ;
		CCouplingsInfo();
	};
	CCouplingsInfo m_cpls;			// output values (optional)

private:
	CHEPParticleLite* m_ChOrderBuf;	//[internal] used to avoid new[]/delete[] operations at every ChangeOrder call
	int m_szChOrderBuf;				//[internal] used to avoid new[]/delete[] operations at every ChangeOrder call

public:
	CHEPEvtLite();
	~CHEPEvtLite();

	//TODO: involve 'bBinaryAsDouble' mode (save/load as 'double' in binary mode), probably via CVMFileIn/Out.
	bool LoadEvent(std::ifstream* ifs); // 'binary' mode autodetected
	bool SaveEvent(std::ofstream* ofs, bool _binary, bool bSaveCouplingsCache=false);
	bool IsEventOK();

	//Change order/Insert particles
	bool ChangeOrder(CHEPParticleOrderLite* order);			//slow, clears couplings
	void InsertParticle(int pos,const CHEPParticleLite& p);	//e.g. to insert nuclon before parton line, clears couplings

public:
	//View/Debug procedures
	void LogEvent(std::ostream& os=std::cout);
	void LogCouplings(std::ostream& os=std::cout);

public:
	CHEPParticleLite* GetParticle(int i);
	int GetFirstInitialId();		// 'id(first=0)' of first particle with status == INITIAL
	int GetFirstFinalId();			// 'id(first=0)' of first particle with status == FINAL

public:
	tReal GetCpl(int i, int j) const; // uses m_Cpls if available
	void SetCpl(int i, int j, tReal val); // fills m_Cpls
	void FillCouplingsCache(bool bKeepFilled=false, bool bFillQCpls=true, int iFillAfter=0); // fills m_Cpls
	tReal GetQCpl(int i) const; // uses m_QCpls if available
	void SetQCpl(int i, tReal val); // fills m_QCpls
	void FillQ2Abs(tReal Q2Abs) { m_cpls.m_Q2abs=Q2Abs; }
	void FillQVec(const CVM4Vec& q) { m_cpls.m_vecQ=q; }
	tReal& GetQ2AbsCache() { return m_cpls.m_Q2abs; }
	const CVM4Vec& GetQVecCache() const { return m_cpls.m_vecQ; }
	void ClearCouplingsCache();
	/*
	*	GetCplSpecial: used by 'CVMPhaseSpaceGen' to update couplings cache (Don't use elsewhere!).
	*		- ignores 'm_CplsFilled' flag
	*		- recalcs and saves coupling if not filles previously (<0)
	*/
	tReal GetCplSpecial(int i, int j);
};

///////////////////////////////////////////// CHEPEvtLiteView
#ifdef Decl_CVMOutput
#define Decl_CHEPEvtLiteView
class CHEPEvtLiteView {
public:
	static const int MODE_YZ;
	static const int MODE_XY;
	static const int MODE_TZ;
	static const int MODE_XZ;
	// static const int MODE_EPS; to be merged with MODE_XX
public:
	// Fast entries
	static void DrawEvent(CHEPEvtLite* evt, const char* fn, int mode, int iMarked=0); /* iMarked=0 -> all finals */
	static void DrawEvent(CHEPEvtLite* evt, CVMOutput* out, int mode, int iMarked=0); /* iMarked=0 -> all finals */
public:
	CHEPEvtLiteView(const char* fn, int mode);
	CHEPEvtLiteView(CVMOutput* out, int mode);
	virtual ~CHEPEvtLiteView();
	void AddEventToView(CHEPEvtLite* evt, int iMarked=0, const char* label=NULL); // SaveView() expected afterwards.
	void SaveView(); // Saves view after several AddEventToView() calls.
private:
	void Init(); // Used by constructors only
	CVMOutput* m_Out;
	bool m_bExternalOutput; // don't delete 'm_Out' object
	int m_Mode;
	bool m_bBeamAxisZ;
	bool m_bOpen;
	int m_nEvtsDrawed;
public:
	double m_xValMax; //GeV
	double m_yValMax; //GeV
	CVMRGB m_clrTracksByPID[40];
	float m_widthTracksByPID[40];
	bool m_bDashedTracksByPID[40];
	CVMRGB m_clrTracksByMark[40];
	bool m_bColorTrackByPID; // otherwise by 'mark' (default: true)
	CVMRGB m_clrLabel; // The color of next event label
	int m_nLabelsPrinted;
	CVMRGB m_clrBackground;
	bool m_bDrawBox;
	float m_InnerRingSize; //(0..1), default: 0.2 for XY-plot, 0 for others
	bool m_bEquXYScale;
	bool m_bAllowOverMargins;
	bool m_bSkipNegativeMarks; //(false): don't draw particles with 'mark<0'
	bool m_bDrawFinalsOnly; //(true): don't draw particles with 'status != 1'
};
#endif //Decl_CVMOutput

///////////////////////////////////////////// Event utilites
class CHEPEvtLiteLib {
public:
	static bool ConvertEvtFile(const char* fnRead, const char* fnWrite, bool bToBinary);
	//static tReal GetRij(CHEPEvtLite* evt, int i, int j);//TODO: update implementation
	static int MarkDetectable(CHEPEvtLite* evt, int iMark);
	static int GetDaughters(CHEPEvtLite* evt, int iPart, int* iDaughter1, int* iDaughter2);
	static int GetNLeptTTbarDecay(CHEPEvtLite* evt); //-1 if no ttbar found or error spotted
	static int GetTTbarDecay(CHEPEvtLite* evt,
		int* iT, int* iTbar, int* iB1, int* iB2, int* iW1, int* iW2, int* iW1q1, int* iW1q2, int* iW2q1, int* iW2q2, int* iW1l, int* iW2l,
		int* iJet1=NULL
		); // returns leptons number
	static void LogEventStatistics(CHEPEvtLite* evt, std::ostream& os=std::cout, bool bFinalOnly=true);
};

#ifdef Decl_CVMGen
#define Decl_CVMPhaseSpaceGen

///////////////////////////////////////////// CHEPEvtLiteCut: Basic interface for cuts
#define Decl_CHEPEvtLiteCut
class CHEPEvtLiteCut {
protected:
	static tReal SmoothCutTest(tReal c, tReal w, tReal x, bool bMin);
public:
	//cut interface
	bool FitEvent(CHEPEvtLite* evt) const {					//true if event is passed, false if cutted
		//return(GetScale(evt)>(tReal)0.9999999);
		return(GetScale(evt)>(tReal)0.5);
	}

	//cut interface
	//virtual bool FitEvent(CHEPEvtLite* evt) const		= 0; //true if event is passed, false if cutted
	virtual tReal GetScale(CHEPEvtLite* evt) const		= 0; //'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
	virtual CHEPEvtLiteCut* Clone() const				= 0; //returns NULL in case of memory problems only
	virtual const char* ClassName() const				= 0; //used for RTTI
	virtual bool SetCutSmooth(bool bSmooth)				= 0; //'false' if not supported
	virtual void ChangeParticleOrder(const CHEPParticleOrderLite* order)	= 0;
};

///////////////////////////////////////////// CHEPProcessLite: Basic interface for processes
#define Decl_CHEPProcessLite
class CHEPProcessLite {
public:
	//CVMGen* m_Gen;
	//CHEPEvtLite* m_GenEvt;

	//process interface
	virtual void AddCut(const CHEPEvtLiteCut* cut)					= 0;	//to be used before 'Prepare'
	virtual bool Prepare(const CVM4Vec* p1, const CVM4Vec* p2,
						double fPrecision)							= 0;	//initializate + calculate c/s
	virtual bool GetCrossSection(tReal& cs, tReal* err=NULL) const	= 0;
	virtual bool GenEvent(CHEPEvtLite* evt)							= 0;
	virtual const char* ProcName() const							= 0;
	virtual int ProcID() const										= 0;	//-1 for sets and unknowns
};

///////////////////////////////////////////// Phase space generator
//
// Notes:
//	1. Uses one 'MC' and 'CVMGenFunc' object for all channels is used
//		To use MC/Func objects per channel --> modify this class
//
//	2. Same final state for all channels is presumed
//		To mix different processes --> use multiprocess classes
//
#define CVMPhSpGen_MAXN		20
class CVMPhaseSpaceGen {
public:
	class CBeamInfo {
	public:
		long id;
		CVM4Vec p;
		tReal mass;
		tReal scaleQ2;
		bool bUsePartons;
		CPDFLite* pdf;					// local CPDFLite-object
		CBeamInfo() { id=0; mass=0.; massSqr=0.; scaleQ2=0.; bUsePartons=false; pdf=NULL; pPabs=0.; }
		CBeamInfo(const CBeamInfo& b) { id=b.id; p=b.p; mass=b.mass; scaleQ2=b.scaleQ2; bUsePartons=b.bUsePartons; pdf=((b.pdf!=NULL)?(b.pdf->Clone()):NULL); pPabs=b.pPabs; }
		CBeamInfo& operator=(const CBeamInfo& b) { id=b.id; p=b.p; mass=b.mass; scaleQ2=b.scaleQ2; bUsePartons=b.bUsePartons; SetPDFConst(b.pdf); pPabs=b.pPabs; return *this; }
		~CBeamInfo() { if(pdf!=NULL){ delete pdf; pdf=NULL; } }
		bool Init();					// Update massSqr && pPabs + make self-check
		void SetPDFConst(const CPDFLite* _pdf) { if(_pdf!=pdf){ if(pdf!=NULL){delete pdf;} pdf=((_pdf!=NULL)?_pdf->Clone():NULL); } }
	public:
		tReal massSqr;					// local cache value, don't fill manually
		tReal pPabs;					// local cache value, don't fill manually
	};
public:
	class CEvtAlignment {				// positions of particles in event
	public:
		long posBeam[2];				// initial particles (or nuclons), filled if 'alg.m_bAddNuclonsToEvent'
		long posIn[2];					// initial particle (or partons)
		long posFin0;					// first final particle
		CEvtAlignment() {
			for(int i=0;i<2;i++){ posBeam[i]=i; posIn[i]=i; } posFin0=2;
		}
	};
public:
	class CAlgorithmInfo {				// Algorithm settings
		//
		// Notes:
		// ======
		//	 - Energy cuts on 'last' particle are ignored
		//	 - valX[2]-value is an interval, filled automatically
		//
	public:
		bool bUseMinMaxLastTheta;		// Apply cuts on the last particle angle
		tReal valMinMaxLastThCos[3];
		//[old] bUseMinMaxLastE=false;	// Apply cuts on the last particle energy (!!! may be combined with m_bUseMinMaxLastTheta)
		bool bUseMinMaxQ2;				// Apply cuts on Q2 (last final-last initial particles)
		tReal valMinMaxQ2[3];
		bool bUsePartFinF0eCuts;		// Apply Emin/Emax-cut on X-final particle (e.g. for IR-photon cut)
		tReal fPartFinF0eMinCuts[CVMPhSpGen_MAXN];
										// Note: eMin>0 is mandatory for usage of 'eMax'!
		tReal fPartFinF0eMaxCuts[CVMPhSpGen_MAXN];
		int iPartFinF0eOrder[CVMPhSpGen_MAXN];
										//dE or dE^-1, used together with 'm_fPartFinF0eMinCuts' only
		bool bInclude1dQ4;				// use 'd(1/Q2)' instead of d(Q^2)
		bool dQdCosTrg;					// elsewhere dQdM2
		bool bReqPreciseMasses;			// re-calc energy for every particle
		bool bCheckGenEvts;				// self-check
		bool bFillCouplingsCache;
		bool m_bAddNuclonsToEvent;		// add nuclon lines to event (as 'DOC' lines)
		//
		//Moved to locals: bool m_bSmartLowQRegion; //Calculate k.k couplings via q.k. Use with (m_bFillCouplingsCache && m_bUseMinMaxQ2)
		//
	public:
		CAlgorithmInfo();
		//[use default]CAlgorithmInfo& operator=(const CAlgorithmInfo& a);
		static CAlgorithmInfo Alg_UseLastThetaLimits(tReal minLastThCos, tReal maxLastThCos) {
			CAlgorithmInfo alg;
			alg.bUseMinMaxLastTheta=true; alg.valMinMaxLastThCos[0]=minLastThCos; alg.valMinMaxLastThCos[1]=maxLastThCos;
			return alg;
		}
		//bool Init();					// returns 'IsConsistent()'
		//TODO: bool InitSpecific...();

		// called by 'CVMPhaseSpaceGen::Init' only
	public:
		bool IsConsistent(				// init & check if parameter combination is allowed
			bool bZAlignedProcess,
			int Nf,
			tReal s
			);
	};

	class CChannelInfo {
	public:
		int idChannel;
		long idParton1;					// 0 if not used
		long idParton2;					// 0 if not used
	public:
		CChannelInfo(int _idChannel, long _idParton1, long _idParton2) { idChannel=_idChannel; idParton1=_idParton1; idParton2=_idParton2; }
	};

public:
	class CVMFuncMultiChannel : public CVMGen::CVMGenFunc {	// abstract class for SelectChannel() interface
	public:
		virtual long SelectChannel(tReal* params) = 0;
		CVMFuncMultiChannel();
	};

private:
	class CVMHepEvtFunc : public CVMGen::CVMGenFunc {
		tReal (vmdecl *m_pFunc)(CHEPEvtLite*); // function pointer
		CVMPhaseSpaceGen* m_GenRef;
	public:
		CVMHepEvtFunc(CVMPhaseSpaceGen* _GenRef, tReal(vmdecl *func)(CHEPEvtLite*));
		virtual tReal Call(tReal* params);
		virtual CVMGen::CVMGenFunc* Clone() const;
	private:
		CVMHepEvtFunc(const CVMHepEvtFunc& func) { m_GenRef=NULL; m_pFunc=NULL; /*use Clone()*/ }
	};
	//Nt: waiting for 'C++0x', now -> public: FillEvent
	//friend class CVMHepEvtFunc;

	///////////////////////////////////////////////////////////////////////////
	//                           generator object                            //
	///////////////////////////////////////////////////////////////////////////
private:
	CVMGen* m_Gen;
	bool m_bGenIsExt;					// m_Gen is external ref, do not delete
public:
	CVMGen* GetGenRef() { return m_Gen; }

	///////////////////////////////////////////////////////////////////////////
	//                             input values                              //
	///////////////////////////////////////////////////////////////////////////
private:
	CBeamInfo m_beam[2];
	bool m_bMultiChannel;				// == m_beam[0].bUsePartons || m_beam[1].bUsePartons
	std::vector<CChannelInfo> channels;	// channels list (used if 'm_bMultiChannel')
	CVMFuncMultiChannel* m_func_multi;	// used (and inited) in 'm_bMultiChannel' mode only

	int m_Nf;							// number of finals
	int m_nDim;							// m_Gen's dimension
	tReal m_MassFin[CVMPhSpGen_MAXN];	// masses of final particles
	long m_idFin[CVMPhSpGen_MAXN];		// IDs of  final particles
	CEvtAlignment align;				// positions of particles in event

	// Internal info
	CAlgorithmInfo alg;
	bool m_bAzimutalSym;				// Azimutal symmetry of the process
	bool m_bInitOk;						// if Init() passed successfully
	tReal m_lastParams[CVMGen_MAXDIM];	// last parameters passed to FillEvent()
										// used to avoid extra computation in various cases:
										//	- by Func in 'SelectChannel' procedure (after 'GenEvent')
										//	- by GenEvent to avoid extra FillEvent call

	///////////////////////////////////////////////////////////////////////////
	//                        output values: event                           //
	///////////////////////////////////////////////////////////////////////////
	/*
	*	Notes on event object conventions:
	*	==================================
	*		1. At every event 'm_GenEvt' object is re-filled with initial particles
	*				==> it may be different at every event (or changed between events)
	*		2. To use external event
	*				call SetEventExternalRef(evt) before 'GenEvent' and
	*				call SetEventExternalRef(NULL) afterwards to avoid further usage of (possibly) broken ref.
	*		3. In 'Init()' the internal instance of event (~1Mb) is created by default ==>
	*				==> delete it after (or avoid before) 'Init' by calling 'SetEventExternalRef(NULL)'
	*				That is important to save memory in multi-process genereator
	*		TODO: set 'const' in 'GetLastEvent' and all 'CalcDiffCS(const evt)'
	*/
private:
	CHEPEvtLite* m_GenEvt;				// event generated (incl. 'm_GenEvt->wgt')
	bool m_bEvtExternalRef;				// if 'm_GenEvt' an external object
public:
	CHEPEvtLite* GetLastEvent() { return m_GenEvt; }
										// get event generated, Note: MAY BE NULL (after SetEventRef(NULL) call)
	void SetEventRefExternal(CHEPEvtLite* evt);	
										//set external event ref., may be NULL
	void SetEventRefInternal();
										//set internal event ref.

	///////////////////////////////////////////////////////////////////////////
	//                     internal/debug output values                      //
	///////////////////////////////////////////////////////////////////////////
private:
	tReal m_EvtWgtDebug;				// Includes 'overweight', to be used for debug ONLY.
	//[old]tReal m_EvtWgt;				// Filled by GenEvent/GenEventWgt()
	//[old]tReal m_EvtWgtErr;			// Filled by GenEvent/GenEventWgt()
public:
	// Nt: 'm_PhSpaceWgt' must be public for Func-classes (whitch may not be friends until 'C++0x')
	tReal m_PhSpaceWgt;					// Filled by FillEvent() for internal use

	///////////////////////////////////////////////////////////////////////////
	//                         generation statictics                         //
	///////////////////////////////////////////////////////////////////////////
public:
	long m_nEvtsFailed;					// Error counter
	void IncEvtsFailed() { if(m_nEvtsFailed<LONG_MAX)m_nEvtsFailed++; }
	tReal m_MaxPhotonMass;
	tReal m_MaxParticleMassError;		// = M/Mcorrect-1
	long m_nFillEvtCalls;
	long m_nEmptyEvents;				// number of events failed to be created (e.g. due to internal cuts)

	///////////////////////////////////////////////////////////////////////////
	//                           calculation cache                           //
	//                                                                       //
	// note: cache is uniform for all parton-channels (m_{parton} === 0)     //
	///////////////////////////////////////////////////////////////////////////
private:
	tReal m_s;							// |pIn1+pIn2|^2, or \hat{s} for parton-based process
	tReal m_sSqrt;						// sqrt(m_s)
										//		- filled in 'UpdateCalculationCache'
										//		- updated in 'FillEvent' for parton processes ( as \hat{s} )
	tReal m_MassFinSqr[CVMPhSpGen_MAXN];
	tReal m_M2kMin[CVMPhSpGen_MAXN];
	tReal m_M1M2Diff2;					//(m0-m1)^2 cache
	bool UpdateCalculationCache(		//TODO: move this block to 'algorithm' object (as virtual)!
		bool bInitChngd, bool bFinChngd);

	///////////////////////////////////////////////////////////////////////////
	//                              debug info                               //
	///////////////////////////////////////////////////////////////////////////
public:
	bool m_bLogEvents;
	int m_nLogNFirstEvents;
	int m_iLogEventsPrecision;			//-1 if use default output
	long m_nHaltAfter;
	long m_nDebugCounter[10];			// debug counters

	///////////////////////////////////////////////////////////////////////////
	//                           basic operations                            //
	///////////////////////////////////////////////////////////////////////////
public:
	CVMPhaseSpaceGen();
	~CVMPhaseSpaceGen();
	bool Init(
		CBeamInfo& beam1, CBeamInfo& beam2,
		int Nf,							//number of final particles
		tReal* _MassFin,				//final particle masses
		long* _idFin,					//final particle IDs (may be NULL)
		const CAlgorithmInfo& _alg,		//
		tReal(vmdecl *func)(CHEPEvtLite*),
		CVMGen* _GenExt = NULL			//external MC core (if required)
		) { return Init(Nf,_MassFin,_idFin,beam1,beam2,_alg,func,NULL,NULL,_GenExt); }
	bool Init(
		CBeamInfo& beam1, CBeamInfo& beam2,
		int Nf,							//number of final particles
		tReal* _MassFin,				//final particle masses
		long* _idFin,					//final particle IDs (may be NULL)
		const CAlgorithmInfo& _alg,		//
		const CVMGen::CVMGenFunc* func,
		CVMGen* _GenExt = NULL			//external MC core (if required)
		) { return Init(Nf,_MassFin,_idFin,beam1,beam2,_alg,NULL,func,NULL,_GenExt); }
	bool InitMultiChannel(
		CBeamInfo& beam1, CBeamInfo& beam2,
		int Nf,							//number of final particles
		tReal* _MassFin,				//final particle masses
		long* _idFin,					//final particle IDs (may be NULL)
		const CAlgorithmInfo& _alg,		//
		CVMFuncMultiChannel* func,
		CVMGen* _GenExt = NULL			//external MC core (if required)
		) { return Init(Nf,_MassFin,_idFin,beam1,beam2,_alg,NULL,NULL,func,_GenExt); }
	bool AddChannel(int id,
					long idParton1,		// 0 if not used
					long idParton2);	// 0 if not used
	tReal Integrate(tReal _Precision=0.01);
	bool GenEvent();

	///////////////////////////////////////////////////////////////////////////
	//                         additional operations                         //
	///////////////////////////////////////////////////////////////////////////
public:
	bool GenEventWgt();							// Generate weighted event
	bool GenEvtDebug();							// Fast & incorrect events for viewing the integration grid.
	bool CheckEvent();							// self check
	void LogStatistics();						// push internal info to std::cout

	///////////////////////////////////////////////////////////////////////////
	//                            implementation                             //
	///////////////////////////////////////////////////////////////////////////
private:
	bool Init(
		int Nf,									//number of final particles
		tReal* _MassFin,						//final particle masses
		long* _idFin,							//final particle IDs (may be NULL)
		CBeamInfo& beam1, CBeamInfo& beam2,
		const CAlgorithmInfo& _alg,				//
		tReal(vmdecl *func)(CHEPEvtLite*),		//use either 'func' or 'gen_func' or 'multi_func'
		const CVMGen::CVMGenFunc* gen_func,		//use either 'func' or 'gen_func' or 'multi_func'
		CVMFuncMultiChannel* func_multi,		//use either 'func' or 'gen_func' or 'multi_func'
		CVMGen* _GenExt							//external MC core (if required)
		);
	void SaveLastParams(tReal* params) {		// save parameters as 'last' (called by 'FillEvent' only)
		for(int iTmp=0;iTmp<m_nDim;iTmp++){ m_lastParams[iTmp]=params[iTmp]; }
	}
	void ClearLastParams() {					// clear 'last' parameters (e.g. if 'FillEvent' failed)
		m_lastParams[0]=-1.;					// that's enough
	}
	//bool FillInitialVectors(tReal* params);	// fills initial particles in m_GenEvt
public:
	int GetNDim() { return m_nDim; }			//
	bool AreLastParams(tReal* params);			// check parameters to previous 'FillEvent' call
	bool FillEvent(								// fills m_GenEvt
		tReal* params,
		int idChannel=-1
		);
	//asdfg: old description!
												//[old] Note:
												//[old]		- doesn't fills parton IDs ==> use 'FillChannelInfo' afterwards
												// Note:
												//		- chnl==-1 is allowed for parton-based process,
												//		  <==> if PDF value is multiplied by user (manually)
												// TODO:
												//		- move this function to 'algorithm' object (as virtual)!
private:
	bool FillChannelInfo(int idChannel);		// fills parton IDs in multichannel mode

	///////////////////////////////////////////////////////////////////////////
	//                                 tests                                 //
	///////////////////////////////////////////////////////////////////////////
public:
	tReal TestPointValue(double* params);		// for tests only
	static tReal EvtFuncUnit(CHEPEvtLite* evt); // To calc phace space & tests
	static void TestVolume(int nFin=3);

	///////////////////////////////////////////////////////////////////////////
	//                                [todo?]                                //
	///////////////////////////////////////////////////////////////////////////
private:
	//[todo?] Set order of particle in final 'evt'-object
	//	Probem is:
	//		- change all m_GenEvt->m_Particles[n] to m_GenEvt->m_Particles[m_SaveParticleOrder[n]]
	//		- ALL SetCpl, etc. functions set to m_SaveParticleOrder-usage
	//		- understand how to deal with Q2-saved
	//	Best(?) solution is to sort event object by evt->ChangeOrder()
	//[todo?]int m_SaveParticleOrder[CVMPhSpGen_MAXN];	//order of storing particles in event (ONLY!)
public:
	//[todo?]bool SetSaveParticleOrder(int* order);		//to be calles BEFORE 'Init' only. {In0,In1,Fin0,Fin1,Fin2...}
};
#endif // #ifdef Decl_CVMGen

#ifdef Decl_CVMPhaseSpaceGen

//
//	Cuts (basic interface see above)
//
#ifdef Decl_CHEPEvtLiteCut
///////////////////////////////////////////// CHEPEvtLiteCutSet: Container class for several cuts
class CHEPEvtLiteCutSet : public CHEPEvtLiteCut {
	std::vector< CHEPEvtLiteCut* > m_Cuts;
public:
	CHEPEvtLiteCutSet();
	virtual ~CHEPEvtLiteCutSet();
	static const char* CLASS_NAME;

	void AddCut(const CHEPEvtLiteCut* cut);
	int GetNCuts() const;
	CHEPEvtLiteCut* GetCut(int i);
	const CHEPEvtLiteCut* GetCutConst(int i) const;
	void RemoveAll();

	void CopyFrom(const CHEPEvtLiteCutSet* cuts);

	//cut interface
	//virtual bool FitEvent(CHEPEvtLite* evt) const;	//true if event is passed, false if cutted
	virtual tReal GetScale(CHEPEvtLite* evt) const;		//'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
	virtual CHEPEvtLiteCut* Clone() const;
	virtual const char* ClassName() const;				//used for RTTI
	virtual bool SetCutSmooth(bool bSmooth);			//'false' if not supported
	virtual void ChangeParticleOrder(const CHEPParticleOrderLite* order);
};

///////////////////////////////////////////// CHEPEvtLiteCutE: Energy cut
class CHEPEvtLiteCutE : public CHEPEvtLiteCut {
	int m_id;										//particle id
	tReal m_EMin;									//-1. if unused
	tReal m_EMax;									//-1. if unused
	bool m_bSmooth;
public:
	CHEPEvtLiteCutE(int id, tReal EMin, tReal EMax);
	virtual ~CHEPEvtLiteCutE();
	static const char* CLASS_NAME;
	static CHEPEvtLiteCutE* CreateCutEMin(int id, tReal EMin);
	static CHEPEvtLiteCutE* CreateCutEMax(int id, tReal EMax);

	//
	int GetId() const { return m_id; }
	tReal GetEMin() const { return m_EMin; }
	tReal GetEMax() const { return m_EMax; }


	//cut interface
	//virtual bool FitEvent(CHEPEvtLite* evt) const;	//true if event is passed, false if cutted
	virtual tReal GetScale(CHEPEvtLite* evt) const;		//'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
	virtual CHEPEvtLiteCut* Clone() const;
	virtual const char* ClassName() const;				//used for RTTI
	virtual bool SetCutSmooth(bool bSmooth);			//'false' if not supported
	virtual void ChangeParticleOrder(const CHEPParticleOrderLite* order);
};

///////////////////////////////////////////// CHEPEvtLiteCutPt: Pt cut
class CHEPEvtLiteCutPt : public CHEPEvtLiteCut {
	int m_id;										//particle id
	tReal m_PtMin;									//-1. if unused
	tReal m_PtMax;									//-1. if unused
	bool m_bSmooth;
public:
	CHEPEvtLiteCutPt(int id, tReal PtMin, tReal PtMax);
	CHEPEvtLiteCutPt(int id, tReal PtMin);
	virtual ~CHEPEvtLiteCutPt();
	static const char* CLASS_NAME;
	static CHEPEvtLiteCutPt* CreateCutPtMin(int id, tReal PtMin);
	static CHEPEvtLiteCutPt* CreateCutPtMax(int id, tReal PtMax);

	//cut interface
	//virtual bool FitEvent(CHEPEvtLite* evt) const;	//true if event is passed, false if cutted
	virtual tReal GetScale(CHEPEvtLite* evt) const;		//'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
	virtual CHEPEvtLiteCut* Clone() const;
	virtual const char* ClassName() const;				//used for RTTI
	virtual bool SetCutSmooth(bool bSmooth);			//'false' if not supported
	virtual void ChangeParticleOrder(const CHEPParticleOrderLite* order);
};

///////////////////////////////////////////// CHEPEvtLiteCutEta: cut on abs(Eta)
class CHEPEvtLiteCutEta : public CHEPEvtLiteCut {
	int m_id;										//particle id
	tReal m_EtaMin;									//-1. if unused
	tReal m_EtaMax;									//-1. if unused
	bool m_bSmooth;
public:
	CHEPEvtLiteCutEta(int id, tReal EtaMin, tReal EtaMax);
	CHEPEvtLiteCutEta(int id, tReal EtaMinMax);
	virtual ~CHEPEvtLiteCutEta();
	static const char* CLASS_NAME;
	static CHEPEvtLiteCutEta* CreateCutPtMin(int id, tReal EtaMin);
	static CHEPEvtLiteCutEta* CreateCutPtMax(int id, tReal EtaMax);

	//cut interface
	//virtual bool FitEvent(CHEPEvtLite* evt) const;	//true if event is passed, false if cutted
	virtual tReal GetScale(CHEPEvtLite* evt) const;		//'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
	virtual CHEPEvtLiteCut* Clone() const;
	virtual const char* ClassName() const;				//used for RTTI
	virtual bool SetCutSmooth(bool bSmooth);			//'false' if not supported
	virtual void ChangeParticleOrder(const CHEPParticleOrderLite* order);
};

///////////////////////////////////////////// CHEPEvtLiteCutDR: DeltaR cut
class CHEPEvtLiteCutDR : public CHEPEvtLiteCut {
	int m_id1;										//particle id #1
	int m_id2;										//particle id #2
	tReal m_DRMin;									//-1. if unused
	tReal m_DRMax;									//-1. if unused
	bool m_bSmooth;
public:
	CHEPEvtLiteCutDR(int id1, int id2, tReal DRMin, tReal DRMax);
	virtual ~CHEPEvtLiteCutDR();
	static const char* CLASS_NAME;
	static CHEPEvtLiteCutDR* CreateCutDRMin(int id1, int id2, tReal DRMin);
	static CHEPEvtLiteCutDR* CreateCutDRMax(int id1, int id2, tReal DRMax);

	//cut interface
	//virtual bool FitEvent(CHEPEvtLite* evt) const;	//true if event is passed, false if cutted
	virtual tReal GetScale(CHEPEvtLite* evt) const;		//'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
	virtual CHEPEvtLiteCut* Clone() const;
	virtual const char* ClassName() const;				//used for RTTI
	virtual bool SetCutSmooth(bool bSmooth);			//'false' if not supported
	virtual void ChangeParticleOrder(const CHEPParticleOrderLite* order);
};
#endif //#ifdef Decl_CHEPEvtLiteCut

//
//	Processes (basic interface see above)
//
#ifdef Decl_CHEPProcessLite
///////////////////////////////////////////// CHEPProcessLiteSet: Container class for several processes
class CHEPProcessLiteSet : public CHEPProcessLite {
	std::vector< CHEPProcessLite* > m_subProcs;
public:
	CHEPProcessLiteSet();
	virtual ~CHEPProcessLiteSet();
	static const char* CLASS_NAME;

	void AddProcess(CHEPProcessLite* proc);			// deletes process after use
	const CHEPProcessLite* GetProcess(int i) const;
	int GetProcessNum() const;

	//process interface
	virtual void AddCut(const CHEPEvtLiteCut* cut);
	virtual bool Prepare(const CVM4Vec* p1, const CVM4Vec* p2, double fPrecision);
	virtual bool GetCrossSection(tReal& cs, tReal* err=NULL) const;
	virtual bool GenEvent(CHEPEvtLite* evt);
	virtual const char* ProcName() const;
	virtual int ProcID() const;
};

///////////////////////////////////////////// CHEPProcessLiteInst: single process class
class CHEPProcessLiteInst : public CHEPProcessLite {
protected:
	class CHEPProcessLiteInstFunc : public CVMGen::CVMGenFunc {
		tReal (vmdecl *m_pFunc)(CHEPEvtLite*);
		CHEPProcessLiteInst* m_proc;
		CVMPhaseSpaceGen* m_ph;
	public:
		CHEPProcessLiteInstFunc(CVMPhaseSpaceGen* ph, CHEPProcessLiteInst* proc, tReal(vmdecl *func)(CHEPEvtLite*));
		virtual tReal Call(tReal* params);
		virtual CVMGen::CVMGenFunc* Clone() const;
	};
protected:
	CHEPEvtLiteCutSet m_cuts;
	CVMGen* m_gen;							//contains cross section
	CVMPhaseSpaceGen* m_ph;					//constains 'm_gen' as external generator
	//tReal(vmdecl *m_func)(CHEPEvtLite*);	//static function ref
public:
	CHEPProcessLiteInst();
	virtual ~CHEPProcessLiteInst();

	//process interface
	virtual void AddCut(const CHEPEvtLiteCut* cut);
	virtual bool Prepare(const CVM4Vec* p1, const CVM4Vec* p2, double fPrecision);
	virtual bool GetCrossSection(tReal& cs, tReal* err=NULL) const;
	virtual bool GenEvent(CHEPEvtLite* evt);
	virtual const char* ProcName() const;
	virtual int ProcID() const;
};
#endif //#ifdef Decl_CHEPProcessLite

#endif // #ifdef Decl_CVMPhaseSpaceGen

#endif // !defined(HEPEvtLite_H_INCLUDED_)
