////////////////////////////////////////////////////////
//	VMJetReco.h
//	Author: V.Makarenko (Vladimir.Makarenko@cern.ch)
////////////////////////////////////////////////////////

//[not here!]#ifdef VM_HEP_LIB_FULL
#if !defined(VMJetReco_H_INCLUDED_)
#define VMJetReco_H_INCLUDED_
/*
*	Pre-includes:
*	#include <iostream>
*	#include <fstream>
*	#include "HEPEvtLite.h"
*/

#define NJETS_MAX	2000
//////////////////////////////////////////////////////////////////////////////////////	CVMJetAlgorithm
class CVMJetAlgorithm {
public:
	class CVMJetRec{
	public:
		CVM4Vec m_Mom;
		int m_nParticles;
		//int m_iCombineMode; // 0: momentum only, 1: default, 2: FJ
		bool m_bUseFullRecord; // calc Eta,Phi,Et etc.
		// Warning: m_Eta, m_Et & m_Phi are combined 'fast' (not precisely!)
		double m_Eta;
		double m_Et;
		double m_Phi; //azimutal angle
		CVMJetRec();
		CVMJetRec(const CVMJetRec& jet);
		CVMJetRec& operator=(const CVMJetRec& jet);
		CVMJetRec& operator+=(const CVMJetRec& jet); //combine jets
		void Set(const CHEPParticleLite& part); // Set as '1-particle' jet
	};
	class CVMParticleFilter{
	public:
		// Particle selection	//default value
		double m_PMaxEta;		//-1	// max. particle eta (-1 if no aplicable)
		double m_PMinMomAbs;	//-1	// min. particle momentum (-1 if no aplicable)
		double m_PMinPT;		//-1	// min. particle Pt (-1 if no aplicable)
		bool m_PNoLeptons;		//false	// no leptons in jets (like in FastJet example
		int m_UseMarked;		//0		// use marked ('mark'==m_UseMarked) particles only
		CVMParticleFilter();
	};
protected:
	// input
	CHEPEvtLite* m_evt;	// volatile ref., do not use it outside Run() or delete
public:
	// Options
	CVMParticleFilter m_PartFilter;	// Particle selection filter
	bool m_bSortFinalJets;			// Sort m_Jets[] by Pt (default: false)
	double m_MinJetPt;				// Makes sorting by Pt faster (default: 0)
	bool m_bKeepPart2JetInfo;		// Keep particle-jet information (default: false)

	// Output
	int m_nJets;					// final number of jets
	int m_nParticlesUsed;			// to check NJETS_MAX selection
	double m_SumEnergy;				// Total energy of jets (all 'detectabe' particles)
	const CVMJetRec& Jet(int i) { return m_Jets[i]; }
	int GetJetForParticle(int i) { return m_idJetForParticle[i]; }

	// Misc. output operations
	int CalcJetsNumber(double minJetPt, int minNParts=1); //To be used after Run()
	void FillJetEvent(CHEPEvtLite* evtJets); //Fill 'evtJets' with jets (e.g. for further save&load), nParticles->'Mark'
	void DrawEvent(const char* title, bool bDrawUnusedParticles=true); //including jets
public:
	CVMJetAlgorithm(bool bUse2JetsCache);
	virtual ~CVMJetAlgorithm();
	virtual bool Run(CHEPEvtLite* evt); // Default: pushes all 'detectable' particles to different jets
private:
	CVMJetRec m_Jets[NJETS_MAX];	// jets info. To be modified by local functions only!
	int m_idJetForParticle[EVTL_MAX_PART]; // Particle-in-jet information (-1: unused or 'not-detectable')
protected:
	virtual void SelectParticlesToJets(); //initial jets filling
	double Get2JetsRes(int i, int j) { return ((m_bUse2jCache)?(m_Cache2Jets[i*NJETS_MAX+j]):(Calc2JetsRes(i,j))); }
	virtual double Calc2JetsRes(int i, int j)=0;
	void CombineJets(int i, int j);
	void FixateJet(int i);
	void SwapJets(int i, int j);
	void SortFinalJets();	// by Pt
protected:
	int m_nFixedJets;		// Number of 'final'-marked jets (to ignore cache values at i<m_nFixedJets)
	bool m_bMomOnlyUsed;	// For faster evaluation of Durham-like algorithms
	bool m_bUse2jCache;		// ==(m_Cache2Jets!=NULL);
	double* m_Cache2Jets;	// 2-jets resolutions cache (allocated upto NJETS_MAX*NJETS_MAX*sizeof(double))
};

//////////////////////////////////////////////////////////////////////////////////////	CVMJetKT
class CVMJetKT : public CVMJetAlgorithm {
public:
	// Parameters			//(default:)
	double m_R;				//0.7
public:
	CVMJetKT(bool bUseCache=true);
	virtual ~CVMJetKT();
	virtual bool Run(CHEPEvtLite* evt);
protected:
	virtual double Calc2JetsRes(int i, int j);
	double CalcJetRes(int i);// { return m_Jets[i].m_Et*m_R; }
};

//////////////////////////////////////////////////////////////////////////////////////	CVMJetDurham
class CVMJetDurham : public CVMJetAlgorithm {
public:
	/*
	*	WARNING: algorithm currently stops when 'm_yCut' is reached.
	*	I.e. the 'm_nJetsRequired' is ignored if 'm_yCut>0'.
	*	Because the continuation of combining will destroy 'final' jeys & jet-to-particle information.
	*	To fix it one need to implement 'SaveJetsCopy' & 'RestoreJetsCopy' in CVMJetAlgorithm.
	*	Or create separated 'working' and 'final' jet&particles arrays in CVMJetAlgorithm
	*	And operate with 'working' copy of jets throughout the reconstruction.
	*/
	// Parameters			//(default:)
	double m_yCut;			//-1	// yCut to form the final jets array
	int m_nJetsRequired;	//1		// fill m_yResN array down to {m_nJetsRequired+1} jets

	// Output
	double m_yResN[NJETS_MAX];		// Minimal resolution of (N+1) jets (filled from [1] for 2 jets to [m_nParticlesUsed-1] for m_nParticlesUsed jets)

	// Additional parameters (unused by default)
	int m_nMinPartInJet;	//1		// store N{jets with Npart>m_nMinPartInJet} into m_nMultiPartJets[N jets total] 
	double m_HardJetMinMom;	//0.	// minimal 'hard jet' momentum to fill m_nHardJets[N]

	// Additional output
	int m_nMultiPartJets[NJETS_MAX];// Number of jets with Npart>m_nMinPartInJet
	int m_nHardJets[NJETS_MAX];		// Number of 'hard' jets (P>m_HardJetMinMom)
public:
	CVMJetDurham(bool bUseYCache=true);
	virtual ~CVMJetDurham();
	virtual bool Run(CHEPEvtLite* evt);
public:
	int GetNJetsRes(double yCut);	// get Njets from m_yResN[] array
protected:
	virtual double Calc2JetsRes(int i, int j);	// Durham Algorithm yres
};

/*
class CVMJetDurham {
public:
	// input
	CHEPEvtLite* m_evt;	// volatile ref., do not use it outside run() or delete

	// jet selection		(default)
	int m_nMinJets;			//2		// minimum number of jets
	int m_nMinPartInJet;	//1		// store N{jets with Npart>m_nMinPartInJet} into m_nMultiPartJets[N jets total] 
	double m_HardJetMom;	//5.	// minimal 'hard jet' momentum to fill m_nHardJets[N]

	// particle selection	(default)
	double m_PMaxEta;		//-1	// max. particle eta (-1 if no aplicable)
	double m_PMinMomAbs;	//0.05	// min. particle momentum (-1 if no aplicable)
	double m_PMinPT;		//-1	// min. particle Pt (-1 if no aplicable)
	bool m_PNoLeptons;		//false	// no leptons in jets (like in FastJet example)

	// additional
	bool m_bSortFinalJets;	//fakle	// (by Pt) --> m_pJetFnl[] and m_nPartJetFnl[]

	// output
	bool m_bFinalJetsFilled;		// if m_nJetFnl, m_pJetFnl[] and m_nPartJetFnl[] are filled
	int m_nJetFnl;					// final number of jets
	int m_nPartJetFnl[NJETS_MAX];	// number of particles in each jet
	CVM4Vec m_pJetFnl[NJETS_MAX];	// 4-vector of each jet
	double m_yResN[NJETS_MAX];		// Minimal resolution of (N+1) jets (filled from [1] to [m_nDetectables-1])
	int m_nMultiPartJets[NJETS_MAX];// Number of jets with Npart>m_nMinPartInJet
	int m_nHardJets[NJETS_MAX];		// Number of 'hard' jets (P>m_HardJetMom)
	int m_nDetectables;				// to check NJETS_MAX selection
	double m_Evis;					// visible energy

	// debug
	long m_nEventsNJETS_MAX;		// to check NJETS_MAX value selection
public:
	bool run(CHEPEvtLite* evt, double yCut=-1);
	int GetNJetsRes(double yCut);	// get Njets from m_yResN[] array
public:
	CVMJetDurham();
	~CVMJetDurham();
private:
	double* m_yCache;
	double CalcYRes(CVM4Vec* p1, CVM4Vec* p2, double evs); // Durham Algorithm yres
	bool RunDurhamImpl(double yCut);
	void SortFinalJets(); // by Pt
};
*/

#endif // !defined(VMJetReco_H_INCLUDED_)
//[not here!]#endif //#ifdef VM_HEP_LIB_FULL
