////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// VMOutput.h: interface for the CVMOutput class.
//
//////////////////////////////////////////////////////////////////////
//[not here!]#ifdef VM_HEP_LIB_FULL
#if !defined(CVMOutput_H_INCLUDED_)
#define CVMOutput_H_INCLUDED_
/*
*	Pre-includes:
*	#include <iostream>
*	#include <fstream>
*/

class CVMRGB {
public:
	float m_Red; float m_Green; float m_Blue;
public:
	CVMRGB();
	CVMRGB(float r, float g, float b);
	CVMRGB(const CVMRGB& c);
	CVMRGB &operator=(const CVMRGB& c);
	bool operator==(const CVMRGB& c) const;
	bool operator!=(const CVMRGB& c) const;
	CVMRGB& operator+=(const CVMRGB& c);		// MIX with other color
	CVMRGB operator+(const CVMRGB& c) const;	// MIX with other color. Note: "+=" is faster
	void Set(float r, float g, float b);
	bool IsValid() { return ((m_Red>=0.f)&&(m_Green>=0.f)&&(m_Blue>=0.f)); }
};

#define Decl_CVMOutput
class CVMOutputSettings {
public:
	static bool m_bSaveOutputTime; //Def: true, use 'false' for simplier comparison of plots
};

class CVMOutput {
	bool m_bStarted;
	bool m_bFinalized;
protected:
	double m_aspect_ratio;
	double m_xMin; // logical values for MoveTo/LineTo functions
	double m_xMax;
	double m_yMin;
	double m_yMax;
	double m_fPad2LogicalCoordScale; //scale from pad (0,200;0,100) to logical coordinates (0,1;0,0.6), usually <<1.
	double m_xCurr;
	double m_yCurr;
	CVMRGB m_clrDraw;
	CVMRGB m_clrBackground;
protected:
	double m_dLineWidth;
	bool m_bLineWidthChanged;
	float m_FontSize;
	bool m_bFontSizeChanged;
	int m_TextAlignX;	//(def:0) -1(left),0(center),1(right)
	int m_TextAlignY;	//(def:1) -1(bottom),0(center),1(up)
	float m_fTxtAngle;	//0..360, default: 0
	bool m_bUseMathTextMacros;		// Use Mathematica-like macros in eps for text printing
	int m_iFontFamily;	//FF_COURIER[default]
	bool m_bFontFamilyChanged;
	//
	//[Unused: 's' is written BEFORE next output]bool m_bBlockEnded;	//'stroke' is just written (for EPS-output)
	//
	float m_fBlockRotationAngle;	//0..360
public:
	static int FF_COURIER;
	static int FF_SYMBOL;
	static int FF_ARIAL;
	static const char* GetFontFamilyID(int iFontFamily);
public:
	static float LW_THIN;
	static float LW_NORMAL;
	static float LW_THICK;
public:
	CVMOutput();
	virtual ~CVMOutput();
	virtual void Start(); // e.g. open file
	virtual void Finalize(); // e.g. close file
	virtual void MoveTo(double x, double y)=0;
	virtual void LineTo(double x, double y)=0;
	//
	// Note: 'out::ArrowTo' moved 'CVMDrawLib::Arrow'
	//
	void Rect(double x1, double y1, double x2, double y2, bool bFill=false); // moves cursor to {x1,y1}
	virtual void Arc(double x, double y, double r, double phi1, double phi2); // angle in degee
	void Circle(double x, double y, double r) { Arc(x,y,r,0.,360.); }
	virtual void TextOut(double x, double y, const char* text, int len=-1)=0;
	void TextOutAdv(double x, double y, const char* text, bool bAllowYFit=false); //Parse TeX line (^,_,\gamma etc.)
	virtual double GetLetterWidth()=0; // average letter width
	virtual double GetLetterHeight()=0;
	float GetFontSize(){ return m_FontSize; }
	virtual float SetFontSize(float sz);
	void SetFontFamily(int iFontFamily);
	//
	double GetFontSizeScaled();		//in logical scale
	double GetFontWidthFactor();	//in logical scale
	//
	static const int TA_LEFT;
	static const int TA_BOTTOM;	//==TA_LEFT
	static const int TA_CENTER;
	static const int TA_RIGHT;
	static const int TA_UP;		//==TA_RIGHT
	int GetTextAlignX() { return m_TextAlignX; }
	int GetTextAlignY() { return m_TextAlignY; }
	int SetTextAlignX(int xAlign); //x{-1:left, 0:(def)center, 1:right)
	int SetTextAlignY(int yAlign); //y{-1:bottom, 0:center, 1:(def)up)
	//
	float SetTextDirection(float fTxtAngle);
	bool UseMathTextMacros(bool bUse);
	//
	virtual void StartBlockRotation(double x, double y, float fBlockAngle);	//TODO: allow recursion here!
	virtual void StopBlockRotation();
	//
	virtual void SetLineWidth(double w);
	double GetLineWidth() { return m_dLineWidth; }
	//
	static const int LINE_SOLID;
	static const int LINE_DASH;
	static const int LINE_DOT;
	static const int LINE_DASHDOT;
	static const int LINE_OTHER;	// set indirectly for specific styles (e.g. dots with non-default rate)
	/*
	*	Dots:
	*		fRate=1: (default) long-distance dots, good for small zoom
	*		fRate=1.3: medium-distance dots, good for high zoom
	*/
	void SetLineStyle(int lStyle, float fRate=1.f); // 'fRate' is relative: '0.5' mean 2 times denser as default.
	//void SetLineDashed() { SetLineStyle(CVMOutput::LINE_DASH); }
	//void SetLineDotted(float fDotRate=1.f) { SetLineStyle(CVMOutput::LINE_DOT,fDotRate); } // set short dash
	//void SetLineSolid() { SetLineStyle(CVMOutput::LINE_SOLID); } // remove 'dash'/'dots' style
	CVMRGB GetColorBackground() { return m_clrBackground; }
	CVMRGB SetColorBackground(CVMRGB rgb); //Doesn't fill background! Use 'Fill...Contour,..()' to fill.
	void FillBackground(CVMRGB rgb);	//Whole background
	//
	virtual void SetLineDash(int nElements, double* len)=0;
	virtual void SetColor(CVMRGB rgb);
	virtual void FillColor(double x1, double y1, double x2, double y2, CVMRGB rgb)=0;
	virtual void FillContour(CVMRGB rgb);
	void FillContour(); //uses 'm_clrBackground'
	virtual void BlockEnd()=0; // e.g. for sub-save or 'show' command. May be unused.
	//
	bool IsOpen(); // if m_bStarted && (!m_bFinalized)
	void GetMargins(double* xMin, double* xMax, double* yMin, double* yMax);
	void SetMargins(double xMin, double xMax, double yMin, double yMax);
public:
	static const CVMRGB clrRed;
	static const CVMRGB clrBlue;
	static const CVMRGB clrGreen;
	static const CVMRGB clrWhite;
	static const CVMRGB clrBlack;
	static const CVMRGB clrYellow;
	static const CVMRGB clrMagenta;
	static const CVMRGB clrCyan;
	static const CVMRGB clrGrey; //(0.75,0.75,0.75)
	static const CVMRGB clrGreyDark; //(0.5,0.5,0.5)
	static const CVMRGB clrGreyLight; //(0.90,0.90,0.90)
	static const CVMRGB clrGreyLightLight; //(0.95,0.95,0.95)
	static const CVMRGB clrOrange; //1,0.65,0
	static const CVMRGB clrLawnGreen; //'salat' (0.49,0.98,0)
	static const CVMRGB clrKhaki; //(0.9,0.82,0,58)
	static const CVMRGB clrBrown; //(0.58,0.25,0,11)
	static const CVMRGB clrPink; //(1,0.4,0.7) Original: (1,0.75,0.8)
	static const CVMRGB clrViolet; //(0.93,0.5,0.93)
	static const CVMRGB clrYellowGreen; //(0.6,0.8,0.2)
	//clrDefault: in some cases may works incorrectly
	static const CVMRGB clrDefault; //(-1.,-1.,-1.)
};

class CVMFileOutput : public CVMOutput {
	char* m_fn;
	std::ofstream* m_pOfs;
	bool m_bKeepOpened;
	char* m_buf;
	long m_lBufPos;
	long m_lBufLen;
protected:
	char m_TmpBuf[10000];
public:
	CVMFileOutput(const char* fn, bool bKeepOpened);
	virtual ~CVMFileOutput();
	virtual void SetUseBuffer(long lBufLen);
	virtual void Start();
	virtual void Finalize();
protected:
	virtual void Write(const char* str);
	virtual std::ofstream* GetStreamPtr();
};

class CVMOutputMath : public CVMFileOutput {
	int m_idCurrCell; //-1(no),0(graphics),1(text)
	int m_nCells;
public:
	CVMOutputMath(const char* fn);
	virtual ~CVMOutputMath();
	virtual void Start();
	virtual void Finalize();
	virtual void MoveTo(double x, double y);
	virtual void LineTo(double x, double y);
	virtual void Arc(double x, double y, double r, double phi1, double phi2); // angle in degee
	virtual void TextOut(double x, double y, const char* text, int len=-1);
	virtual void StartBlockRotation(double x, double y, float fBlockAngle);
	virtual void StopBlockRotation();
	virtual double GetLetterWidth();
	virtual double GetLetterHeight();
	virtual void SetLineDash(int nElements, double* len);
	virtual void SetColor(CVMRGB rgb);
	virtual void FillColor(double x1, double y1, double x2, double y2, CVMRGB rgb);
	virtual void FillContour(CVMRGB rgb);
	virtual void BlockEnd();
	void PrepareCell(); //create/update cell before output
	void NewPlotCell();
	void NewTextCell();
private:
	void FinalizeCell();
};

class CVMOutputEPS : public CVMFileOutput {
	// output layout margins (BoundingBox). Do not use it with MoveTo/LineTo functions.
	double m_xPadMin;
	double m_xPadMax;
	double m_yPadMin;
	double m_yPadMax;
public:
	CVMOutputEPS(const char* fn);
	virtual ~CVMOutputEPS();
	void GetBoundingBox(double* xBoxMin, double* xBoxMax, double* yBoxMin, double* yBoxMax);
	void SetBoundingBox(double xBoxMin, double xBoxMax, double yBoxMin); //scaled to current aspect ratio
	void SetBoundingBox(double xBoxMin, double xBoxMax, double yBoxMin, double yBoxMax); //change aspect ratio
	virtual void Start();
	virtual void Finalize();
	virtual void MoveTo(double x, double y);
	virtual void LineTo(double x, double y);
	virtual void Arc(double x, double y, double r, double phi1, double phi2); // angle in degee
	virtual void TextOut(double x, double y, const char* text, int len=-1);
	virtual void StartBlockRotation(double x, double y, float fBlockAngle);
	virtual void StopBlockRotation();
	virtual double GetLetterWidth();
	virtual double GetLetterHeight();
	virtual float SetFontSize(float sz);
	virtual void SetLineWidth(double w);
	virtual void SetLineDash(int nElements, double* len);
	virtual void SetColor(CVMRGB rgb);
	virtual void FillColor(double x1, double y1, double x2, double y2, CVMRGB rgb);
	virtual void FillContour(CVMRGB rgb);
	virtual void BlockEnd();
};

/*
class CVMOutputRecord : public CVMOutput {
private:
	class COutRec {
	public:
		int m_iType;
		int m_nArgs;
		double* m_Args;
	public:
		COutRec(int iType, int nArgs, double* args);
		virtual ~COutRec() { if(m_Args!=NULL)delete[]m_Args; }
	};
public:
	CVMOutputRecord();
	virtual ~CVMOutputRecord();
public:
	//TODO: add any non-virtual CVMOutput-functions into PlayRecord() implementation
	void PlayRecord(CVMOutput* out); //Incl. Get/Set Margins
private:
	void SaveRecord(int iType, int nargs, double* args);
public:
	//TODO: add any virtual CVMOutput-functions here implementation
	virtual void Start() { SaveRecord(1,0,NULL); }
	virtual void Finalize() { SaveRecord(2,0,NULL); }
	virtual void MoveTo(double x, double y)=0;
	virtual void LineTo(double x, double y)=0;
	virtual void Arc(double x, double y, double r, double phi1, double phi2); // angle in degee
	virtual void TextOut(double x, double y, const char* text, int len=-1)=0;
	virtual double GetLetterWidth()=0; // average letter width
	virtual double GetLetterHeight()=0;
	virtual float SetFontSize(double sz)=0;
	virtual void SetLineWidth(double w);
	virtual void SetLineDash(int nElements, double* len)=0;
	virtual void SetColor(CVMRGB rgb);
	virtual void FillColor(double x1, double y1, double x2, double y2, CVMRGB rgb)=0;
	virtual void BlockEnd()=0; // e.g. for sub-save or 'show' command. May be unused.
};
*/

class CVMDrawLib {
public:
	static void Arrow(CVMOutput& out, double x1, double y1, double x2, double y2,
					  double sz=-1., int iStyle=-1, bool bFillBcgr=false);
	//
	static double fAxisY;		//def: -1 for '0 at bottom' (+1 for '0 at top')
	static bool bParseTeX;		//def: false
	static double szShadow;		//def: 0
	static CVMRGB clrShadow;	//def: CVMRGB(0.5,0.5,0.5)
	static double gapLines;		//(def: 0.6) distance between lines in high-of-letter units
	static void Box(CVMOutput& out, double x1, double y1, double dx, double dy, const char* text);
};

class CVMOutputTest {
public:
	static void LetterSizeTest(CVMOutput* out);
	static void TestTextSizeAndAlign(CVMOutput* out);
};

#endif // !defined(CVMOutput_H_INCLUDED_)
//[not here!]#endif //#ifdef VM_HEP_LIB_FULL
