////////////////////////////////////////////////////////
//	HEPEvtLite.cpp
//	Author: V.Makarenko (Vladimir.Makarenko@cern.ch)
////////////////////////////////////////////////////////

#include <math.h> // for sqrt() in mass()
#include <iostream>
#include <fstream>
#include <iomanip>

#include "VMLib/include/vmlib.h"

#ifdef VM_USE_LOOPTOOLS
#ifdef VM_LOOPTOOLS_QUAD
#include "clooptools-q.h"
#else
#include "clooptools.h"
#endif //VM_LOOPTOOLS_QUAD
#endif //VM_USE_LOOPTOOLS

#if defined VM_USE_QD_REAL
tReal CVMConst::m_Pi			= qd_real::_pi;
bool CVMConst::m_bRecalcPi			= false;
#elif defined VM_USE_DD_REAL
tReal CVMConst::m_Pi			= dd_real::_pi;
bool CVMConst::m_bRecalcPi			= false;
#else
tReal CVMConst::m_Pi			= 3.1415926535897932385;
bool CVMConst::m_bRecalcPi			= true;	// or false if (sizeof(tReal)==8)
#endif //VM_USE_QD_REAL/VM_USE_DD_REAL
tReal CVMConst::Pi() { if(m_bRecalcPi){ m_Pi=atan((tReal)1.)*((tReal)4.); m_bRecalcPi=false; } return m_Pi; }

const tReal CVMConst::Euler			= 2.718281828459045;
const tReal CVMConst::Alpha			= 1./137.035999679;	//PDG'09
//Old: const tReal CVMConst::M_e = 5.109989e-4;	//PDG'09: 0.0005109989(10) +/- 0.00000000013
const tReal CVMConst::M_e			= 5.10998910e-4;	//PDG'09: 0.0005109989(10) +/- 0.00000000013
//const tReal CVMConst::M_e			= 1e-2;				//for tests
const tReal CVMConst::M_mu			= 0.105658367e0;	//PDG'09: 0.105658367 +/- 0.000000004
const tReal CVMConst::M_p			= 0.938272013;		//PDG'09: 938.272013 +/- 0.000023 MeV
const tReal CVMConst::ReClassic		= Alpha/M_e;		//similar to Alpha from PDG'09, 14.2806 CeV-1
const tReal CVMConst::GeV2pb		= 3.8937966e8;		//1 pb =10^-12 b =10^-9 mb
//Old: const tReal CVMConst::El = 0.302822;	//=sqrt(Alpha*4*Pi). Old value: 0.30286
const tReal CVMConst::El			= 0.302822120214;	//=sqrt(Alpha*4*Pi). Old value: 0.30286
//
const tReal CVMConst::MMom_p		= 2.792847356;		//magnetic moment of proton, PGD'09: 2.792847356 +/- 0.000000023
//const tReal CVMConst::MMom_p		= 2.79;				//[old]
//const tReal CVMConst::MMom_p		= 0.0;				//[debug]
//tReal mup=1.792; //Anomalous magnetic moment of proton
//tReal mup=2.79; //Magnetic moment of proton

/*
*	--- CAlHepLib ---
*/
bool CAlHepLib::m_Inited=false; //to avoid repetion of 'Init()' call
unsigned int CAlHepLib::m_QD_oldcw=0; //used by QD package initialization procedure
CVMFileIn* CAlHepLib::m_DebugIn		= NULL;
CVMFileOut* CAlHepLib::m_DebugOut	= NULL;
const int CAlHepLib::NUM_RND_BANKS	= 100;				//may be safely changed in 'HEPEvtLite.cpp' (but must be uniform for all runs!)
#ifdef VM_USE_PYTHIA8
Pythia8::Pythia* CAlHepLib::pythia	= NULL;				//Pythia instance
#endif //#ifdef VM_USE_PYTHIA8

bool CAlHepLib::SelectRndBank(const char* bank_no) {	//to be called first (before any RND-use), BEFORE 'CAlHepLib::Init()'!
	if(bank_no==NULL)return false;
	if(m_Inited) { std::cout<<"Error in CAlHepLib::SelectRndBank(). Must be called BEFORE 'CAlHepLib::Init()'! RND-Bank is unaffected"<<std::endl; return false; }
	bool bArgOk=true;
	for(int i=0;;){
		//Nt. first char may not be '\0'
		if((bank_no[i]<'0')||(bank_no[i]>'9')) { bArgOk=false; break; }
		i++; if(bank_no[i]=='\0')break;
	}
	if(!bArgOk) { std::cout<<"Error in CAlHepLib::SelectRndBank(). RND-Bank ID parameter \""<<bank_no<<"\" is not a number. "<<std::endl; return false; }
	int idRndBank=atol(bank_no);
	if((idRndBank < 0) || (idRndBank >= CAlHepLib::NUM_RND_BANKS)) { std::cout<<"Error in CAlHepLib::SelectRndBank(). RND-Bank ID parameter \""<<bank_no<<"\" mut be within [0.."<<CAlHepLib::NUM_RND_BANKS-1<<"] interval. "<<std::endl; return false; }
	std::cout<<"---- Rnd bank set to "<<idRndBank<<" (max: "<<CAlHepLib::NUM_RND_BANKS<<") ----"<<std::endl;
	CVMRnd::ResetRndGlobal(idRndBank,CAlHepLib::NUM_RND_BANKS);
	return true;
}
bool CAlHepLib::Init() { //Inits LoopTools, QD, Pythia etc.
	if(m_Inited)return true;

	//Init QD
#if defined VM_USE_QD_REAL
	fpu_fix_start(&m_QD_oldcw);
#elif defined VM_USE_DD_REAL
	fpu_fix_start(&m_QD_oldcw);
#endif //VM_USE_QD_REAL,VM_USE_DD_REAL,

	//Init LoopTools (if used)
	bool bLTOk=CAlHepLib::LTInit(); //'false' if no LT are used ==> ignore 'bLTOk'

	//Init Pythia
#ifdef VM_USE_PYTHIA8
	if(CAlHepLib::pythia != NULL){ delete CAlHepLib::pythia; CAlHepLib::pythia=NULL; }
#ifdef VM_PYTHIA8_XML_PATH
	CAlHepLib::pythia = new Pythia8::Pythia(VM_PYTHIA8_XML_PATH);
#else //#ifdef VM_PYTHIA8_XML_PATH
	CAlHepLib::pythia = new Pythia8::Pythia()
#endif //#ifdef VM_PYTHIA8_XML_PATH
#endif //#ifdef VM_USE_PYTHIA8

	//Others
	CVMConst::Pi(); //Pre-calculate 'Pi'
	//Debug
	//m_DebugIn = new CVMFileIn("debug.txt");
	//m_DebugOut = new CVMFileOut("debug.txt");
	//
	unsigned long ver=GetVersion(); std::cout<<"ALHEP library ver.: "<<((ver>>16)&0xff)<<"."<<((ver>>8)&0xff)<<"."<<(ver&0xff)<<std::endl;
	m_Inited=true; return true;
}
void CAlHepLib::Finalize() { //Finalizes LoopTools, QD, etc.
	m_Inited=false;

	//Finalize QD
#if defined VM_USE_QD_REAL
	fpu_fix_end(&m_QD_oldcw);
#elif defined VM_USE_DD_REAL
	fpu_fix_end(&m_QD_oldcw);
#endif //VM_USE_QD_REAL,VM_USE_DD_REAL,

	//Finalize LoopTools
#ifdef VM_USE_LOOPTOOLS
	ffexi(); m_bLTInited=false;
#endif //VM_USE_LOOPTOOLS

	//Finalize Pythia
#ifdef VM_USE_PYTHIA8
	if(CAlHepLib::pythia != NULL){ delete CAlHepLib::pythia; CAlHepLib::pythia=NULL; }
#endif //#ifdef VM_USE_PYTHIA8

	//Debug
	if(m_DebugIn!=NULL){ delete m_DebugIn; m_DebugIn=NULL; }
	if(m_DebugOut!=NULL){ delete m_DebugOut; m_DebugOut=NULL; }
}

bool CAlHepLib::TestLibrary(bool bSilent) { //calls all tests: TestReal, LTTest, TestLi2
	if(!bSilent){
		std::cout<<"[CAlHepLib::TestLibrary]"<<std::endl;
		std::cout<<"'Real' type size: "<<sizeof(tReal)<<std::endl;
	}

	if(!TestReal(!bSilent))return false;
#ifdef VM_USE_LOOPTOOLS
	if(!LTTest(bSilent))return false;
#endif //VM_USE_LOOPTOOLS
	if(!TestLi2(bSilent))return false;
	if(!TestRnd(bSilent))return false;
	if(!bSilent){ std::cout<<"[CAlHepLib::TestLibrary] Done"<<std::endl; }
	return true;
}
tReal CAlHepLib::F1(tReal Q2, tReal M2) {
	tReal _Gm=Gm(Q2); return Q2*_Gm*_Gm;
}
tReal CAlHepLib::F2(tReal Q2, tReal M2) {
	tReal _Ge=Ge(Q2); tReal _Gm=Gm(Q2); tReal taup=Q2/(4.*M2); return 4.*M2*(_Ge*_Ge+taup*_Gm*_Gm)/(1.+taup);
}
tReal CAlHepLib::F3(tReal Q2, tReal M2) {
	tReal _Ge=Ge(Q2); tReal _Gm=Gm(Q2); return -2.*M2*_Ge*_Gm;
}
tReal CAlHepLib::F4(tReal Q2, tReal M2) {
	tReal _Ge=Ge(Q2); tReal _Gm=Gm(Q2); tReal taup=Q2/(4.*M2); return -M2*_Gm*(_Ge-_Gm)/(1.+taup);
}
tReal CAlHepLib::Ge(tReal Q2) {
	tReal Q02=0.71; tReal tmp=(1.+Q2/Q02); return 1./(tmp*tmp);
}
tReal CAlHepLib::Gm(tReal Q2) {
	tReal Q02=0.71;
	tReal tmp=(1.+Q2/Q02); return CVMConst::MMom_p/(tmp*tmp); //times 'magnetic moment of proton'
}
const char* CAlHepLib::GetParticleTitle(int idAbs) { // For debug purposes only!
	if(idAbs<0)idAbs=-idAbs;
	const char* res=NULL;
	switch(idAbs) {
	case 1: res="d"; break; case 2: res="u"; break; case 3: res="s"; break; case 4: res="c"; break; case 5: res="b"; break; case 6: res="t"; break; case 7: res="b'"; break; case 8: res="t'"; break;
	case 11: res="e"; break; case 12: res="nu_e"; break; case 13: res="mu"; break; case 14: res="nu_m"; break; case 15: res="tau"; break; case 16: res="nu_t"; break; case 17: res="tau'"; break; case 18: res="nu_t'"; break;
	case 21: res="g"; break; case 22: res="A"; break; case 23: res="Z"; break; case 24: res="W"; break; case 25: res="H"; break;
	case 32: res="Z'"; break; case 33: res="Z''"; break; case 34: res="W'"; break; case 35: res="H+"; break;
	case 111: res="pi0"; break; case 211: res="pi+"; break;
	case 221: res="eta"; break;
	case 130: res="KL0"; break; case 310: res="KS0"; break; case 311: res="K0"; break; case 321: res="K+"; break;
	case 411: res="D+"; break; case 421: res="D0"; break;
	case 511: res="B+"; break; case 521: res="B0"; break;
	case 2212: res="p"; break; case 2112: res="n"; break;
	case 2224: res="Delta++"; break; case 2214: res="Delta+"; break; case 2114: res="Delta0"; break; case 1114: res="Delta-"; break;
	case 3122: res="Lambda"; break;
	case 3222: res="Sigma+"; break; case 3212: res="Sigma0"; break; case 3112: res="Sigma-"; break;
	case 3322: res="Theta0"; break; case 3312: res="Theta-"; break;
	case 3334: res="Omega-"; break;
	}
	return res;
}
///*
tReal CAlHepLib::Pow(tReal val, int p) {
	if(p<0)return 1./Pow(val, -p);
	if(p>10000)return pow(val,p); // huge powers --> to series calculation
	tReal res=1.;
	if(p%2==1){ res*=val; p-=1; }
	if(p>0){ tReal subRes=Pow(val,p/2); res*=subRes*subRes; }
	return res;
}
//*/

// Find single solufion of func(x)==0, very slow implementation, for debug only
bool CAlHepLib::NSolve(tReal(vmdecl *func)(tReal), tReal xMin, tReal xMax, tReal& res) {
	if (xMin > xMax) { tReal tmp=xMin; xMin=xMax; xMax=tmp; }
	if ((func==NULL) || (xMin == xMax)) return false;

	tReal xLeft=xMin; tReal xRight=xMax;
	tReal fLeft=(*func)(xLeft); if ((fLeft!=fLeft) || (CAlHepLib::IsInf(DBLtR(fLeft)))) return false;
	if(fLeft==0.){ res=xLeft; return true; }
	tReal fRight=(*func)(xRight); if ((fRight!=fRight) || (CAlHepLib::IsInf(DBLtR(fRight)))) return false;
	if(fRight==0.){ res=xRight; return true; }
	if(fRight*fLeft >= 0.){
		//std::cout<<"f[min]*f[max]>=0"<<std::endl;
		//std::cout<<"f[min]= "<<fLeft<<std::endl;
		//std::cout<<"f[max]= "<<fRight<<std::endl;
		return false;
	}
	bool bResFound=false;
	for(;;) {
		tReal xMid=(xLeft+xRight)/2.;
		if ((xRight-xLeft) < 10.*epsReal){ res=xMid; bResFound=true; break; }
		tReal fMid=(*func)(xMid); if ((xMid!=xMid) || (CAlHepLib::IsInf(DBLtR(xMid)))) break;
		if(fMid==0.){ res=xMid; bResFound=true; break; }
		if(fLeft*fMid < 0.){ fRight=fMid; xRight=xMid; }else{ fLeft=fMid; xLeft=xMid; }
	}
	return bResFound;
}

tReal CAlHepLib::m_SoftPhotonE	= 5.e-6; //default: ~0.01*m_e
tReal CAlHepLib::SoftBremsInt1(const CVM4Vec& k1, tReal m1) {
	//Soft bremsstrahlung integral (see Denner'93 vol.7.2)
	//I = Int{ d3k/2w 2k1k1/(k1k*k1k) }
	//Note: k1^2 may differ from m^2 (especially in 0-angle scattering)
	//HowTo fix it: compile with high precision numbers or use 'E=sqrt(P^2+m^2)' instead of k.m_E value

	tReal lambda2=GetIRLambdaSqr();
	if((CAlHepLib::m_SoftPhotonE<=0.) || (lambda2<=0.))return 0.; //error
	tReal kE=k1.m_E; tReal kP=k1.GetPAbs(); if((kP==0.)||(kP==kE))return 0.; //error

	tReal kRatio=m1/(kE+kP); kRatio*=kRatio; //kRatio=(kE-kP)/(kE+kP)
	tReal res = 2.*CVMConst::Pi()*(
		log(CAlHepLib::m_SoftPhotonE*CAlHepLib::m_SoftPhotonE*4./lambda2)
		+(kE/kP)*log(kRatio)
		);

	if(IsNaN(DBLtR(res)) || IsInf(DBLtR(res))) {
		std::cout<<"SoftBremsInt1 error. res: "<<res<<std::endl;
		std::cout<<"k1: "<<k1<<std::endl;
		std::cout<<"kE: "<<kE<<std::endl;
		std::cout<<"kP: "<<kP<<std::endl;
		std::cout<<"CAlHepLib::m_SoftPhotonE: "<<CAlHepLib::m_SoftPhotonE<<std::endl;
		std::cout<<"lambda^2: "<<lambda2<<std::endl;
		std::cout<<"(kE-kP)/(kE+kP): "<<(kE-kP)/(kE+kP)<<std::endl;
		exit(0);
	}
	return res;
}
tReal CAlHepLib::SoftBremsInt2(const CVM4Vec& k1, const CVM4Vec& k2, tReal m1, tReal m2) {
	//Soft bremsstrahlung integral (see Denner'93 vol.7.2)
	//I = Int{ d3k/2w 2k1k2/(k1k*k2k) }
	//Note: k1^2 may differ from m^2 (especially in 0-angle scattering)
	//HowTo fix it: compile with high precision numbers or use 'E=sqrt(P^2+m^2)' instead of k.m_E value

	//TODO: make 'm' params optional (def: m=-1)

	if(k1==k2)return SoftBremsInt1(k1, m1);
	tReal lambda2=GetIRLambdaSqr();
	if((CAlHepLib::m_SoftPhotonE<=0.) || (lambda2<=0.))return 0.; //error

	tReal k1Sqr=m1*m1; tReal k2Sqr=m2*m2;								//tReal k1Sqr=k1.GetSqr(); tReal k2Sqr=k2.GetSqr();
	tReal k1P=k1.GetPAbs(); tReal k1E=sqrt(k1Sqr+k1P*k1P);				//tReal k1E=k1.m_E;
	tReal k2P=k2.GetPAbs(); tReal k2E=sqrt(k2Sqr+k2P*k2P);				//tReal k2E=k2.m_E;
	tReal k1k2=k1E*k2E-k1.m_Pz*k2.m_Pz-k1.m_Px*k2.m_Px-k1.m_Py*k2.m_Py;	//tReal k1k2=k1.GetCpl(k2);
	tReal dk1=k1Sqr/(k1E+k1P); //=(k2E-k2P)
	tReal dk2=k2Sqr/(k2E+k2P); //=(k2E-k2P)

	if((k1Sqr<=0.) || (k2Sqr<=0.)){
		return 0.; //error
	}
	if((k1E<=k1P) || (k2E<=k2P)){
		return 0.; //error
	}

	/*
	std::cout<<"k1: "<<k1<<std::endl;
	std::cout<<"k1E: "<<k1E<<std::endl;
	std::cout<<"k1P: "<<k1P<<std::endl;
	std::cout<<"(k1E-k1P)/(k1E+k1P): "<<(k1E-k1P)/(k1E+k1P)<<std::endl;
	std::cout<<"k2: "<<k2<<std::endl;
	std::cout<<"k2E: "<<k2E<<std::endl;
	std::cout<<"k2P: "<<k2P<<std::endl;
	std::cout<<"(k2E-k2P)/(k2E+k2P): "<<(k2E-k2P)/(k2E+k2P)<<std::endl;
	std::cout<<"k1k2: "<<k1k2<<std::endl;
	std::cout<<"k1Sqr: "<<k1Sqr<<std::endl;
	std::cout<<"k2Sqr: "<<k2Sqr<<std::endl;
	std::cout<<"m2: "<<m1*m1<<std::endl;
	*/

	tReal ad=k1k2*k1k2-k1Sqr*k2Sqr; tReal adSqrt=((ad>0.)?sqrt(ad):0.); tReal a1=(k1k2+adSqrt)/k1Sqr; tReal a2=(k1k2-adSqrt)/k1Sqr;
	//std::cout<<"a1: "<<a1<<", fit check: "<<((a1*k1E-k2E)/k2E>0.)<<std::endl;
	//std::cout<<"a2: "<<a2<<", fit check: "<<((a2*k1E-k2E)/k2E>0.)<<std::endl;
	tReal a=0.; if((a1*k1E-k2E)/k2E>0.){a=a1;}else if((a2*k1E-k2E)/k2E>0.){a=a2;}else{return 0.;}//else error
	if(a==0.)return 0.;//error
	//std::cout<<"a: "<<a<<std::endl;

	tReal v=(a*a*k1Sqr-k2Sqr)/2./(a*k1E-k2E);
	//std::cout<<"v: "<<v<<std::endl;
	if(v==0.)return 0.;//error

	tReal res=0.;
	//CVM4Vec u2=k1; u2*=a; tReal u2E=u2.m_E; tReal u2P=u2.GetPAbs(); //u1=k2;
	//tReal u2E=k1E*a; tReal u2P=k1P*a; //u1=k2;
	//tReal rk1=dk1/(k1E+k1P); //=(k1E-k1P)/(k1E+k1P)=(u2E-u2P)/(u2E+u2P)
	//tReal rk2=dk2/(k2E+k2P); //=(k2E-k2P)/(k2E+k2P)
	//(u2E-u2P)==a*dk1;

	//tReal tLn2Sqr=log(fabs((u2E-u2P)/(u2E+u2P))); res+=( tLn2Sqr*tLn2Sqr/4. + Li2(1.-(u2E+u2P)/v) + Li2(1.-(u2E-u2P)/v) );
	tReal tLn2Sqr=log(fabs(dk1/(k1E+k1P))); res+=( tLn2Sqr*tLn2Sqr/4. + Li2(1.-a*(k1E+k1P)/v) + Li2(1.-a*dk1/v) );
	tReal tLn1Sqr=log(fabs(dk2/(k2E+k2P))); res-=( tLn1Sqr*tLn1Sqr/4. + Li2(1.-(k2E+k2P)/v) + Li2(1.-dk2/v) );

	tReal logLa = log(CAlHepLib::m_SoftPhotonE*CAlHepLib::m_SoftPhotonE*4./lambda2);
	res+=log(a*a*k1Sqr/k2Sqr)*logLa/2.;

	res*=a*k1k2/(a*a*k1Sqr-k2Sqr);
	res*=4.*CVMConst::Pi();

	//std::cout<<"SoftBremsInt1 res: "<<res<<std::endl;
	//exit(0);

	if(IsNaN(DBLtR(res)) || IsInf(DBLtR(res))) {
		std::cout<<"SoftBremsInt1 error. res: "<<res<<std::endl;
		std::cout<<"k1: "<<k1<<std::endl;
		std::cout<<"k1E: "<<k1E<<std::endl;
		std::cout<<"k1P: "<<k1P<<std::endl;
		std::cout<<"(k1E-k1P)/(k1E+k1P): "<<(k1E-k1P)/(k1E+k1P)<<std::endl;
		std::cout<<"k2: "<<k2<<std::endl;
		std::cout<<"k2E: "<<k2E<<std::endl;
		std::cout<<"k2P: "<<k2P<<std::endl;
		std::cout<<"(k2E-k2P)/(k2E+k2P): "<<(k2E-k2P)/(k2E+k2P)<<std::endl;
		std::cout<<"k1k2: "<<k1k2<<std::endl;
		exit(0);
	}
	return res;
}

//Algebraic functions
int CAlHepLib::m_iNormGenMethod	= 3; //default: 3;
tReal CAlHepLib::GenNormDistr() { //N(0,1), Note: N(mu,sigma)=mu+sigma*N(0,1)
	CVMRnd* rnd=CVMRnd::GetRndRef(); if(rnd==NULL)return 0.;
	tReal res;
	//
	//Several algorithms available
	//The cumulative distribution function (CDF) is expressed via 'Erf()' and is not fast ==> use approximation here

	switch(m_iNormGenMethod){
	case 0:
		//Precise
		//res=(1.+Erf(rnd->Rnd()/sqrt((tReal)2.)))/2.;
		std::cout<<"Error in CAlHepLib::GenNormDistr()"<<std::endl; exit(0);
		break;
	case 1:
		//1.[Approximation] IrwinHall[n==12] distribution
		res=0.; for(int i=0;i<12;i++){res+=rnd->Rnd();} res-=6.;
		break;
	case 2: {
		//2.[Precise] BoxMuller method
		tReal u=rnd->Rnd(); tReal v=rnd->Rnd(); for(;u==0.;)u=rnd->Rnd(); //(0.,1.]
		tReal resA=sqrt(fabs(-2.*log(u))); res=resA*cos(2.*CVMConst::Pi()*v); //res2=resA*sin(2.*CVMConst::Pi()*v);
			} break;
	case 3:
	default: {
		//3.[Precise] Marsaglia polar method
		bool bLoop=true;
		for(;bLoop;){
			tReal u=rnd->Rnd()*2.-1.; tReal v=rnd->Rnd()*2.-1.; tReal s=u*u+v*v; if((s>=1.)||(s<=0.))continue;
			tReal resA=sqrt(fabs(-2.*log(s)/s)); res=resA*u; bLoop=false; //res2=resA*v;
		}
			 } //break;
	//default: std::cout<<"Error in CAlHepLib::GenNormDistr()"<<std::endl; exit(0);
	}
	return res;
}
tReal CAlHepLib::Li2(tReal x) { //Spence function
	//debug: if((x<-1.5)||(x>2.5))return 0.;
	tReal pi=CVMConst::Pi(); tReal pi2=pi*pi;
	tReal res;
	if(x>=0.) {
		if(x <= (tReal)1./(tReal)2.){ res=Li2_01(x); }
		else {
			if( x<=(tReal)1. ){ res=pi2/6.-log(x)*log(1.-x+epsReal*10.)-Li2(1.-x); }//'+epsReal*10' is neglected due to 'log(x)'
			else {
				if( x<=2. ){ res=pi2/6.-log(x)*log((x-1.)*(x-1.)/x)/2.+Li2(1.-1./x); }
				else{ res=pi2/3.-log(x)*log(x)/2.-Li2(1./x); }
			}
		}
	} else {
		if( x>=(tReal)(-1.) ){ res=-log(1.-x)*log(1.-x)/2.-Li2(x/(x-1.)); }
		else{ res=-pi2/6.-log(1.-x)*log(x*x/(1.-x))/2.+Li2(1./(1.-x)); }
	}
	return res;
}
tReal CAlHepLib::Li2_01(tReal x) { //Note: infinite loop here if fabs(x)>=1., very slow at x->1
	tReal res=0.; tReal eps=epsReal;
	tReal xPowN=1.; long n=0;
	for(;;){ n++; xPowN*=x; tReal item=xPowN/(n*n); res+=item; if(fabs(item)<=eps) break; }
	return res;
}
bool CAlHepLib::TestLi2(bool bSilent) { //saves 'TestLi2.eps'
	bool bOK=true; tReal val1,val2,d; tReal errMax=0.00001; if(!bSilent){ std::cout<<"--- TestLi2 started ---"<<std::endl; }
	val1=Li2(1.); val2=CVMConst::Pi()*CVMConst::Pi()/6.; d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>errMax))bOK=false;
	if(!bSilent){ std::cout<<"Li2[1]: "<<val1<<", rel. error: "<<d<<"."<<std::endl; }
	val1=Li2(1.-10*epsReal); val2=CVMConst::Pi()*CVMConst::Pi()/6.; d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>errMax))bOK=false;
	if(!bSilent){ std::cout<<"Li2[1-10*eps]: "<<val1<<", Li[1]-Li[1-10*eps]: "<<d<<"."<<std::endl; }
	if(!bSilent){
#ifdef Decl_CVMHistLite
		const char* hFileName="TestLi2.eps"; int i;
		CVMHistLite hst("Li2 test",100,-2.,3.); for(i=0; i<100; i++)hst.Fill(hst.GetBinXCenter(i),DBLtR(Li2(hst.GetBinXCenter(i))));
		CVMHistLite hstX("Li2 series",100,-1.,1); for(i=0; i<100; i++)hstX.Fill(hstX.GetBinXCenter(i),DBLtR(Li2_01(hstX.GetBinXCenter(i))));
		CVMHistLiteView hview(&hst);
		hview.AddPlotToView(&hstX); hview.SetHistStyle(1,CVMRGB(1,0,0),CVMOutput::LINE_DASH);
#ifdef VM_USE_LOOPTOOLS
		CVMHistLite hstLT("LoopTools",100,-2.,3.);
#ifdef VM_LOOPTOOLS_QUAD
		for(i=0; i<100; i++)hstLT.Fill(hstLT.GetBinXCenter(i),to_double(Re(::Li2(hstLT.GetBinXCenter(i)))));
#else
		for(i=0; i<100; i++)hstLT.Fill(hstLT.GetBinXCenter(i),Re(::Li2(hstLT.GetBinXCenter(i))));
#endif //VM_LOOPTOOLS_QUAD
		hview.AddPlotToView(&hstLT); hview.SetHistStyle(2,CVMRGB(0,1,0),CVMOutput::LINE_SOLID,CVMOutput::LW_THIN);
#endif //VM_USE_LOOPTOOLS
		hview.SaveAsEPS(hFileName);
		std::cout<<"CAlHepLib::TestLi2(): histogram saved as '"<<hFileName<<"'."<<std::endl;
#else
		std::cout<<"CAlHepLib::TestLi2() test skipped: no histograms available."<<std::endl;
#endif //Decl_CVMHistLite
	}
	if(!bSilent){ std::cout<<"--- TestLi2 "<<(bOK?"completed OK":"FAILED")<<" ---"<<std::endl; }
	return bOK;
}
bool CAlHepLib::TestRnd(bool bLog) { //Test default RND
	CVMRnd* rnd = CVMRnd::GetRndRef(); bool bOK=true; int i; long N=100000; tReal ave=0.;
#ifdef Decl_CVMHistLite
	CVMHistLite hst("rnd",100,0,1);
#endif //Decl_CVMHistLite
	for(i=0;i<N;i++){
		tReal r=rnd->Rnd();
#ifdef Decl_CVMHistLite
		hst.Fill(DBLtR(r),1./N);
#endif //Decl_CVMHistLite
		ave+=r;
	}
#ifdef Decl_CVMHistLite
	CVMHistLiteView hview(&hst); hview.SaveAsEPS("TestRnd.eps");
#endif //Decl_CVMHistLite
	ave/=N; if((ave<0.4)||(ave>0.6))bOK=false;
	if(bOK){ std::cout<<"Random numbers test passed successfully"<<std::endl; }
	else{ std::cout<<"Random numbers test FAILED (Average: "<<ave<<")"<<std::endl; }
	return bOK;
}
//Other debug functions
char* CAlHepLib::LogConstantValues() { //Allocates memory using CVMUtil::strAlloc()
	char* res=CVMUtil::strAlloc(10000); if(res==NULL)return NULL;
	int l=0;
	l+=sprintf(res+l,"CVMConst::Euler = %f\n",DBLtR(CVMConst::Euler));
	l+=sprintf(res+l,"CVMConst::Alpha = %f\n",DBLtR(CVMConst::Alpha));
	l+=sprintf(res+l,"CVMConst::M_e = %f\n",DBLtR(CVMConst::M_e));
	l+=sprintf(res+l,"CVMConst::M_mu = %f\n",DBLtR(CVMConst::M_mu));
	l+=sprintf(res+l,"CVMConst::M_p = %f\n",DBLtR(CVMConst::M_p));
	l+=sprintf(res+l,"CVMConst::ReClassic = %f\n",DBLtR(CVMConst::ReClassic));
	l+=sprintf(res+l,"CVMConst::GeV2pb = %f\n",DBLtR(CVMConst::GeV2pb));
	l+=sprintf(res+l,"CVMConst::El = %f\n",DBLtR(CVMConst::El));
	l+=sprintf(res+l,"CVMConst::MMom_p = %f\n",DBLtR(CVMConst::MMom_p));
	return res;
}

bool CAlHepLib::IsNaN(double val) {
	return (val!=val);
}
double CAlHepLib::GetNaN() {
	double res=0; ((unsigned long*)&res)[1]=0x7ff80000; return res;
	//Ver.2: return sqrt(-1.);
}
bool CAlHepLib::IsInf(double val) {
	return ( (((unsigned long*)&val)[1] & 0x7ff80000) == 0x7ff00000 );
	//Ver.2: return ((val == 1./0.) || (-val == 1./0.));
}
double CAlHepLib::GetInf() {
	double res=0; ((unsigned long*)&res)[1]=0x7ff00000; return res;
	//Ver.2: return (1./0.);
}
/*
*	Fill 'real(kind=16)' variable (as byte array)
*/
void CAlHepLib::ConvertToReal16(double val, unsigned char* res16) {
	if(res16==NULL)return;
	unsigned char* pVal=(unsigned char*)(&val); int i; unsigned char c;
	//Clear result
	for(i=0;i<16;i++){ res16[i]=0; }
	//Copy mantissa
	for(i=6;i>=0;i--){ c=pVal[i]; c=c<<4; res16[7+i]+=c; if(i<=5){ c=pVal[i]; c=c>>4; res16[8+i]+=c; } }
	//Copy exponent
	unsigned int uExp=(pVal[7] & 0x7f); uExp=uExp<<4; c=pVal[6]; c=c>>4; uExp+=c; //uExp is now 'exp+0x3ff' ('7ff' is Inf/NaN)
	if(uExp>0x7ff){ std::cout<<"Error in CAlHepLib::ConvertToReal16()"<<std::endl; uExp=0x7ff; }
	if(uExp<0x7ff){
		if(uExp!=0)uExp+=0x3c00; //if(val==0.0): uExp==0 and must remain ==0
	}else{ uExp=0x7fff; }
	//uExp is now 'exp+0x3fff' (or 0 for val==0.0)
	res16[14]=((unsigned char*)(&uExp))[0];
	res16[15]=((unsigned char*)(&uExp))[1];
	//Copy sign
	if(pVal[7] & 0x80){ res16[15]|=0x80; }else{ res16[15]&=0x7f; } //'else' is useless here
}
double CAlHepLib::ConvertFromReal16(const unsigned char* res16){
	if(res16==NULL)return CAlHepLib::GetNaN();
	double res=0.; unsigned char* pRes=(unsigned char*)(&res); int i; unsigned char c;
	//Copy mantissa
	for(i=6;i>=0;i--){ c=res16[7+i]; c=c>>4; pRes[i]=c; if(i<=5){ c=res16[8+i]; c=c<<4; pRes[i]+=c; } }
	//Copy exponent
	unsigned int uExp=(res16[15] & 0x7f); uExp=uExp<<8; uExp+=res16[14]; //uExp is now 'exp+0x3fff' ('7fff' is Inf/NaN)
	if(uExp==0){ //TODO: check mantissa for 0 here?
		//do nothing, it's probably 0.0
	}else if(uExp<0x7fff){
		if(uExp<0x3c00){
			std::cout<<"Error in CAlHepLib::ConvertFromReal16(): power is too SMALL for double ("<<uExp<<")"<<std::endl;
			//Old: uExp=0x3c00; //set exponent to minimal possible value
			//New: return '0'
			return 0.;
		}
		uExp-=0x3c00;
		if(uExp>=0x7ff){
			std::cout<<"Error in CAlHepLib::ConvertFromReal16(): power is too BIG for double ("<<uExp<<")"<<std::endl;
			//Old: uExp=0x7fe; //set exponent to maximal possible value
			//New: return '+/-Inf'
			res=CAlHepLib::GetInf(); if(res16[15]&0x80)res=-res; return res;
		}
	}else{ uExp=0x7ff; } //Copy 'Inf/NaN' id
	//uExp is now 'exp+0x3ff' ('7ff' is Inf/NaN)
	c=((unsigned char*)(&uExp))[0]; c=c<<4; pRes[6]+=c; uExp=uExp>>4; pRes[7]=((unsigned char*)(&uExp))[0];
	//Copy sign
	if(res16[15] & 0x80){ pRes[7]|=0x80; }else{ pRes[7]&=0x7f; } //'else' is useless here
	return res;
}
long CAlHepLib::dbg_FirstStackPos=0; // Used by 'ViewStackUsed'
void CAlHepLib::ViewStackUsed(const char* message, void* pLastLocalVar) { //Writes 'Stack used: xx bytes' (since first call) to 'std::cout'
	int iTmp=0;
	if(CAlHepLib::dbg_FirstStackPos==0){ //First init
		CAlHepLib::dbg_FirstStackPos=((pLastLocalVar!=NULL)?(long)pLastLocalVar:(long)&iTmp);
		//[useless output: stack pos. is random]std::cout<<"Stack position saved: "<<CAlHepLib::dbg_FirstStackPos<<std::endl;
	}else{
		std::cout<<"Stack "<<((pLastLocalVar!=NULL)?"position":"used"); if(message!=NULL)std::cout<<" ["<<message<<"]"; std::cout<<": "<<abs(((pLastLocalVar!=NULL)?(long)pLastLocalVar:(long)&iTmp)-CAlHepLib::dbg_FirstStackPos)<<std::endl;
	}
}
tReal CAlHepLib::AsymmProd(const CVM4Vec& vec1, const CVM4Vec& vec2, const CVM4Vec& vec3, const CVM4Vec& vec4) {
	//TODO: implement it via 'det()' function
	tReal tmp,res;
	tmp=vec2.m_Py*(vec3.m_Pz*vec4.m_E-vec3.m_E*vec4.m_Pz);
	tmp-=vec2.m_Pz*(vec3.m_Py*vec4.m_E-vec3.m_E*vec4.m_Py);
	tmp+=vec2.m_E*(vec3.m_Py*vec4.m_Pz-vec3.m_Pz*vec4.m_Py);
	res=vec1.m_Px*tmp;
	tmp=vec2.m_Px*(vec3.m_Pz*vec4.m_E-vec3.m_E*vec4.m_Pz);
	tmp-=vec2.m_Pz*(vec3.m_Px*vec4.m_E-vec3.m_E*vec4.m_Px);
	tmp+=vec2.m_E*(vec3.m_Px*vec4.m_Pz-vec3.m_Pz*vec4.m_Px);
	res-=vec1.m_Py*tmp;
	tmp=vec2.m_Px*(vec3.m_Py*vec4.m_E-vec3.m_E*vec4.m_Py);
	tmp-=vec2.m_Py*(vec3.m_Px*vec4.m_E-vec3.m_E*vec4.m_Px);
	tmp+=vec2.m_E*(vec3.m_Px*vec4.m_Py-vec3.m_Py*vec4.m_Px);
	res+=vec1.m_Pz*tmp;
	tmp=vec2.m_Px*(vec3.m_Py*vec4.m_Pz-vec3.m_Pz*vec4.m_Py);
	tmp-=vec2.m_Py*(vec3.m_Px*vec4.m_Pz-vec3.m_Pz*vec4.m_Px);
	tmp+=vec2.m_Pz*(vec3.m_Px*vec4.m_Py-vec3.m_Py*vec4.m_Px);
	res-=vec1.m_E*tmp;
	res=-res;
	return res;
}

//LoopTools settings and tests (empty calls if 'VM_USE_LOOPTOOLS' is not defined)
tReal CAlHepLib::m_IRLambda2	= 1.e-8;	// Used if LT is unavailable only.
tReal CAlHepLib::m_UVChi		= 0.;		// Used if LT is unavailable only.
tReal CAlHepLib::m_DimMuSqr		= 1.;		// Used if LT is unavailable only.
bool CAlHepLib::m_bLTInited		= false;
bool CAlHepLib::LTInit() {
	std::cout<<"------------------------ CAlHepLib::LTInit ------------------------"<<std::endl;
#ifdef VM_USE_LOOPTOOLS
	if(m_bLTInited)return true;
	m_bLTInited=true;
	ffini();

	//View default values (to be changed later)
	std::cout<<"[Default value from LT] m_IRLambda2= "<<(double)getlambda()<<std::endl;
	std::cout<<"[Default value from LT] m_UVChi=     "<<(double)getdelta()<<std::endl;
	std::cout<<"[Default value from LT] m_DimMuSqr=  "<<(double)getmudim()<<std::endl;

	//Constants re-init:
	tReal me2=CVMConst::M_e*CVMConst::M_e;
	SetIRLambdaSqr(me2); //SetIRLambdaSqr(1.);
	CAlHepLib::SetUVChi(0.);
	//SetUVChi(-15.1583); //debug
	CAlHepLib::SetDimMuSqr(me2); //CAlHepLib::SetDimMuSqr(1.);

	//Save constants copies (useless)
	m_IRLambda2=getlambda(); m_UVChi=getdelta(); m_DimMuSqr=getmudim(); //useless init => 'm_[LT]' variables are never used
	std::cout<<"[LT parameter changed] m_IRLambda2= "<<m_IRLambda2<<" , expected: "<<me2<<std::endl;
	std::cout<<"[LT parameter changed] m_UVChi=     "<<m_UVChi<<" , expected: "<<0.<<std::endl;
	std::cout<<"[LT parameter changed] m_DimMuSqr=  "<<m_DimMuSqr<<" , expected: "<<me2<<std::endl;
	return true;
#else
	m_IRLambda2 = 1.e-16; //default init
	return false;
#endif //VM_USE_LOOPTOOLS
}
void CAlHepLib::LTClearCache() {
#ifdef VM_USE_LOOPTOOLS
	//setcachelast(Ccache, 0);
	//setcachelast(Dcache, 0);
	clearcache();
#endif //VM_USE_LOOPTOOLS
}

bool CAlHepLib::LTTest(bool bSilent) {
	bool bOK=true; if(!bSilent){std::cout<<"LoopTools test started"<<std::endl;}
#ifdef VM_USE_LOOPTOOLS
	if(!m_bLTInited)LTInit();

	tReal l2Old=CAlHepLib::GetIRLambdaSqr(); tReal xiOld=CAlHepLib::GetUVChi(); tReal mu2Old=CAlHepLib::GetDimMuSqr();

	tReal Pi=CVMConst::Pi(); tReal a4Pi=CVMConst::Alpha/(((tReal)4.)*Pi);

	if(!bSilent){std::cout<<"alpha/4Pi="<<a4Pi<<std::endl; }
	if(!bSilent){std::cout<<"[Global constant] lambda^2="<<l2Old<<std::endl; }
	if(!bSilent){std::cout<<"[Global constant] chi="<<xiOld<<std::endl; }
	if(!bSilent){std::cout<<"[Global constant] mudim^2="<<mu2Old<<std::endl; }
	if(!bSilent){std::cout<<"[Global constant] chi-log(me^2/mu^2)="<<xiOld - log(CVMConst::M_e*CVMConst::M_e/mu2Old)<<std::endl; }

	//SetUVChi(100.); //test
	SetDimMuSqr(1.); //test
	SetIRLambdaSqr(1.); //test

	const tReal l2=GetIRLambdaSqr(); const tReal xi=GetUVChi(); const tReal mu2=GetDimMuSqr();
	const tReal me=CVMConst::M_e; const tReal me2=me*me; const tReal xiM=xi-log(me2/mu2);
	const tReal logML=log(me2/l2);

	tReal val1,val2,d;
	const tReal p2=me2*100; const tReal p2Small=me2/10000.;
	const tReal f1d2=((tReal)1.)/2.;

	if(!bSilent){ std::cout<<"lambda^2: "<<l2<<", mudim^2: "<<mu2<<", chi: "<<xi<<std::endl; }
	if(!bSilent){std::cout<<"[In-test constant] lambda^2="<<l2<<std::endl; }
	if(!bSilent){std::cout<<"[In-test constant] chi="<<xi<<std::endl; }
	if(!bSilent){std::cout<<"[In-test constant] mudim^2="<<mu2<<std::endl; }
	if(!bSilent){std::cout<<"[In-test constant] chi-log(me^2/mu^2)="<<xiM<<std::endl; }
	if(!bSilent){std::cout<<"[In-test constant] log(me^2/l^2)="<<logML<<std::endl; }

	//A(m) = m^2(x+1-ln(m^2/mu^2))
	val1=(tReal)Re(A0(DBLtR(me2))); val2=me2*(xi+1-log(me2/mu2)); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"A0 value test #1:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}
	//A(0) = 0
	val1=(tReal)Re(A0(0.)); val2=0.; d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"A0 value test #2:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B(0,m,m) = x-ln(m^2/mu^2)
	val1=(tReal)Re(B0(0.,DBLtR(me2),DBLtR(me2))); val2=xi-log(me2/mu2); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B0 value test #1:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B(m^2,0,m) = x-ln(m^2/mu^2)+2
	//Note: LT says 'IR-divergent B0 is involved', but BO is IR-safe, only dB0 is divergent!
	val1=(tReal)Re(B0(DBLtR(me2),0.,DBLtR(me2))); val2=xi+2.-log(me2/mu2); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B0 value test #2:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B(m^2,0,m) =[if l<<1 only!]= B(m^2,l,m)
	//val1=(tReal)Re(B0(DBLtR(me2),0.,DBLtR(me2))); val2=(tReal)Re(B0(DBLtR(me2),DBLtR(l2),DBLtR(me2))); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	//if(!bSilent){std::cout<<"B0 value test #3:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B(0,m,0) = ( x-ln(m^2/mu^2)+1 )
	val1=(tReal)Re(B0(0.,DBLtR(me2),0.)); val2=xiM+(tReal)1.; d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B0(0,m,0):    "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//Must be: B1(p2,m1,m2) == 1/(2*p2)*{ A0(m1)-A0(m2)-(p2+m1^2-m2^2)*B0(p2,m1,m2) }
	//Must be: B1(p2,m1,m2) == 1/(2*p2)*{ (m1^2-m2^2)*B0(0,m1,m2)-(p2+m1^2-m2^2)*B0(p2,m1,m2) }

	//B1(p2,m,m) == 1/(2*p2)*( -p2*B0(p2,m,m) )
	val1=(tReal)Re(B1(DBLtR(p2),DBLtR(me2),DBLtR(me2))); val2=(((tReal)1.)/(2.*p2))*( -p2*(tReal)Re(B0(DBLtR(p2),DBLtR(me2),DBLtR(me2))) ); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B1 test #1:   "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B1(m2,m,m) == 1/(2*m2)*( A0(m)-A0(m)-l2*B0(m2,m,m) )
	//val1=(tReal)Re(B1(DBLtR(me2),DBLtR(me2),DBLtR(me2))); val2=(((tReal)1.)/(2.*me2))*( (tReal)Re(A0(DBLtR(me2)))-(tReal)Re(A0(DBLtR(me2)))-me2*(tReal)Re(B0(DBLtR(me2),DBLtR(me2),DBLtR(me2))) ); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	//if(!bSilent){std::cout<<"B1 test #1x:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B1(m2,0,m) == 1/(2*m2)*( -A0(m) )
	val1=(tReal)Re(B1(DBLtR(me2),0.,DBLtR(me2))); val2=(((tReal)1.)/(2.*me2))*( -(tReal)Re(A0(DBLtR(me2))) ); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B1 test #1b:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B1(m2,m,0) == 1/(2*m2)*( A0(m) - 2*m2*B0(m2,m,0) )
	val1=(tReal)Re(B1(DBLtR(me2),DBLtR(me2),0.)); val2=f1d2/me2*( (tReal)Re(A0(DBLtR(me2))) - 2*me2*Re(B0(DBLtR(me2),DBLtR(me2),0.)) ); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B1 test #1c:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B1(m2,m,l) == 1/(2*m2)*( A0(m)-A0(l)-(2*me2-l2)*B0(m2,m,l) )
	val1=(tReal)Re(B1(DBLtR(me2),DBLtR(me2),DBLtR(l2))); val2=(((tReal)1.)/(2.*me2))*( (tReal)Re(A0(DBLtR(me2)))-(tReal)Re(A0(DBLtR(l2)))-(2*me2-l2)*(tReal)Re(B0(DBLtR(me2),DBLtR(me2),DBLtR(l2))) ); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B1 test #2:   "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B1(m2,m,l) == 1/2*(1.-l2/m2)*B0(0,m,l)-(1-l2/(2*m2))*B0(m2,m,l)
	val1=(tReal)Re(B1(DBLtR(me2),DBLtR(me2),DBLtR(l2))); val2=f1d2*(((tReal)1.)-l2/me2)*Re(B0(0.,DBLtR(me2),DBLtR(l2)))-(((tReal)1.)-l2/(2*me2))*Re(B0(DBLtR(me2),DBLtR(me2),DBLtR(l2))); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"B1 test #3:   "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//B0 auto-lambda tests
	//Note: LT=> 'dB0(0,m,0)' is IR-safe
	//val1=(tReal)Re(DB0(0.,DBLtR(me2),DBLtR(l2))); val2=(tReal)Re(DB0(0.,DBLtR(me2),0.)); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	//if(!bSilent){std::cout<<"dB0 [lambda]:  "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}
	//dB0(m,m,l) =[if l<<1 only!]= dB0(m,m,0)
	//val1=(tReal)Re(DB0(DBLtR(me2),DBLtR(me2),DBLtR(l2))); val2=(tReal)Re(DB0(DBLtR(me2),DBLtR(me2),0.)); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	//if(!bSilent){std::cout<<"dB0 [lambda]2: "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//[Wrong] dB0(p2Small,m,m) =?= Xi[m]+2-(1-4*m2/p2)ln( 1-p2/2m2 )
	//val1=Re(B0(DBLtR(p2Small),DBLtR(me2),DBLtR(me2)));
	//val2=xiM+(tReal)2.-((tReal)1.-(4.*me2)/p2Small)*log(fabs((tReal)1.-p2Small/(2*me2)));
	//d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	//if(!bSilent){std::cout<<"B0(x,m,m):   "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//[Wrong] dB0(0,m,m) =?= 1/m2
	//val1=Re(DB0(0.,DBLtR(me2),DBLtR(me2))); val2=((tReal)1.)/me2; d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	//if(!bSilent){std::cout<<"dB0 test:    "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//dB0(m2,m,0) == 1/(2*m2)*(log(m2/l2)-2) <--ok!
	val1=Re(DB0(DBLtR(me2),DBLtR(me2),0.)); val2=f1d2/me2*(logML-(tReal)2.); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001) && (fabs(val1+val2)>1.e-14))bOK=false;
	if(!bSilent){std::cout<<"dB0(m2,m,0): "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//dB0(0,m,m) == 1/(6*m2) <--ok!
	val1=Re(DB0(0.,DBLtR(me2),DBLtR(me2))); val2=((tReal)1.)/(((tReal)6.)*me2); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001) && (fabs(val1+val2)>1.e-14))bOK=false;
	if(!bSilent){std::cout<<"dB0(0,m,m): "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//dMe == (-a/4Pi)*m*(3*Xi+4)  <--ok!
	val1=LTCalcDMe_QED(); val2=(-a4Pi)*me*((tReal)3.*xiM + (tReal)4.); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001) && (fabs(val1+val2)>1.e-14))bOK=false;
	if(!bSilent){std::cout<<"dMe test:    "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//dZaa == (-a/4Pi)*(4/3)*(Xi)
	val1=LTCalcDZaa_QED(); val2=(-a4Pi)*(((tReal)4.)/3.)*(xiM); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001) && (fabs(val1+val2)>1.e-14))bOK=false;
	if(!bSilent){std::cout<<"dZaa test:    "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//dZe == (-a/4Pi)*(-2/3)*(Xi)
	val1=LTCalcDZe_QED(); val2=(-a4Pi)*(-((tReal)2.)/3.)*(xiM); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001) && (fabs(val1+val2)>1.e-14))bOK=false;
	if(!bSilent){std::cout<<"dZe test:    "<<val1<<" : "<<val2<<" (delta: "<<d<<")"<<std::endl;}

	//dFe == (-a/4Pi)*( Xi[m] - 2*ln(m2/l2) + 4 )
	val1=LTCalcDFe_QED(); val2=(-a4Pi)*(xiM - ((tReal)2.)*logML + 4.); d=fabs((val1-val2)/((fabs(val1+val2)>0)?fabs(val1+val2):1.)); if(bOK && (d>0.001))bOK=false;
	if(!bSilent){std::cout<<"dFe test:    "<<val1<<" : "<<val2<<" (delta: "<<d<<", abs: "<<val1-val2<<")"<<std::endl;}

	//restore
	CAlHepLib::SetIRLambdaSqr(l2Old); CAlHepLib::SetUVChi(xiOld); CAlHepLib::SetDimMuSqr(mu2Old);

	//if(false){ //Additional tests
	if(true){ //Additional tests
		//std::cout<<"getversionkey(): "<<getversionkey()<<std::endl;
		//setversionkey(-1);

////////////////////////////////////// LoopTools NaN points
//1. D0i(dd002,me2,q2,0,me2,me2+2*k_k2,2*me2-2*k1_k2,0,me2,me2,me2)
//[Event <-- LT_NaN_evt1.bin/evt]
//0. [11] 2.760000e+01, 0.000000e+00, 0.000000e+00, 2.760000e+01  M=5.109989e-04
//1. [2212] 9.200000e+02, 0.000000e+00, 0.000000e+00, -9.199995e+02  M=9.382720e-01
//2. [11] 1.257693e+01, 7.103989e-03, 7.606153e-03, 1.257692e+01  M=5.109989e-04
//3. [22] 1.502308e+01, 8.289951e-03, -8.557227e-03, 1.502307e+01  M=-5.836838e-28
//4. [2212] 9.200000e+02, -1.539394e-02, 9.510741e-04, -9.199995e+02  M=9.382720e-01
//q2=-2.37877940526132028554e-04
//(k+k2)^2=2.61189030306149902781e-04
//kk1=1.30463955971735953881e-04
//kk2=1.30463955210064357355e-04
//(k1-k2)^2=-2.37877939002788835503e-04
//k1_k2=1.19200089387415605822e-04
//(k1-k2)^2-q2=1.52334319305100381691e-12
		double me2=2.61119886e-7;
		//NaN
		if(!bSilent)std::cout<<"D0i dd002: "<<D0i(dd002,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779390e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//Output: D0i dd002: (nan,nan)
		if(!bSilent)std::cout<<"D0i dd001: "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779390e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//Output: D0i dd001: (nan,nan)
		if(!bSilent)std::cout<<"D0i[002] ..86 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779386e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..86 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..87 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779387e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..87 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..88 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779388e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..88 (nan,nan)
		if(!bSilent)std::cout<<"D0i[002] ..89 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779389e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..89 (nan,nan)
		if(!bSilent)std::cout<<"D0i[002] ..90 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779390e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..90 (nan,nan)
		if(!bSilent)std::cout<<"D0i[002] ..91 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779391e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..91 (nan,nan)
		if(!bSilent)std::cout<<"D0i[002] ..92 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779392e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..92 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..93 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779393e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..93 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..94 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779394e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..94 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..95 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779395e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..95 (nan,nan)
		if(!bSilent)std::cout<<"D0i[002] ..96 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779396e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..96 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..97 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779397e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..97 (825.008,645.411)
		if(!bSilent)std::cout<<"D0i[002] ..98 "<<D0i(dd001,me2,-2.378779405e-4,0.,me2,2.611890303e-4,-2.378779398e-4,0,me2,me2,me2)<<std::endl; clearcache();
		//D0i[002] ..98 (nan,nan)

		double q2=-2.37877940526132028554e-04;
		me2=2.61119886021188070506e-07;
		double k1mk2P2=-2.37877939002788835503e-04;
		double kpk2P2=2.61189030306149902781e-04;
		double kk1=1.30463955971735953881e-04;
		double kk2=1.30463955210064357355e-04;
		double k1k2=1.19200089387415605822e-04;
		//(k1-k2)^2-q2=1.523343193051e-12

		me2=2.61119886e-7;
		if(!bSilent)std::cout<<"C0i g: "<<C0i(cc11,-2.37877939e-04,0,-2.37877940526e-04,me2,me2,me2)<<std::endl;
		//C0i g: (-8.08587e+11,-0)
		//(REAL*16) C0i g: (-7460.7,-0)
		if(!bSilent)std::cout<<"C0i h: "<<C0i(cc11,-2.37877939e-04,0,-2.378779405e-04,me2,me2,me2)<<std::endl;
		//C0i h: (4.77356e+12,-0)
		//(REAL*16) C0i h: (-7460.7,-0)
		if(!bSilent)std::cout<<"C0i i: "<<C0i(cc11,-2.37877939e-04,0,-2.3787794e-04,me2,me2,me2)<<std::endl;
		//C0i i: (nan,nan)
		//(REAL*16) C0i i: (-7460.7,-0)
		if(!bSilent)std::cout<<"C0i j: "<<C0i(cc1,-2.37877939e-04,0,-2.3787794e-04,me2,me2,me2)<<std::endl;
		//C0i j: (inf,nan)
		//(REAL*16) C0i j: (12226.3,0)
	}
	//////////////////////////////////////
#else
	bOK=false;
#endif //VM_USE_LOOPTOOLS
	if(!bSilent){std::cout<<"LoopTools test "<<(bOK?"finished successfully":"failed")<<std::endl;} return bOK;
}

// IR-delimiter
tReal CAlHepLib::GetIRLambdaSqr() {
#ifdef VM_USE_LOOPTOOLS
	return (tReal)getlambda();
#else
	return m_IRLambda2;
#endif //VM_USE_LOOPTOOLS
}
void CAlHepLib::SetIRLambdaSqr(tReal lambdaSqr) {
	if(lambdaSqr<=0.){ std::cout<<"ERROR: SetIRLambdaSqr("<<lambdaSqr<<") call."; return; }
#ifdef VM_USE_LOOPTOOLS
	setlambda(DBLtR(lambdaSqr)); m_IRLambda2=getlambda();
#else
	m_IRLambda2=lambdaSqr;
#endif //VM_USE_LOOPTOOLS
}

//UV-delimiters
tReal CAlHepLib::GetUVChi() {
#ifdef VM_USE_LOOPTOOLS
	return (tReal)getdelta();
#else
	return m_UVChi;
#endif //VM_USE_LOOPTOOLS
}
tReal CAlHepLib::GetDimMuSqr() {
#ifdef VM_USE_LOOPTOOLS
	return (tReal)getmudim();
#else
	return m_DimMuSqr;
#endif //VM_USE_LOOPTOOLS
}
void CAlHepLib::SetUVChi(tReal chi) {
#ifdef VM_USE_LOOPTOOLS
	setdelta(DBLtR(chi)); m_UVChi=getdelta();
#else
	m_UVChi=chi;
#endif //VM_USE_LOOPTOOLS
}
void CAlHepLib::SetDimMuSqr(tReal mudim2) {
	if(mudim2<=0.){ std::cout<<"ERROR: SetDimMuSqr("<<mudim2<<") call."; return; }
#ifdef VM_USE_LOOPTOOLS
	setmudim(DBLtR(mudim2)); m_DimMuSqr=getmudim();
#else
	m_DimMuSqr=mudim2;
#endif //VM_USE_LOOPTOOLS
}

/*
*	Calculate renorm. constants (currently in QED)
*	Note: we don't use lambda as integral parameter. It will be inserted automatically by LoopTools.
*	Integrals with manually inserted lambda are correct ONLY if (lambda << any).
*/
tReal CAlHepLib::LTCalcDFe_QED() {
//QED: delta{F_e}= (-a/4Pi)*(-1)*( 1 - B0(0,m,l) + 4*m2*dB0dP2(m2,m,l) )
#ifdef VM_USE_LOOPTOOLS
	tReal me = CVMConst::M_e; tReal me2 = me*me;
	tReal f = CVMConst::Alpha/(4.*CVMConst::Pi()); //(-a/4Pi)*(-1)
	tReal res = f*( ((tReal)1.)-(tReal)Re(B0(0.,DBLtR(me2),0.))+4.*me2*(tReal)Re(DB0(DBLtR(me2),DBLtR(me2),0.)) );
	return res;
#else
	return 0.;
#endif //VM_USE_LOOPTOOLS
}
tReal CAlHepLib::LTCalcDMe_QED() {
//QED: delta{m_e}= (-a/4Pi)*m*(2B0(m2,m,l)+B0(0,m,l)-1)
#ifdef VM_USE_LOOPTOOLS
	tReal me = CVMConst::M_e; tReal me2 = me*me;
	tReal f = -CVMConst::Alpha/(4.*CVMConst::Pi())*me;
	tReal res = f*(2*(tReal)Re(B0(DBLtR(me2),DBLtR(me2),0.)) + (tReal)Re(B0(0.,DBLtR(me2),0.)) - 1.);
	//res*=(-1.);//debug
	return res;
#else
	return 0.;
#endif //VM_USE_LOOPTOOLS
}
tReal CAlHepLib::LTCalcDZaa_QED() {
//QED: delta{Z_{AA}}= (-a/4Pi)*(4/3)*(B0(0,m,m)+2*m2*dB0dk2(0,m,m)-1/3) == (-a/4Pi)*(4/3)*(B0(0,m,m))
	//asdfg: use simplified expression after debug!
#ifdef VM_USE_LOOPTOOLS
	tReal me = CVMConst::M_e; tReal me2 = me*me;
	tReal f = (-CVMConst::Alpha/(4.*CVMConst::Pi()))*(((tReal)4.)/3.);
	//tReal res = f*( (tReal)Re(B0(0.,DBLtR(me2),DBLtR(me2))) + 2.*me2*(tReal)Re(DB0(0.,DBLtR(me2),DBLtR(me2))) - ((tReal)1.)/3. );
	tReal res = f*( (tReal)Re(B0(0.,DBLtR(me2),DBLtR(me2))) );
	//res*=-1.; //asdfg: debug
	return res;
#else
	return 0.;
#endif //VM_USE_LOOPTOOLS
}
tReal CAlHepLib::LTCalcDZe_QED() {
//QED: delta{Z_e} = -1/2*delta{Z_{AA}} (QED only)
	tReal res= -LTCalcDZaa_QED()/2.;
	return res;
}

//Self tests
bool CAlHepLib::TestReal(bool bLog) { //Test if precision of tReal is reduced by compiler setting
	if(sizeof(tReal)<=8)return true; //double or float
	int iPrecExpected=sizeof(tReal)*2-1;
	tReal epsCalc=pow((tReal)10.,-iPrecExpected);
	if(bLog)std::cout<<"[prec. test] --- Precision test --- (sizeof(real)=="<<sizeof(tReal)<<")"<<std::endl;
	if(bLog)std::cout<<"[prec. test] Mantissa digits expected: "<<iPrecExpected<<std::endl;
	tReal a,b,c;
	//1.
	if(bLog)std::cout<<"[prec. test] 'eps' declared: "<<epsReal<<", expected: "<<epsCalc<<std::endl;
	if(epsReal>1000*epsCalc){ std::cout<<"[err] declared precision ('eps') is smaller than: "<<epsReal<<", expected: "<<epsCalc<<std::endl; return false; }
	//2.
	bool bRes=true;
	a=2.0; b=sqrt(a); b=b*b; //same if a=qd_real("2.0");
	if(bLog)std::cout<<"[prec. test] 'sqrt' function test. |{sqrt(x)}^2-x|/x: "<<fabs(b-a)/((a>0.)?a:1.)<<" at x="<<a<<std::endl;
	if(fabs(b-a)>1000*a*epsReal){ std::cout<<"[err] 'sqrt' is not precise. Diff: "<<fabs(a-b)/((a>0.)?a:1.)<<" at x="<<a<<std::endl; bRes=false; }
	a=2.1; b=sqrt(a); b=b*b; //same if a=qd_real("2.1");
	if(bLog)std::cout<<"[prec. test] 'sqrt' function test. |{sqrt(x)}^2-x|/x: "<<fabs(b-a)/((a>0.)?a:1.)<<" at x="<<a<<std::endl;
	if(fabs(b-a)>1000*a*epsReal){ std::cout<<"[err] 'sqrt' is not precise. Diff: "<<fabs(a-b)/((a>0.)?a:1.)<<" at x="<<a<<std::endl; bRes=false; }
	a=2.1; b=exp(log(a)/((tReal)2.)); b=b*b; //same if a=qd_real("2.1");
	if(bLog)std::cout<<"[prec. test] 'exp(log(a)/2)' function test. |{exp(log(x)/2)}^2-x|/x: "<<fabs(b-a)/((a>0.)?a:1.)<<" at x="<<a<<std::endl;
	if(fabs(b-a)>1000*a*epsReal){ std::cout<<"[err] 'exp(log(a)/2)' is not precise. Diff: "<<fabs(a-b)/((a>0.)?a:1.)<<" at x="<<a<<std::endl; bRes=false; }
	if(!bRes)return false;
	//3.
	a=CVMConst::M_p; b=920.; //a=CVMConst::M_e; b=27.6;
	if(bLog){ int iCoutPrec=std::cout.precision(20); std::cout<<"[prec. test] Constant test. Me: "<<a<<std::endl; std::cout.precision(iCoutPrec); }
	CVM4Vec v(b,0.,0.,sqrt(b*b-a*a)); c=fabs((v.GetSqr()-(a*a))/((a>0.)?(a*a):1.));
	if(bLog)std::cout<<"[prec. test] Test vector square relative error: "<<c<<std::endl;
	if(c>1e7*epsReal){ std::cout<<"[err] Vector precision error. Square error: "<<c<<", allowed: "<<1e7*epsReal<<std::endl; return false; }
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
CVM4Vec::CVM4Vec() { m_E=0.; m_Px=0., m_Py=0., m_Pz=0.; }
CVM4Vec::CVM4Vec(tReal _E, tReal _Px, tReal _Py, tReal _Pz) { m_E=_E; m_Px=_Px, m_Py=_Py, m_Pz=_Pz; }
CVM4Vec::CVM4Vec(const CVM4Vec& vec) { m_E=vec.m_E; m_Px=vec.m_Px, m_Py=vec.m_Py, m_Pz=vec.m_Pz; }
CVM4Vec& CVM4Vec::operator=(const CVM4Vec& vec) { m_E=vec.m_E; m_Px=vec.m_Px, m_Py=vec.m_Py, m_Pz=vec.m_Pz; return *this; }
CVM4Vec& CVM4Vec::operator+=(const CVM4Vec& vec) { m_E+=vec.m_E; m_Px+=vec.m_Px, m_Py+=vec.m_Py, m_Pz+=vec.m_Pz; return *this; }
CVM4Vec& CVM4Vec::operator-=(const CVM4Vec& vec) { m_E-=vec.m_E; m_Px-=vec.m_Px, m_Py-=vec.m_Py, m_Pz-=vec.m_Pz; return *this; }
CVM4Vec& CVM4Vec::operator*=(const tReal& f) { m_E*=f; m_Px*=f, m_Py*=f, m_Pz*=f; return *this; }
CVM4Vec CVM4Vec::operator+(const CVM4Vec& vec) const { return CVM4Vec(m_E+vec.m_E,m_Px+vec.m_Px,m_Py+vec.m_Py,m_Pz+vec.m_Pz); }
CVM4Vec CVM4Vec::operator-(const CVM4Vec& vec) const { return CVM4Vec(m_E-vec.m_E,m_Px-vec.m_Px,m_Py-vec.m_Py,m_Pz-vec.m_Pz); }
bool CVM4Vec::operator==(const CVM4Vec& vec) const { return ((m_E==vec.m_E)&&(m_Pz==vec.m_Pz)&&(m_Px==vec.m_Px)&&(m_Py==vec.m_Py)); }
//CVM4Vec operator+(const CVM4Vec& vec1, const CVM4Vec& vec2) { return CVM4Vec(vec1.m_E+vec2.m_E,vec1.m_Px+vec2.m_Px,vec1.m_Py+vec2.m_Py,vec1.m_Pz+vec2.m_Pz); }
void CVM4Vec::Set(tReal _E, tReal _Px, tReal _Py, tReal _Pz) { m_E=_E; m_Px=_Px, m_Py=_Py, m_Pz=_Pz; }
bool CVM4Vec::Load(std::ifstream* ifs, bool _binary) {
	if(ifs==NULL)return false;
	if(_binary){ ifs->read((char*)&m_E,sizeof(m_E));ifs->read((char*)&m_Px,sizeof(m_Px));ifs->read((char*)&m_Py,sizeof(m_Py));ifs->read((char*)&m_Pz,sizeof(m_Pz)); }else{ *ifs>>m_E>>m_Px>>m_Py>>m_Pz; }
	return true;
}
bool CVM4Vec::Save(std::ostream* ofs, bool _binary) const {
	if(ofs==NULL)return false;
	if(_binary){ ofs->write((char*)&m_E,sizeof(m_E));ofs->write((char*)&m_Px,sizeof(m_Px));ofs->write((char*)&m_Py,sizeof(m_Py));ofs->write((char*)&m_Pz,sizeof(m_Pz)); }else{ *ofs<<m_E<<" "<<m_Px<<" "<<m_Py<<" "<<m_Pz; }
	return true;
}
tReal CVM4Vec::GetMassSigned() const {
	tReal m2=m_E*m_E-m_Px*m_Px-m_Py*m_Py-m_Pz*m_Pz; bool bInvSign=false; if(m2<0.){ m2=-m2; bInvSign=true; }
	tReal res=sqrt((tReal)m2); if(bInvSign)res=-res; return res;
}
tReal CVM4Vec::GetSqr() const { return (m_E*m_E-m_Px*m_Px-m_Py*m_Py-m_Pz*m_Pz); }
tReal CVM4Vec::GetCpl(const CVM4Vec& vec) const { return (m_E*vec.m_E-m_Px*vec.m_Px-m_Py*vec.m_Py-m_Pz*vec.m_Pz); }
tReal CVM4Vec::GetPAbs() const { return sqrt(m_Px*m_Px+m_Py*m_Py+m_Pz*m_Pz); }
tReal CVM4Vec::GetPSqr() const { return m_Px*m_Px+m_Py*m_Py+m_Pz*m_Pz; }
tReal CVM4Vec::GetPt() const { return sqrt(m_Px*m_Px+m_Py*m_Py); /*no sign check is required here*/ }
tReal CVM4Vec::GetPtSqr() const { return (m_Px*m_Px+m_Py*m_Py); }
tReal CVM4Vec::GetPhi() const { return ( ((m_Py!=0.)||(m_Px!=0.)) ? (atan2(m_Py,m_Px)) : 0.); }
tReal CVM4Vec::GetThetaCos() const { //polar angle cosine
	tReal P=GetPAbs(); if(P<=0.){ std::cout<<"ERROR in CVM4Vec::GetThetaCos(): ZERO momentum"<<std::endl; return 0.; }
	return (m_Pz/P);
}
tReal CVM4Vec::GetTheta() const { //polar angle (uses sin() or cos() depending on angle value)
	tReal P=GetPAbs(); if(P<=0.){ std::cout<<"ERROR in CVM4Vec::GetTheta(): ZERO momentum"<<std::endl; return 0.; }
	tReal theta;
	if(m_Pz>(0.9*P)){
		tReal pt=GetPt();
		if(fabs(pt)>P){ //fix eps-level error [Unused: Pabs is calculated via Pi ==> there may NOT be eps-errors]
			if(fabs(fabs(pt)-P)>P*1.e-10){ std::cout<<"ERROR in CVM4Vec::GetTheta(): Pt>Pabs"<<std::endl; return 0.; }
			pt=((pt>0.)?P:(-P));
		}
		theta=asin(pt/P);
	}else{
		theta=acos(m_Pz/P); //Nt. Pabs is calculated via Pz ==> there may NOT be Pz>Pabs]
	}
	return theta;
}
tReal CVM4Vec::GetEta() const { //= -log(tan(Theta/2)) = 0.5*log( (P+Pz)/(P-Pz) )
	//Note: "0.5*log( (E+Pz)/(E-Pz) )" is a different value, ~~ but != eta
	tReal absPz=fabs(m_Pz);
	if ((m_Px==0.) && (m_Py==0.)) {
		//infinite rapidity -> UNIQUE huge number [FastJet comments: 'unique' numbers can be relevant at parton-level]
		return ((m_Pz>0.)?(1.e10+absPz):(-1.e10-absPz));
	}
	tReal pt2=m_Px*m_Px+m_Py*m_Py; tReal P=sqrt(pt2+m_Pz*m_Pz);
	tReal eta=-0.5*log(pt2/((P+absPz)*(P+absPz)));
	if(m_Pz<0.)eta=-eta;
	return eta;
	//Calculation of 0.5*log( (E+Pz)/(E-Pz) )
	//Note:
	//*	May occur (m_E==absPz) && (Pt>0) due to save/load precision lost (like "777.777,0.0001,0.0001,777.777")
	//*	ToDo: always save momentum as binary!
	//*	Next line fails if (m_E==absPz)
	//*	tReal eta=-0.5*log(fabs((m_E-absPz)/(m_E+absPz))); if(m_Pz<0.)eta=-eta; return eta;
	// tReal m2=m_E*m_E-m_Pz*m_Pz-pt2; if(m2<0.)m2=0.; //'Fix' (m_E==absPz) case
	// tReal etaBis=-0.5*log((m2+pt2)/((m_E+absPz)*(m_E+absPz))); if(m_Pz<0.)etaBis=-etaBis; return etaBis;
}
tReal CVM4Vec::GetAngleCos3Vec(const CVM4Vec& vec) const { // Cosine of angle between 3-vectors
	tReal p1mod=sqrt(m_Px*m_Px+m_Py*m_Py+m_Pz*m_Pz); // no sign check is required here
	tReal p2mod=sqrt(vec.m_Px*vec.m_Px+vec.m_Py*vec.m_Py+vec.m_Pz*vec.m_Pz); // no sign check is required here
	tReal p12cpl=m_Px*vec.m_Px+m_Py*vec.m_Py+m_Pz*vec.m_Pz;
	tReal res=0.; if((p1mod>0.)&&(p2mod>0.))res=p12cpl/(p1mod*p2mod); // 1/0 check
	return res;
}
tReal CVM4Vec::GetDRSqr(const CVM4Vec& vec) const { // delta-R squared
	tReal dEta=(GetEta()-vec.GetEta());
	tReal dPhi=(GetPhi()-vec.GetPhi()); tReal pi=CVMConst::Pi(); while(dPhi>pi){dPhi-=pi*2;} while(dPhi<=-pi){dPhi+=pi*2;}
	return (dEta*dEta+dPhi*dPhi);
}
void CVM4Vec::LorMToK(tReal m, const CVM4Vec& k) { //Lorentz transform with system transform: [m,0,0,0]->k
	if(m==0.)return;
	//debug: if(k.GetSqr()-m*m > m*m*1000*epsReal) { std::cout<<"CVM4Vec::LorMToK(): invalid parameter"<<std::endl; return; }
	//debug: if(k.m_E<0) { std::cout<<"CVM4Vec::LorMToK(): Invalid k= "<<k<<std::endl; }

	//Note: Both space- & time-like vetors are transformed here
	tReal resE=(m_E*k.m_E + m_Px*k.m_Px + m_Py*k.m_Py + m_Pz*k.m_Pz)/m; tReal x=(m_E+resE)/(k.m_E+m);
	m_E=resE; m_Px+=k.m_Px*x; m_Py+=k.m_Py*x; m_Pz+=k.m_Pz*x;

	//Truncate ghost components: tReal qeps=2.*m_E*epsReal; if((m_Px<qeps)&&(m_Px>-qeps))m_Px=0.; if((m_Py<qeps)&&(m_Py>-qeps))m_Py=0.; if((m_Pz<qeps)&&(m_Pz>-qeps))m_Pz=0.;
}
void CVM4Vec::LorKToM(const CVM4Vec& k, tReal m) { //Lorentz transform with system transform: k->[m,0,0,0]
	if(m==0.)return;
	//debug: if(k.GetSqr()-m*m > m*m*1000*epsReal) { std::cout<<"CVM4Vec::LorKToM(): invalid parameter"<<std::endl; return; }
	//debug: if(k.m_E<0) { std::cout<<"CVM4Vec::LorKToM(): Invalid k= "<<k<<std::endl; }

	//Note: Both space- & time-like vetors are transformed here
	tReal resE=(m_E*k.m_E-m_Px*k.m_Px-m_Py*k.m_Py-m_Pz*k.m_Pz)/m; tReal x=(m_E+resE)/(k.m_E+m);
	m_E=resE; m_Px-=k.m_Px*x; m_Py-=k.m_Py*x; m_Pz-=k.m_Pz*x;

	//Truncate ghost components: tReal qeps=2.*m_E*epsReal; if((m_Px<qeps)&&(m_Px>-qeps))m_Px=0.; if((m_Py<qeps)&&(m_Py>-qeps))m_Py=0.; if((m_Pz<qeps)&&(m_Pz>-qeps))m_Pz=0.;
}
void CVM4Vec::RotateXY(tReal phi) { RotateXY(sin(phi),cos(phi)); } //Rotate vector in XY-plane (preserving pT)
void CVM4Vec::RotateXY(tReal cosPhi, tReal sinPhi) { //Rotate vector in XY-plane (preserving pT)
	//cos(x+y)=cosx*cosy-sinx*siny; sin(x+y)=sinx*cosy+cosx*siny;
	tReal px=m_Px*cosPhi-m_Py*sinPhi; tReal py=m_Py*cosPhi+m_Px*sinPhi; m_Px=px; m_Py=py;
}
void CVM4Vec::RotateXZ(tReal phi) { RotateXZ(sin(phi),cos(phi)); } //Rotate vector in XZ-plane
void CVM4Vec::RotateXZ(tReal cosPhi, tReal sinPhi) { //Rotate vector in XZ-plane
	tReal px=m_Px*cosPhi-m_Pz*sinPhi; tReal pz=m_Pz*cosPhi+m_Px*sinPhi; m_Px=px; m_Pz=pz;
}
void CVM4Vec::RotateYZ(tReal phi) { RotateYZ(sin(phi),cos(phi)); } //Rotate vector in YZ-plane
void CVM4Vec::RotateYZ(tReal cosPhi, tReal sinPhi) { //Rotate vector in YZ-plane
	tReal py=m_Py*cosPhi-m_Pz*sinPhi; tReal pz=m_Pz*cosPhi+m_Py*sinPhi; m_Py=py; m_Pz=pz;
}
/*
*	RotateZYYX: rotate vector in ZY+YX planes accroding to {0,0,zPAbs}->{zRes} transformation.
*	Nt. {0,0,zPAbs}->{0,0,-zPAbs} transformation is treated as pi-rotation in YZ-plane
*	'zPAbs' and 'zResPt' are used to avoid extra 'sqrt()' operations.
*/
void CVM4Vec::RotateZYYX(const CVM4Vec& zRes, tReal zPAbs, tReal zResPt) {
	if(zPAbs<0.){ zPAbs=zRes.GetPAbs(); }
	if(zResPt<0.){ zResPt=zRes.GetPt(); }
	if(zPAbs<=0.)return; //Unable to rotate
	if(zResPt>0.){
		tReal x=m_Px; tReal y=m_Py; tReal z=m_Pz; tReal zResX=zRes.m_Px; tReal zResY=zRes.m_Py; tReal zResZ=zRes.m_Pz;
		m_Px= x*zResY/zResPt +y*zResZ*zResX/zResPt/zPAbs +z*zResX/zPAbs;
		m_Py=-x*zResX/zResPt +y*zResZ*zResY/zResPt/zPAbs +z*zResY/zPAbs;
		m_Pz=                -y*zResPt/zPAbs             +z*zResZ/zPAbs;
	}else if((zResPt==0.) && (zRes.m_Pz<0.)){ //Pi-rotate in YZ-plane if Pz is changed
		m_Py=-m_Py; m_Pz=-m_Pz;
	}
}

///////////////////////////////////////////////////////////// CHEPParticleLite
const int CHEPParticleLite::STATUS_EMPTY	= 0;	//hepevt.isthep
const int CHEPParticleLite::STATUS_FINAL	= 1;	//hepevt.isthep
const int CHEPParticleLite::STATUS_INTERNAL	= 2;	//hepevt.isthep
const int CHEPParticleLite::STATUS_INITIAL	= -1;	//hepevt.isthep == -1 (as in MG)
const int CHEPParticleLite::STATUS_DOC		= 21;	//hepevt.isthep

CHEPParticleLite::CHEPParticleLite() {
	m_PID=-1; m_Mark=0;
	m_Status=CHEPParticleLite::STATUS_EMPTY;
	m_Mother1_f0=0; m_Mother2_f0=0; m_Daughter1_f0=0; m_Daughter2_f0=0; m_Pt=0.; m_Eta=0.; m_Phi=0.; m_Rapidity=0.;
}
CHEPParticleLite::CHEPParticleLite(const CHEPParticleLite& p) {
	//*this = p;
	m_PID=p.m_PID; m_Status=p.m_Status;
	m_Mother1_f0=p.m_Mother1_f0; m_Mother2_f0=p.m_Mother2_f0; m_Daughter1_f0=p.m_Daughter1_f0; m_Daughter2_f0=p.m_Daughter2_f0;
	m_Mom=p.m_Mom; m_Pt=p.m_Pt; m_Eta=p.m_Eta; m_Phi=p.m_Phi; m_Rapidity=p.m_Rapidity; m_Pos=p.m_Pos;
	m_Mark=0;
}

CHEPParticleLite &CHEPParticleLite::operator=(const CHEPParticleLite& p) {
	//TODO: re-implement using 'mmove'
	m_PID=p.m_PID; m_Status=p.m_Status;
	m_Mother1_f0=p.m_Mother1_f0; m_Mother2_f0=p.m_Mother2_f0; m_Daughter1_f0=p.m_Daughter1_f0; m_Daughter2_f0=p.m_Daughter2_f0;
	m_Mom=p.m_Mom; m_Pt=p.m_Pt; m_Eta=p.m_Eta; m_Phi=p.m_Phi; m_Rapidity=p.m_Rapidity; m_Pos=p.m_Pos;
	return *this;
}

//"id status m1 m2 d1 d2 E px py pz pt eta phi rap T x y z"
//-12 1 542 -1 -1 -1 E534.972 px0.198941 py-0.0840353 pz-534.972 pt0.215962 eta-8.50802 phi-0.399678 -inf 593.166 0.00613992 0.249814 -593.165
bool CHEPParticleLite::Load(std::ifstream* ifs, bool _binary) {
	if(ifs==NULL)return false;
	if(_binary) {
		*ifs>>m_PID>>m_Status>>m_Mother1_f0>>m_Mother2_f0>>m_Daughter1_f0>>m_Daughter2_f0;
		char c=0; *ifs>>c; for(int iSp=0;((iSp<10)&&(c==' '));iSp++){*ifs>>c;} if(c!='{')return false;
		m_Mom.Load(ifs,_binary);
		ifs->read((char*)&m_Pt, sizeof(m_Pt)); ifs->read((char*)&m_Eta,sizeof(m_Eta));
		ifs->read((char*)&m_Phi,sizeof(m_Phi)); ifs->read((char*)&m_Rapidity,sizeof(m_Rapidity));
		m_Pos.Load(ifs,_binary);
		*ifs>>c; if(c!='}')return false;
	} else {
		*ifs>>m_PID>>m_Status>>m_Mother1_f0>>m_Mother2_f0>>m_Daughter1_f0>>m_Daughter2_f0;
		m_Mom.Load(ifs,false); *ifs>>m_Pt>>m_Eta>>m_Phi>>m_Rapidity; m_Pos.Load(ifs,false);
	}
	return true;
}

bool CHEPParticleLite::Save(std::ofstream* ofs, bool _binary) {
	if(ofs==NULL)return false;
	if (_binary) {
		*ofs<<m_PID<<" "<<m_Status<<" "<<m_Mother1_f0<<" "<<m_Mother2_f0<<" "<<m_Daughter1_f0<<" "<<m_Daughter2_f0;
		*ofs<<'{'; //start of binary block
		m_Mom.Save(ofs,_binary);
		ofs->write((char*)&m_Pt,sizeof(m_Pt)); ofs->write((char*)&m_Eta,sizeof(m_Eta));
		ofs->write((char*)&m_Phi,sizeof(m_Phi)); ofs->write((char*)&m_Rapidity,sizeof(m_Rapidity));
		m_Pos.Save(ofs,_binary);
		*ofs<<'}'; //end of binary block
	} else {
		*ofs<<m_PID<<" "<<m_Status<<" "<<m_Mother1_f0<<" "<<m_Mother2_f0<<" "<<m_Daughter1_f0<<" "<<m_Daughter2_f0<<" ";
		m_Mom.Save(ofs,_binary); *ofs<<" "<<m_Pt<<" "<<m_Eta<<" "<<m_Phi<<" "<<m_Rapidity<<" "; m_Pos.Save(ofs,_binary);
	}
	return true;
}

bool CHEPParticleLite::IsParticleOK() { return (m_PID>=0); }
tReal CHEPParticleLite::GetMass() { tReal m=m_Mom.GetMassSigned(); if(m<0.){ m=0.; /*something like '-1e-16'*/ } return m; }
void CHEPParticleLite::Log(bool bShortLog) {
	std::cout<<CAlHepLib::GetParticleTitle(m_PID)<<": ";
	if(bShortLog) {
		std::cout<<" Eta="<<m_Eta<<", Pt= "<<m_Pt<<", Phi="<<m_Phi<<", Mom="<<m_Mom;
	} else {
		std::cout<<" Eta="<<m_Eta<<" ("<<m_Mom.GetEta()<<"), Rap="<<m_Rapidity<<", Pt="<<m_Pt<<" ("<<m_Mom.GetPt()<<"), Phi="<<m_Phi<<" ("<<m_Mom.GetPhi()<<"), Mom="<<m_Mom;
	}
	std::cout<<std::endl;
}

///////////////////////////////////////////////////////////// Particle order
CHEPParticleOrderLite::CHEPParticleOrderLite(int nParticles) {
	m_nParticles=nParticles; if(m_nParticles<0)m_nParticles=2;
	m_order=new int[m_nParticles]; for(int i=0;i<m_nParticles;i++){m_order[i]=i;}
}
CHEPParticleOrderLite::CHEPParticleOrderLite(const CHEPParticleOrderLite& order) {
	m_nParticles=order.m_nParticles;
	m_order=new int[m_nParticles]; for(int i=0;i<m_nParticles;i++){m_order[i]=order.m_order[i];}
}
CHEPParticleOrderLite::CHEPParticleOrderLite(int nParticles, int* orderIToX) {
	m_nParticles=nParticles; if(m_nParticles<0)m_nParticles=2;
	m_order=new int[m_nParticles]; for(int i=0;i<m_nParticles;i++){ m_order[i]=((orderIToX!=NULL)?orderIToX[i]:i); }
}
CHEPParticleOrderLite::~CHEPParticleOrderLite() { if(m_order!=NULL){ delete[] m_order; m_order=NULL; } }
void CHEPParticleOrderLite::ResetOrder() { //to straight order
	if(m_order!=NULL){ for(int i=0;i<m_nParticles;i++){m_order[i]=i;} }
}
void CHEPParticleOrderLite::SetOrder(int* orderIToX) {
	if((m_order!=NULL)&&(orderIToX!=NULL)){ for(int i=0;i<m_nParticles;i++){ m_order[i]=orderIToX[i]; } }
}
void CHEPParticleOrderLite::Swap(int i, int j){ int tmp=m_order[i]; m_order[i]=m_order[j]; m_order[j]=tmp; }
int CHEPParticleOrderLite::GetOrigin(int i) const {
	int res=-1; for(int iTmp=0;iTmp<m_nParticles;iTmp++){ if(i==m_order[iTmp]){ res=iTmp; break; } } return res;
}
bool CHEPParticleOrderLite::CheckConsistency() const {
	int* tmp=new int[m_nParticles]; if(tmp==NULL)return false; int i; bool bRes=true;
	for(i=0;i<m_nParticles;i++){ tmp[i]=0; }
	for(i=0;i<m_nParticles;i++){ int pos=m_order[i]; if((pos<0)||(pos>=m_nParticles)||(tmp[pos]!=0)){tmp=false;break;} tmp[pos]++; }
	for(i=0;i<m_nParticles;i++){ if(tmp[i]!=1){bRes=false;break;} } delete[] tmp; return bRes;
}
CHEPParticleOrderLite* CHEPParticleOrderLite::CreateBackwardOrder() const {
	int* tmp=new int[m_nParticles]; if(tmp==NULL)return NULL; int i; bool bConsistent=true;
	for(i=0;i<m_nParticles;i++){ tmp[i]=-1; }
	for(i=0;i<m_nParticles;i++){ int pos=m_order[i]; if((pos<0)||(pos>=m_nParticles)||(tmp[pos]>=0)){bConsistent=false;break;} tmp[pos]=i; }
	CHEPParticleOrderLite* res=(bConsistent?(new CHEPParticleOrderLite(m_nParticles,tmp)):NULL);
	delete[] tmp; return res;
}

///////////////////////////////////////////////////////////// CPDFLite
CPDFLite::CPDFLite() {}
CPDFLite::~CPDFLite() {}
CPDFLite* CPDFLite::Create(long gup, long sup, int idBeam) {
#ifdef VM_USE_PYTHIA8
	if((gup==4) && (sup==46)) {
		return new CPDFPyCTEQ5L(idBeam);
	}
#endif //#ifdef VM_USE_PYTHIA8
	return NULL;
}

///////////////////////////////////////////////////////////// CPDFLite-derived
#ifdef VM_USE_PYTHIA8
CPDFPyCTEQ5L::CPDFPyCTEQ5L(int idBeamIn) : CPDFLite(), pdf(idBeamIn) { idBeam=idBeamIn; }
CPDFPyCTEQ5L::~CPDFPyCTEQ5L() {}
bool CPDFPyCTEQ5L::IsReady() { return pdf.isSetup(); }
double CPDFPyCTEQ5L::GetDensity(int id, double x, double Q2) {
	return ((x>0.)?(pdf.xf(id, x, Q2)/x):0.);
}
CPDFLite* CPDFPyCTEQ5L::Clone() const { return new CPDFPyCTEQ5L(idBeam); }
#endif //#ifdef VM_USE_PYTHIA8

///////////////////////////////////////////////////////////// CHEPEvtLite
CHEPEvtLite::CCouplingsInfo::CCouplingsInfo() {
	m_CplsFilled=false; m_iCplsFilledAfter=0; m_QCplsFilled=false; m_Q2abs=-1.;
	//m_Cpls[EVTL_MAX_CPL_PART][EVTL_MAX_CPL_PART];
	//tReal m_QCpls[EVTL_MAX_CPL_PART]; //used by GetQCpl()
	//tReal m_UserCpls[EVTL_MAX_CPL_PART];
	//m_vecQ;
}
CHEPEvtLite::CHEPEvtLite() {
	m_nParticles=-1;
	idProc=0; wgt=-1.; wgtErr=0.; scale=1e3; a_QED=0.7546772e-2; a_QCD=0.1178469; //default values
	m_ChOrderBuf=NULL; m_szChOrderBuf=0;
}
/*CHEPEvtLite::CHEPEvtLite(const CHEPEvtLite& evt) {
	m_nParticles=evt.m_nParticles;
	for(int i=0; i<m_nParticles; i++) { m_Particles[i]=evt.m_Particles[i]; }
	m_Q2abs=evt.m_Q2abs;
	m_CplsFilled=false; m_iCplsFilledAfter=0; m_QCplsFilled=false;
}*/
CHEPEvtLite::~CHEPEvtLite() {
	if(m_ChOrderBuf!=NULL){ delete[] m_ChOrderBuf; m_ChOrderBuf=NULL; }
}

#define EVTLITE_CNTRL_NUMBER_TXT	255
#define EVTLITE_CNTRL_NUMBER_BIN	254
#define EVTLITE_CNTRL_CACHE_SYM		'C'
bool CHEPEvtLite::LoadEvent(std::ifstream* ifs) {
	if(ifs==NULL)return false;
	if(!ifs->good())return false;
	int cntrlN=-1;
	*ifs>>cntrlN;
	bool bBinary;
	if(cntrlN==EVTLITE_CNTRL_NUMBER_BIN) { bBinary=true; }
	else if(cntrlN==EVTLITE_CNTRL_NUMBER_TXT) { bBinary=false; }
	else return false;
	if(bBinary) {
		//int smOK=_setmode(_fileno(*ifs),_O_BINARY); if(smOK==-1) perror( "Cannot set mode" ); else printf( "'stdin' successfully changed to binary mode\n" );
	}
	bool bLoadedOK=true; int i; char c;
	*ifs>>m_nParticles;
	//
	c=ifs->peek();
	for(;c==' ';){*ifs>>c;} //skip spaces
	if((c!=0x0d)&&(c!=0x0a)) {
		//read additional values in first line
		if(bBinary) {
			ifs->read((char*)&wgt, sizeof(wgt)); ifs->read((char*)&wgtErr, sizeof(wgtErr));
			ifs->read((char*)&scale,sizeof(scale));
			ifs->read((char*)&a_QED,sizeof(a_QED)); ifs->read((char*)&a_QCD,sizeof(a_QCD));
		}else{
			*ifs>>wgt>>wgtErr>>scale>>a_QED>>a_QCD;
		}
	}
	for(i=0;i<m_nParticles; i++) { if(!m_Particles[i].Load(ifs,bBinary)){bLoadedOK=false;break;} }

	//Skip possible spaces
	if(bLoadedOK){ for(int iSp=0;(iSp<10);iSp++){ char c; *ifs>>c; if((c==' ')||(c==0x0d)||(c==0x0a)){continue;} ifs->putback(c); break; } }
	//Load couplings cache
	if(bLoadedOK && (ifs->peek() == EVTLITE_CNTRL_CACHE_SYM)) {
		*ifs>>c; //skip symbol
		if(bBinary){
			//Skip spaces and '{'-bracket
			for(int iSp=0;(iSp<10);iSp++){ *ifs>>c; if(c!=' ')break; }
			if(c=='{'){
				tReal val; ifs->read((char*)&val,sizeof(val)); m_cpls.m_Q2abs=val; m_cpls.m_vecQ.Load(ifs,bBinary);
				ifs->read((char*)&c,sizeof(c)); m_cpls.m_CplsFilled=(c>0); m_cpls.m_iCplsFilledAfter=((c>0)?(c-1):0);
				ifs->read((char*)&c,sizeof(c)); m_cpls.m_QCplsFilled=(c>0);
				if(m_cpls.m_CplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ for(int j=m_cpls.m_iCplsFilledAfter;j<m_nParticles; j++){ ifs->read((char*)&val,sizeof(val)); m_cpls.m_Cpls[i][j]=val; } } }
				if(m_cpls.m_QCplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ ifs->read((char*)&val,sizeof(val)); m_cpls.m_QCpls[i]=val; } }
				for(i=0;i<m_nParticles; i++){ ifs->read((char*)&val,sizeof(val)); m_cpls.m_UserCpls[i]=val; }
				*ifs>>c; if(c!='}'){ bLoadedOK=false; } //Check '}'-bracket
			}else{ bLoadedOK=false; }
		}else{
			*ifs>>m_cpls.m_Q2abs; m_cpls.m_vecQ.Load(ifs,bBinary); tReal val;
			*ifs>>c; m_cpls.m_CplsFilled=(c>0); m_cpls.m_iCplsFilledAfter=((c>0)?(c-1):0);
			*ifs>>c; m_cpls.m_QCplsFilled=(c>0);
			if(m_cpls.m_CplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ for(int j=m_cpls.m_iCplsFilledAfter;j<m_nParticles; j++){ *ifs>>val; m_cpls.m_Cpls[i][j]=val; } } }
			if(m_cpls.m_QCplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ *ifs>>val; m_cpls.m_QCpls[i]=val; } }
			for(i=0;i<m_nParticles; i++){ *ifs>>val; m_cpls.m_UserCpls[i]=val; }
		}
	}

	//Check control number
	if(bLoadedOK){ int cntrlN2; *ifs>>cntrlN2; if(cntrlN2!=cntrlN){bLoadedOK=false;} }
	if((!bLoadedOK) && bBinary) { const char* mess="ERROR loading event. Possible reason: file is open in non-binary mode."; std::cerr<<mess<<std::endl; std::cout<<mess<<std::endl; }
	return bLoadedOK;
}

bool CHEPEvtLite::SaveEvent(std::ofstream* ofs, bool _binary, bool bSaveCouplingsCache) {
	if(ofs==NULL)return false;
	int cntrlN=((_binary)?EVTLITE_CNTRL_NUMBER_BIN:EVTLITE_CNTRL_NUMBER_TXT);
	*ofs<<cntrlN<<" "<<m_nParticles<<" ";
	*ofs<<idProc<<" ";
	if(_binary){
		ofs->write((char*)&wgt,sizeof(wgt)); ofs->write((char*)&wgtErr,sizeof(wgtErr));
		ofs->write((char*)&scale,sizeof(scale));
		ofs->write((char*)&a_QED,sizeof(a_QED)); ofs->write((char*)&a_QCD,sizeof(a_QCD));
	}else{
		*ofs<<wgt<<" "<<wgtErr<<" "<<scale<<" "<<a_QED<<" "<<a_QCD;
	}
	*ofs<<std::endl;//if(!_binary)*ofs<<std::endl;
	int i;
	for(i=0;i<m_nParticles; i++) {
		m_Particles[i].Save(ofs,_binary);
		*ofs<<std::endl; //if(!_binary)*ofs<<std::endl;
	}
	//Save couplings cache
	if(bSaveCouplingsCache){
		*ofs<<EVTLITE_CNTRL_CACHE_SYM;
		if (_binary) {
			*ofs<<'{'; //start of binary block
			tReal val; val=m_cpls.m_Q2abs; ofs->write((char*)&val,sizeof(val)); m_cpls.m_vecQ.Save(ofs,_binary);
			if(m_cpls.m_iCplsFilledAfter<0)m_cpls.m_iCplsFilledAfter=0; //fix internal error
			char cTmp=0; if(m_cpls.m_CplsFilled){ cTmp=1+(char)m_cpls.m_iCplsFilledAfter; } ofs->write((char*)&cTmp,sizeof(cTmp));
			cTmp=(m_cpls.m_QCplsFilled?1:0); ofs->write((char*)&cTmp,sizeof(cTmp));
			if(m_cpls.m_CplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ for(int j=m_cpls.m_iCplsFilledAfter;j<m_nParticles; j++){ val=m_cpls.m_Cpls[i][j]; ofs->write((char*)&val,sizeof(val)); } } }
			if(m_cpls.m_QCplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ val=m_cpls.m_QCpls[i]; ofs->write((char*)&val,sizeof(val)); } }
			for(i=0;i<m_nParticles; i++){ val=m_cpls.m_UserCpls[i]; ofs->write((char*)&val,sizeof(val)); }
			*ofs<<'}'; //end of binary block
		} else {
			*ofs<<" "<<m_cpls.m_Q2abs<<" "; m_cpls.m_vecQ.Save(ofs,_binary);
			*ofs<<" "<<(int)(m_cpls.m_CplsFilled?1:0)<<" ";
			if(m_cpls.m_iCplsFilledAfter<0)m_cpls.m_iCplsFilledAfter=0; //fix internal error
			long cTmp=0; if(m_cpls.m_CplsFilled){ cTmp=1+m_cpls.m_iCplsFilledAfter; } *ofs<<cTmp;
			if(m_cpls.m_CplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ for(int j=m_cpls.m_iCplsFilledAfter;j<m_nParticles; j++){ *ofs<<" "<<m_cpls.m_Cpls[i][j]; } } }
			if(m_cpls.m_QCplsFilled){ for(i=m_cpls.m_iCplsFilledAfter;i<m_nParticles; i++){ *ofs<<" "<<m_cpls.m_QCpls[i]; } }
			for(i=0;i<m_nParticles; i++){ *ofs<<" "<<m_cpls.m_UserCpls[i]; }
		}
	}
	*ofs<<std::endl;
	*ofs<<cntrlN; *ofs<<std::endl; //if(!_binary)*ofs<<std::endl;
	return true;
}

CHEPParticleLite* CHEPEvtLite::GetParticle(int i) {
	return (((i<0)||(i>=m_nParticles)) ? NULL : (&(m_Particles[i])));
}
int CHEPEvtLite::GetFirstInitialId() { // 'id(first=0)' of first particle with status == INITIAL
	for(int i=0; i<m_nParticles; i++) { if(m_Particles[i].m_Status==CHEPParticleLite::STATUS_INITIAL) return i; }
	return -1;
}
int CHEPEvtLite::GetFirstFinalId() { // 'id(first=0)' of first particle with status == FINAL
	for(int i=0; i<m_nParticles; i++) { if(m_Particles[i].m_Status==CHEPParticleLite::STATUS_FINAL) return i; }
	return -1;
}
tReal CHEPEvtLite::GetCpl(int i, int j) const { // uses m_Cpls if available
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	if ((i<0) || (j<0) || (i>=nMax) || (j>=nMax)) return -1.;
	tReal res=-1.;
	if ((m_cpls.m_CplsFilled) && (i!=j) && (i>=m_cpls.m_iCplsFilledAfter) && (j>=m_cpls.m_iCplsFilledAfter)) {
		//TODO: why don't we keep squares (i==j)?	//asdfg
		res=((i<j)?(m_cpls.m_Cpls[i][j]):(m_cpls.m_Cpls[j][i]));
	} else {
		res=((i==j)? (m_Particles[i].m_Mom.GetSqr()) : (m_Particles[i].m_Mom.GetCpl(m_Particles[j].m_Mom)));
		//TODO: cache couplings automatically?
	}
	return res;
}
/*
*	GetCplSpecial: used by 'CVMPhaseSpaceGen' to update couplings cache (Don't use elsewhere!).
*		- ignores 'm_CplsFilled' flag
*		- recalcs and saves coupling if not filles previously (<0)
*/
tReal CHEPEvtLite::GetCplSpecial(int i, int j) {
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	if((i<0) || (j<0) || (i>=nMax) || (j>=nMax)) return -1.;
	if(i==j) return m_Particles[i].m_Mom.GetSqr();
	tReal res=((i<j)?(m_cpls.m_Cpls[i][j]):(m_cpls.m_Cpls[j][i]));
	if(res<0.) { res=m_Particles[i].m_Mom.GetCpl(m_Particles[j].m_Mom); if(i<j){ m_cpls.m_Cpls[i][j]=res; }else{ m_cpls.m_Cpls[j][i]=res; } }
	return res;
}
void CHEPEvtLite::SetCpl(int i, int j, tReal val) { // fills m_Cpls
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	if ((i<0) || (j<0) || (i>=nMax) || (j>=nMax)) return;
	if(i<j){ m_cpls.m_Cpls[i][j]=val; }else{ m_cpls.m_Cpls[j][i]=val; }
}
tReal CHEPEvtLite::GetQCpl(int i) const { // uses m_QCpls if available
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	if ((i<0) || (i>=nMax)) return 0.;
	return ((m_cpls.m_QCplsFilled)?(m_cpls.m_QCpls[i]):(m_cpls.m_vecQ.GetCpl(m_Particles[i].m_Mom)));
}
void CHEPEvtLite::SetQCpl(int i, tReal val) { // fills m_QCpls
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	if((i<0)||(i>=nMax))return; m_cpls.m_QCpls[i]=val;
}
void CHEPEvtLite::FillCouplingsCache(bool bKeepFilled, bool bFillQCpls, int iFillAfter) { // fills m_Cpls
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	if(iFillAfter>=nMax) { m_cpls.m_CplsFilled=false; m_cpls.m_iCplsFilledAfter=iFillAfter; return; }

	for(int i=iFillAfter; i<nMax; i++){
		for(int j=i+1; j<nMax; j++){
			if ((bKeepFilled) && (m_cpls.m_Cpls[i][j]>=0.)) continue;
			m_cpls.m_Cpls[i][j]=m_Particles[i].m_Mom.GetCpl(m_Particles[j].m_Mom);
		}
	}
	m_cpls.m_CplsFilled=true; m_cpls.m_iCplsFilledAfter=iFillAfter;

	if(bFillQCpls){ // && (m_vecQ.GetSqr()<0.)
		for(int j=iFillAfter; j<nMax; j++){
			if((bKeepFilled)&&(m_cpls.m_QCpls[j]!=0.))continue; m_cpls.m_QCpls[j]=m_cpls.m_vecQ.GetCpl(m_Particles[j].m_Mom);
		}
		m_cpls.m_QCplsFilled=true;
	}
}
//tReal CHEPEvtLite::GetQ2Abs() const {
//	tReal res = m_cpls.m_Q2abs;
//	//if(res<0.){ /*recalc 'm_cpls.m_Q2abs'*/ }
//	return res;
//}
//const CVM4Vec* CHEPEvtLite::GetQVec() const {
//	//if(m_cpls.m_vecQ.GetSqr() >= 0.){ /*recalc 'm_cpls.m_vecQ'*/}
//	return &(m_cpls.m_vecQ);
//}
void CHEPEvtLite::ClearCouplingsCache() {
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	for(int i=0;i<nMax;i++){ for(int j=i+1;j<nMax;j++){ m_cpls.m_Cpls[i][j]=-1; } m_cpls.m_QCpls[i]=0.; }
	m_cpls.m_vecQ.Set(0,0,0,0);
	m_cpls.m_Q2abs=-1.;
	m_cpls.m_CplsFilled=false; m_cpls.m_iCplsFilledAfter=0; m_cpls.m_QCplsFilled=false;
}

bool CHEPEvtLite::IsEventOK() { return (m_nParticles>=0); }

//order: {0<-Pos,1<-Pos,...}
//Warning: function doesn't check 'order'!
//TODO #1: leave some couplings if available (e.g. Q^2)
bool CHEPEvtLite::ChangeOrder(CHEPParticleOrderLite* order) {
	if(order==NULL)return false;
	//allocate temporary buffer (if not done yet)
	if(m_ChOrderBuf==NULL){
		m_ChOrderBuf = new CHEPParticleLite[m_nParticles]; m_szChOrderBuf=m_nParticles;
	}else{
		if(m_szChOrderBuf<m_nParticles) {
			delete[] m_ChOrderBuf; m_ChOrderBuf=NULL;
			m_ChOrderBuf = new CHEPParticleLite[m_nParticles]; m_szChOrderBuf=m_nParticles;
		}
	}
	if(m_ChOrderBuf==NULL)return false;

	ClearCouplingsCache();

	memcpy(m_ChOrderBuf,m_Particles,m_nParticles*sizeof(CHEPParticleLite));
	bool bRes=true;
	for(int i=0; i<m_nParticles; i++) {
		int posOld=order->Get(i); if((posOld<0)||(posOld>=m_nParticles)) { bRes=false; break; }
		memcpy(&m_Particles[i],&m_ChOrderBuf[posOld],sizeof(CHEPParticleLite));
	}
	return bRes;
}

//e.g. to insert nuclon before parton line
//clears couplings
void CHEPEvtLite::InsertParticle(int pos, const CHEPParticleLite& p) {
	ClearCouplingsCache();
	if(pos<0)pos=0;
	if(m_nParticles<EVTL_MAX_PART){
		if(pos<m_nParticles) {				//insert
			memcpy(&m_Particles[pos+1],&m_Particles[pos],(m_nParticles-pos)*sizeof(CHEPParticleLite));
			memcpy(&m_Particles[pos],&p,sizeof(CHEPParticleLite));
		} else {							//add
			memcpy(&m_Particles[m_nParticles],&p,sizeof(CHEPParticleLite));
		}
		m_nParticles++;
	}else{
		std::cerr<<"CHEPEvtLite ERROR: too many particles in event. The last is missed"<<std::endl;
		if(pos<m_nParticles) {				//insert & truncate last (or replace last)
			if(pos<(m_nParticles-1)) {
				memcpy(&m_Particles[pos+1],&m_Particles[pos],(m_nParticles-1-pos)*sizeof(CHEPParticleLite));
			}
			memcpy(&m_Particles[pos],&p,sizeof(CHEPParticleLite));
		}
	}
}

void CHEPEvtLite::LogEvent(std::ostream& os) {
	CVM4Vec pSum(0,0,0,0);			// initials-finals, if all initials are before finals
	bool bDrawSum=true;
	CVM4Vec pSumFinals(0,0,0,0);	// finals only
	int i=10; int nWidth=1; for(;m_nParticles>=i;){ nWidth++; i*=10; }
	os<<"================================ Event listing ================================"<<std::endl;
	for(i=1;i<nWidth;i++){os<<" ";} os<<"N   KF KS IDM"; for(i=1;i<nWidth;i++){os<<"  ";} os<<" p{E,Px,Py,Pz,m}"<<std::endl;
	bool bFinalSpotted=false;
	for(int i=0; i<m_nParticles; i++) {
		CHEPParticleLite* part=&m_Particles[i];
		os.width(nWidth); os<<(i+1); os.width();
		os<<" "; os.width(4); os<<part->m_PID; os.width();
		os<<" "; os.width(2); os<<part->m_Status; os.width();
		os<<" "; os.width(nWidth); os<<(part->m_Mother1_f0+1)<<" "<<(part->m_Mother2_f0+1); os.width();
		os<<" "<<part->m_Mom<<"  M="<<part->m_Mom.GetMassSigned()<<std::endl;
		//
		bool bIsFinal=(part->m_Status == CHEPParticleLite::STATUS_FINAL); if(bIsFinal)bFinalSpotted=true;
		bool bIsInitial=(part->m_Status == CHEPParticleLite::STATUS_INITIAL); if(bIsInitial && bFinalSpotted)bDrawSum=false;
		if(bDrawSum) { if(bIsInitial){ pSum+=part->m_Mom; }else if (bIsFinal){ pSum-=part->m_Mom; } }
		if(bIsFinal) { pSumFinals+=part->m_Mom; }
	}
	os<<std::endl;
	if(bDrawSum){
		for(i=1;i<nWidth;i++){os<<"   ";}
		os<<"      In-Fin:"<<" "<<pSum<<"  M="<<pSum.GetMassSigned()<<std::endl;
	}else{
		for(i=1;i<nWidth;i++){os<<"   ";}
		os<<"      Finals:"<<" "<<pSumFinals<<"  M="<<pSumFinals.GetMassSigned()<<std::endl;
	}
	os<<std::endl;
	if(m_cpls.m_CplsFilled){
		for(i=1;i<nWidth;i++){os<<"   ";}
		os<<"           Q: "<<m_cpls.m_vecQ<<" Q2="<<m_cpls.m_Q2abs<<" (Calc: "<<-m_cpls.m_vecQ.GetSqr()<<")"<<std::endl;
	}else{
		os<<"No couplings pre-calculated"<<std::endl;
	}
	os<<"==============================================================================="<<std::endl;
}
void CHEPEvtLite::LogCouplings(std::ostream& os) {
	int nMax=m_nParticles; if(nMax>EVTL_MAX_CPL_PART)nMax=EVTL_MAX_CPL_PART;
	os<<"+------------- Event couplings --------------+"<<std::endl;
	int i;
	int idInitials[]={-1,-1};
	for(i=0; i<m_nParticles; i++) {
		if(m_Particles[i].m_Status==CHEPParticleLite::STATUS_DOC)continue;
		if(m_Particles[i].m_Status==CHEPParticleLite::STATUS_INITIAL){ //save initial particle positions
			if(idInitials[0]<0){ idInitials[0]=i; }else if(idInitials[1]<0){ idInitials[1]=i; }
		}
		for(int j=i+1;j<m_nParticles;j++){
			if(m_Particles[j].m_Status==CHEPParticleLite::STATUS_DOC)continue;
			tReal cpl=m_Particles[i].m_Mom.GetCpl(m_Particles[j].m_Mom); os<<"["<<i<<","<<j<<"]: "<<cpl<<", cache: ";
			tReal cch=(((i<nMax)&&(j<nMax))?m_cpls.m_Cpls[i][j]:-1.); if(cch>=0.){os<<cch<<" ( diff: "<<fabs((cch>0.)?(cpl/cch-1.):(cpl-cch))<<" )"<<std::endl;}else{os<<"no"<<std::endl;}
		}
	}
	if(m_cpls.m_QCplsFilled){
		for(i=0; i<m_nParticles; i++) {
			if(m_Particles[i].m_Status==CHEPParticleLite::STATUS_DOC)continue;
			tReal cpl=m_cpls.m_vecQ.GetCpl(m_Particles[i].m_Mom); os<<"[q,"<<i<<"]: "<<cpl<<", cache: ";
			tReal cch=((i<nMax)?m_cpls.m_QCpls[i]:0.); if(cch!=0.){os<<cch<<" ( diff: "<<fabs((cch!=0.)?(cpl/cch-1.):(cpl-cch))<<" )"<<std::endl;}else{os<<"no"<<std::endl;}
		}

		tReal cpl=m_cpls.m_vecQ.GetSqr();
		os<<"[q,q]: "<<cpl<<", cache: "<<-m_cpls.m_Q2abs<<", (diff:"<<fabs((m_cpls.m_Q2abs>0.)?(-cpl/m_cpls.m_Q2abs-1.):(-cpl-m_cpls.m_Q2abs))<<")"<<std::endl;
		if(idInitials[1]>=0){
			CVM4Vec qTmp=m_Particles[m_nParticles-1].m_Mom; qTmp-=m_Particles[idInitials[1]].m_Mom; cpl=qTmp.GetSqr();
			os<<"[q,q]: via p2-p1: "<<cpl<<", cache: "<<-m_cpls.m_Q2abs<<", (diff:"<<fabs((m_cpls.m_Q2abs>0.)?(-cpl/m_cpls.m_Q2abs-1.):(-cpl-m_cpls.m_Q2abs))<<")"<<std::endl;
		}else{
			os<<"Can't re-calculate q-vector: initial particles not marked"<<std::endl;
		}
	}
	os<<"+--------------------------------------------+"<<std::endl;
}

///////////////////////////////////////////// CHEPEvtLiteView
#ifdef Decl_CHEPEvtLiteView
const int CHEPEvtLiteView::MODE_YZ	= 1;
const int CHEPEvtLiteView::MODE_XY	= 2;
const int CHEPEvtLiteView::MODE_TZ	= 3;
const int CHEPEvtLiteView::MODE_XZ	= 4;
// Fast entries
void CHEPEvtLiteView::DrawEvent(CHEPEvtLite* evt, const char* fn, int mode, int iMarked) { /* iMarked=0 -> all finals */
	CHEPEvtLiteView view(fn,mode); view.AddEventToView(evt,iMarked); view.SaveView();
}
void CHEPEvtLiteView::DrawEvent(CHEPEvtLite* evt, CVMOutput* out, int mode, int iMarked) { /* iMarked=0 -> all finals */
	CHEPEvtLiteView view(out,mode); view.AddEventToView(evt,iMarked); view.SaveView();
}
//
CHEPEvtLiteView::CHEPEvtLiteView(const char* fn, int mode) {
	if(fn==NULL)fn="Event.eps";
	m_Mode=mode;
	CVMOutputEPS* outEPS=new CVMOutputEPS(fn);
	if(m_Mode==MODE_XY){ outEPS->SetBoundingBox(0,300,0,300); }else{ outEPS->SetBoundingBox(0,500,0,300); }
	m_Out=outEPS; m_bExternalOutput=false; m_bOpen=false;
	Init();
}
CHEPEvtLiteView::CHEPEvtLiteView(CVMOutput* out, int mode) {
	m_Out=out; m_bExternalOutput=true; m_bOpen=false;
	m_Mode=mode; Init();
}
void CHEPEvtLiteView::Init() { // Used by constructors only
	if(m_Out==NULL) {
		return;
	}
	if((m_Mode<=0)||(m_Mode>4))m_Mode=MODE_YZ;
	m_bBeamAxisZ=((m_Mode==MODE_XZ) || (m_Mode==MODE_YZ) || (m_Mode==MODE_TZ));
	if(m_Mode==MODE_XY)m_Out->SetMargins(0,1,0,0.9);
	m_nEvtsDrawed=0;
	m_xValMax=0; m_yValMax=0;
	int i;
	for(i=0; i<40; i++) { m_clrTracksByPID[i]=CVMOutput::clrBlack; m_widthTracksByPID[i]=0.25; m_bDashedTracksByPID[i]=false; }
	m_clrTracksByPID[11]=CVMOutput::clrGreen; m_clrTracksByPID[13]=CVMOutput::clrGreen; m_clrTracksByPID[15]=CVMOutput::clrGreen;
	m_clrTracksByPID[12]=CVMOutput::clrGrey; m_clrTracksByPID[14]=CVMOutput::clrGrey; m_clrTracksByPID[16]=CVMOutput::clrGrey; //neutrinos
	m_clrTracksByPID[22]=CVMOutput::clrBlue; m_clrTracksByPID[23]=CVMOutput::clrBlue; m_clrTracksByPID[24]=CVMOutput::clrBlue;
	//CVMRGB clrQuark(1,0,0);
	m_clrTracksByPID[0]=CVMOutput::clrRed; m_widthTracksByPID[0]=1;
	m_bDashedTracksByPID[12]=true; m_bDashedTracksByPID[14]=true; m_bDashedTracksByPID[16]=true; //nuetrinos
	m_clrLabel=CVMOutput::clrBlack; m_nLabelsPrinted=0;
	m_clrBackground.Set(0.9f,1.f,1.f); m_bDrawBox=true;
	m_InnerRingSize=0.f; if(!m_bBeamAxisZ)m_InnerRingSize=0.05f;
	m_bEquXYScale=false;
	m_clrTracksByMark[0]=CVMOutput::clrBlack; m_clrTracksByMark[1]=CVMOutput::clrRed; m_clrTracksByMark[2].Set(0,0.6f,0);
	m_clrTracksByMark[3]=CVMOutput::clrBlue; m_clrTracksByMark[4]=CVMOutput::clrYellow; m_clrTracksByMark[5]=CVMOutput::clrMagenta;
	m_clrTracksByMark[6]=CVMOutput::clrCyan; m_clrTracksByMark[7]=CVMOutput::clrOrange; m_clrTracksByMark[8]=CVMOutput::clrGrey;
	m_clrTracksByMark[9]=CVMOutput::clrBrown; m_clrTracksByMark[10]=CVMOutput::clrYellowGreen; m_clrTracksByMark[12]=CVMOutput::clrLawnGreen;
	for(i=12; i<40; i++) { m_clrTracksByMark[i]=m_clrTracksByMark[i-11]; }
	m_bColorTrackByPID=true; // otherwise 'color-by-mark'
	m_bAllowOverMargins=false;
	m_bSkipNegativeMarks=false; //(false): don't draw particles with 'mark<0'
	m_bDrawFinalsOnly=true;
}
CHEPEvtLiteView::~CHEPEvtLiteView() {
	if(m_Out!=NULL){ if(m_bOpen)m_Out->Finalize(); if(!m_bExternalOutput)delete m_Out; m_Out=NULL; }
}
void CHEPEvtLiteView::AddEventToView(CHEPEvtLite* evt, int iMarked, const char* label) { // SaveView() expected afterwards.
	if(m_Out==NULL)return;
	if(!m_bOpen) {
		m_Out->Start(); if(!m_Out->IsOpen()){ std::cerr<<"ERROR: Can't create ouptut file "<<std::endl; if(!m_bExternalOutput)delete m_Out; m_Out=NULL; return; }
		m_bOpen=true;
	}

	//Draw event
	if ((evt==NULL) || (m_Out==NULL))return; if (!m_Out->IsOpen()) return; if(evt->m_nParticles<=0) return;
	double xMin=0; double xMax=1; double yMin=0; double yMax=1; m_Out->GetMargins(&xMin,&xMax,&yMin,&yMax);
	if ((xMin>=xMax) || (yMin>=yMax)) return;
	int i,l; char buf[100]; double lw=m_Out->GetLetterWidth(); double lh=m_Out->GetLetterHeight();

	double xC=(xMin+xMax)/2.; double xA=(xC-xMin);
	double yC=(yMin+yMax)/2.; double yA=(yC-yMin);
	double Amin=((xA>=yA)?yA:xA); if(m_bEquXYScale){ xA=Amin; yA=Amin; }

	if (m_nEvtsDrawed==0) { // fill background
		m_Out->FillBackground(m_clrBackground);
		if(m_bDrawBox){ m_Out->SetColor(CVMOutput::clrBlack); m_Out->SetLineWidth(0.25); m_Out->Rect(xMin,yMin,xMax,yMax); }
		if(m_InnerRingSize > 0.f) { m_Out->SetColor(CVMOutput::clrBlack); m_Out->Circle(xC,yC,Amin*m_InnerRingSize); }
	}

	if (m_nEvtsDrawed==0) {
		if ((m_bEquXYScale) && ((m_xValMax > 0.) || (m_yValMax>0.))) {
			if(m_yValMax<=0.)m_yValMax=m_xValMax; if(m_xValMax<=0.)m_xValMax=m_yValMax;
		}
	}
	if ((m_nEvtsDrawed==0) && ((m_xValMax<=0.) || (m_yValMax<=0.))) { // find max if not pre-defined
		double _xValMax=0; double _yValMax=0;
		double fPMax=0; int nPartUsed=0;
		for(i=0; i<evt->m_nParticles; i++) {
			CHEPParticleLite* part=&evt->m_Particles[i];
			if (m_bSkipNegativeMarks && (part->m_Mark<0))continue;
			if (iMarked!=0) { if(part->m_Mark!=iMarked)continue; }
			//else{ if(m_bDrawFinalsOnly && (part->m_Status!=CHEPParticleLite::STATUS_FINAL))continue; }
			if(m_bDrawFinalsOnly && (part->m_Status!=CHEPParticleLite::STATUS_FINAL))continue;
			double pAbs=DBLtR(part->m_Mom.GetPAbs()); if(pAbs>fPMax)fPMax=pAbs;
			double xVal=0; double yVal=0;
			switch(m_Mode) {
			case MODE_YZ: xVal=DBLtR(part->m_Mom.m_Pz); yVal=DBLtR(part->m_Mom.m_Py); break;
			case MODE_XY: xVal=DBLtR(part->m_Mom.m_Px); yVal=DBLtR(part->m_Mom.m_Py); break;
			case MODE_TZ: xVal=DBLtR(part->m_Mom.m_Pz); yVal=DBLtR(part->m_Pt); break;
			case MODE_XZ: xVal=DBLtR(part->m_Mom.m_Pz); yVal=DBLtR(part->m_Mom.m_Px); break;
			}
			if(_xValMax<fabs(xVal))_xValMax=fabs(xVal); if(_yValMax<fabs(yVal))_yValMax=fabs(yVal);
			nPartUsed++;
		}
		if(m_xValMax<=0)m_xValMax=_xValMax; if(m_yValMax<=0)m_yValMax=_yValMax;
		if ((m_bEquXYScale) && (m_xValMax != m_yValMax)) {
			if(m_yValMax<m_xValMax){ m_yValMax=m_xValMax; }else{ m_xValMax=m_yValMax; }
		}
		sprintf(buf,"Used: %d/%d",nPartUsed,evt->m_nParticles); m_Out->TextOut(xMin+(xMax-xMin)/5,yMin+(yMax-yMin)/20,buf);
		sprintf(buf,"Pmax=%.4g[GeV]",(double)fPMax); m_Out->TextOut(xMin+(xMax-xMin)/5,yMin+(yMax-yMin)*2/20,buf);
	}
	if (m_nEvtsDrawed==0) {
		CVMRGB clrBeam(0.5,0.5,0.5);
		if(m_bBeamAxisZ){ m_Out->SetColor(clrBeam); m_Out->SetLineWidth(1); m_Out->MoveTo(xMin,yC); m_Out->LineTo(xMax,yC); }
		l=sprintf(buf,"%.1fGeV",(double)m_yValMax); m_Out->TextOut(xC+lw*l/2,yMax-lh/2,buf);
		l=sprintf(buf,"%.1fGeV",(double)m_xValMax); m_Out->TextOut(xMax-lw*l/2,yC-lh/2,buf);
	}
	if ((m_yValMax<=0.) || (m_xValMax<=0.)) return;

	CVMRGB clrTrack(0,0,0); float wTrack=0.25;
	CVMRGB clrCurr(-1,-1,-1); float wCurr=-1; bool bDashedCurr=false;
	//m_Out->SetColor(clrTrack);
	for(i=0; i<evt->m_nParticles; i++) {
		CHEPParticleLite* part=&evt->m_Particles[i];
		if (m_bSkipNegativeMarks && (part->m_Mark<0))continue;
		if (iMarked!=0) { if(part->m_Mark!=iMarked)continue; }
		//else{ if(m_bDrawFinalsOnly && (part->m_Status!=CHEPParticleLite::STATUS_FINAL))continue; }
		if(m_bDrawFinalsOnly && (part->m_Status!=CHEPParticleLite::STATUS_FINAL))continue;
		double xVal=0; double yVal=0;
		switch(m_Mode) {
		case MODE_YZ: xVal=DBLtR(part->m_Mom.m_Pz); yVal=DBLtR(part->m_Mom.m_Py); break;
		case MODE_XY: xVal=DBLtR(part->m_Mom.m_Px); yVal=DBLtR(part->m_Mom.m_Py); break;
		case MODE_TZ: xVal=DBLtR(part->m_Mom.m_Pz); yVal=DBLtR(part->m_Pt); break;
		case MODE_XZ: xVal=DBLtR(part->m_Mom.m_Pz); yVal=DBLtR(part->m_Mom.m_Px); break;
		}

		double xValNorm=xVal/m_xValMax; double yValNorm=yVal/m_yValMax;
		double x1=xC; double y1=yC; double x2=xC+xA*xValNorm; double y2=yC+yA*yValNorm;
		if(m_InnerRingSize>0.f) {
			double lValNorm=sqrt(xValNorm*xValNorm+yValNorm*yValNorm);
			if(lValNorm <= 0.)continue; //can't draw particle
			double cosA=xValNorm/lValNorm; double sinA=yValNorm/lValNorm;
			x1+=Amin*m_InnerRingSize*cosA; y1+=Amin*m_InnerRingSize*sinA;
			// Rescale length
			double lVal=sqrt((xA*xValNorm)*(xA*xValNorm) + (yA*yValNorm)*(yA*yValNorm));
			double lValNew=Amin*m_InnerRingSize+lVal*(1.-m_InnerRingSize);
			x2=xC+lValNew*cosA; y2=yC+lValNew*sinA;
		}

		if(!m_bAllowOverMargins) {
			if(y2>yMax){ x2=xC+(x2-xC)*(yMax-yC)/(y2-yC);y2=yMax; }else if(y2<yMin){ x2=xC+(x2-xC)*(yC-yMin)/(yC-y2);y2=yMin; }
			if(x2>xMax){ y2=yC+(y2-yC)*(xMax-xC)/(x2-xC);x2=xMax; }else if(x2<xMin){ y2=yC+(y2-yC)*(xC-xMin)/(xC-x2);x2=xMin; }
		}

		int idPAbs=abs(part->m_PID);
		float w=((idPAbs<40)?(m_widthTracksByPID[idPAbs]):wTrack); if(wCurr!=w){ wCurr=w; m_Out->SetLineWidth(wCurr); }
		CVMRGB* clr;
		if(m_bColorTrackByPID) {
			clr=((idPAbs<40)?(&m_clrTracksByPID[idPAbs]):&clrTrack);
		} else {
			int mrk=part->m_Mark; clr=(((mrk>=0)&&(mrk<40))?(&m_clrTracksByMark[mrk]):&clrTrack);
		}
		if(clrCurr!=*clr){ clrCurr=*clr; m_Out->SetColor(clrCurr); }
		bool bDashed=((idPAbs<40)?(m_bDashedTracksByPID[idPAbs]):false);
		if((bDashedCurr && !bDashed)||(!bDashedCurr && bDashed)) { bDashedCurr=bDashed; if(bDashedCurr){m_Out->SetLineStyle(CVMOutput::LINE_DASH);}else{m_Out->SetLineStyle(CVMOutput::LINE_SOLID);} }

		m_Out->MoveTo(x1,y1); m_Out->LineTo(x2,y2);
		//m_Out->MoveTo(xC,yC); m_Out->LineTo(xC+xA*xVal/m_xValMax,yC+yA*yVal/m_yValMax);
	}
	if(bDashedCurr){ m_Out->SetLineStyle(CVMOutput::LINE_SOLID); bDashedCurr=false; }

	// Draw label
	if(label!=NULL) {
		double yHigh=yMax-lh*0.5; int l=0; for(l=0;l<100;l++){ if(*(label+l)=='\0')break; } // inline length calc
		m_Out->SetColor(m_clrLabel); m_Out->TextOut(xMax-lw-lw*l/2,yHigh-lh*(1.3*m_nLabelsPrinted),label);
		m_nLabelsPrinted++;
	}

	m_Out->SetColor(clrTrack); // restore color
	//m_Out->BlockEnd();
	m_nEvtsDrawed++;
}
void CHEPEvtLiteView::SaveView() { // Saves view after several AddEventToView() calls.
	if((m_Out!=NULL)&&(m_bOpen)){ m_Out->Finalize(); m_bOpen=false; }
}
#endif //Decl_CHEPEvtLiteView

///////////////////////////////////////////////////////////// CHEPEvtLiteLib
bool CHEPEvtLiteLib::ConvertEvtFile(const char* fnRead, const char* fnWrite, bool bToBinary) {
	if((fnRead==NULL) || (fnWrite==NULL))return false;
	std::ifstream ifs; ifs.open(fnRead,std::ios::in|std::ios::binary); if(!ifs.is_open()) { std::cout<<"Error: can't read events file: "<<fnRead<<std::endl; return false; }
	std::ofstream ofs(fnWrite,(bToBinary?(std::ios::out|std::ios::binary):(std::ios::out))); if(!ofs){ std::cout<<"Can't open output file "<<fnWrite<<std::endl; ifs.close(); return false; }
	CHEPEvtLite* evt = new CHEPEvtLite(); //Safer than local variable (0.3Mb)
	int nLogeventsEvery=1000; int nLoggedLast=-1; long n; bool bLoadOK=true; bool bSaveOK=true;
	for(n=0;; n++) { // Loop over all events
		if (n/nLogeventsEvery>nLoggedLast) { nLoggedLast=n/nLogeventsEvery; std::cout<<"Event #"<<n<<std::endl; }
		if (!evt->LoadEvent(&ifs)){ if(!ifs.eof())bLoadOK=false; break; }
		if (!evt->SaveEvent(&ofs,bToBinary)){ bSaveOK=false; break; }
	}
	if(!bLoadOK)std::cout<<"Failed to load event #"<<n<<" (ifs.good()="<<ifs.good()<<")"<<std::endl; 
	if(!bSaveOK)std::cout<<"Failed to save event #"<<n<<" (ofs.good()="<<ofs.good()<<")"<<std::endl; 
	std::cout<<"Total "<<n<<" events converted"<<std::endl;
	ofs.close(); ifs.close(); delete evt;
	return ((bLoadOK)&&(bSaveOK));
}
/*
//TODO: add 2Pi correction
tReal CHEPEvtLiteLib::GetRij(CHEPEvtLite* evt, int i, int j) {
	if (evt==NULL) return -1.;
	if((i<0) || (i>=evt->m_nParticles) || (j<0) || (j>=evt->m_nParticles)) return -1.;
	tReal d_eta=evt->m_Particles[i].m_Eta-evt->m_Particles[j].m_Eta;
	tReal d_phi=evt->m_Particles[i].m_Phi-evt->m_Particles[j].m_Phi;
	tReal Rij=sqrt(d_eta*d_eta+d_phi*d_phi);
	return Rij;
}
*/
int CHEPEvtLiteLib::MarkDetectable(CHEPEvtLite* evt, int iMark) {
	if (evt==NULL) return 0;
	int res=0;
	for(int i=0; i<evt->m_nParticles; i++) {
		CHEPParticleLite* part=&evt->m_Particles[i]; part->m_Mark=0;
		if(part->m_Status!=CHEPParticleLite::STATUS_FINAL)continue;
		int idPAbs=abs(part->m_PID);
		if ((idPAbs==12) || (idPAbs==14) || (idPAbs==16)) continue;	// nu
		if (sqrt(part->m_Mom.GetPSqr()) < 0.05) continue;				// |P|>m_PMinMomAbs {GeV}
		// (ptot > m_PMinMomAbs) &&									// |P|>m_PMinMomAbs {GeV}
		if (part->m_Pt < 10.) continue;								// |Pt|>m_PMinPT {GeV}
		// if (part->m_Eta > m_PMaxEta) bDetectable=false;		// Eta < m_PMaxEta
		// if ((idPAbs==11) || (idPAbs==13) || (idPAbs==15)) bDetectable=false;
		part->m_Mark=iMark; res++;
	}
	return res;
}
int CHEPEvtLiteLib::GetDaughters(CHEPEvtLite* evt, int iPart, int* iDaughter1, int* iDaughter2) {
	if (evt==NULL) return 0;
	int N=evt->m_nParticles; if((iPart<0)||(iPart>=N))return 0;
	if (evt->m_Particles[iPart].m_Status==CHEPParticleLite::STATUS_FINAL)return 0; // final
	int res=0;
	int d1=evt->m_Particles[iPart].m_Daughter1_f0; int d2=evt->m_Particles[iPart].m_Daughter2_f0;
	if((d1<0)||(d1>=N))d1=-1; if((d2<0)||(d2>=N))d2=-1;
	if ((d1>=0) && (d2>=0)) {
		res=2;
	} else {
		for(int i=iPart+1; i<N; i++) {
			if (evt->m_Particles[i].m_Mother1_f0==iPart) { if(d1<0){d1=i;res++;} else {d2=i;res++;break;} }
		}
	}
	if (iDaughter1!=NULL)*iDaughter1=d1; if (iDaughter2!=NULL)*iDaughter2=d2;
	return res;
}
int CHEPEvtLiteLib::GetNLeptTTbarDecay(CHEPEvtLite* evt) { //-1 if no ttbar found or error spotted
	if (evt==NULL) return -1;
	int N=evt->m_nParticles; int noT=-1; int noTbar=-1; int i;
	//std::cout<<"N="<<N<<std::endl;
	for(i=0; i<N; i++) {
		if(evt->m_Particles[i].m_PID==6){ noT=i;if(noTbar>=0)break; }else if(evt->m_Particles[i].m_PID==-6){ noTbar=i;if(noT>=0)break; }
	}
	//std::cout<<"noT="<<noT<<", noTbar="<<noTbar<<std::endl;
	if ((noT<0) || (noTbar<0)) return -1;

	int res=0;
	int d1,d2,d1abs,d2abs,iW;
	if(GetDaughters(evt,noT,&d1,&d2)!=2)return -1;
	iW=-1; if(abs(evt->m_Particles[d1].m_PID)==24){ iW=d1; }else if(abs(evt->m_Particles[d2].m_PID)==24){ iW=d2; }
	//std::cout<<"d1="<<d1<<", d2="<<d2<<", iW="<<iW<<std::endl;
	if (iW>=0) {
		if(GetDaughters(evt,iW,&d1,&d2)!=2)return -1; d1abs=abs(evt->m_Particles[d1].m_PID); d2abs=abs(evt->m_Particles[d2].m_PID);
		if((d1abs==11)||(d2abs==11)||(d1abs==13)||(d2abs==13)||(d1abs==15)||(d2abs==15))res++;
	}
	if(GetDaughters(evt,noTbar,&d1,&d2)!=2)return -1;
	iW=-1; if(abs(evt->m_Particles[d1].m_PID)==24){ iW=d1; }else if(abs(evt->m_Particles[d2].m_PID)==24){ iW=d2; }
	if (iW>=0) {
		if(GetDaughters(evt,iW,&d1,&d2)!=2)return -1; d1abs=abs(evt->m_Particles[d1].m_PID); d2abs=abs(evt->m_Particles[d2].m_PID);
		if((d1abs==11)||(d2abs==11)||(d1abs==13)||(d2abs==13)||(d1abs==15)||(d2abs==15))res++;
	}
	//std::cout<<"res="<<res<<std::endl;
	return res;
}

// returns leptons number
int CHEPEvtLiteLib::GetTTbarDecay(CHEPEvtLite* evt,
								  int* iT, int* iTbar, int* iB1, int* iB2, int* iW1, int* iW2, int* iW1q1, int* iW1q2, int* iW2q1, int* iW2q2, int* iW1l, int* iW2l,
								  int* iJet1)
{
	if (evt==NULL) return -1;
	if ((iT==NULL)||(iTbar==NULL)||(iB1==NULL)||(iW1==NULL)||(iB2==NULL)||(iW2==NULL)||(iW1q1==NULL)||(iW1q2==NULL)||(iW2q1==NULL)||(iW2q2==NULL)||(iW1l==NULL)||(iW2l==NULL)) return -1;
	int N=evt->m_nParticles; int noT=-1; int noTbar=-1; int i;	
	for(i=0; i<N; i++) { if(evt->m_Particles[i].m_PID==6){ noT=i;if(noTbar>=0)break; }else if(evt->m_Particles[i].m_PID==-6){ noTbar=i;if(noT>=0)break; } }	
	if ((noT<0) || (noTbar<0)) return -1;
	*iT=noT; *iTbar=noTbar;

	// TODO: search particle with same Mothers to t&tbar
	// But Madgraph samples are not correct with jet mother ==> simple
	if(iJet1!=NULL)*iJet1=-1; int iTmp=((noT>noTbar)?noT:noTbar)+1; int tmpID=abs(evt->m_Particles[iTmp].m_PID);
	if ((tmpID==21) || ((tmpID>=1) && (tmpID<=5))) { if(iJet1!=NULL)*iJet1=iTmp; }

	*iW1l=-1; *iW2l=-1;
	int nLept=0;
	int d1,d2,d1abs,d2abs,iW,iB;
	if(GetDaughters(evt,noT,&d1,&d2)!=2)return -1;
	iW=-1; iB=-1; if(abs(evt->m_Particles[d1].m_PID)==24){iW=d1;iB=d2;}else if(abs(evt->m_Particles[d2].m_PID)==24){iW=d2;iB=d1;}
	if((iW<0)||(iB<0))return -1;
	if(abs(evt->m_Particles[iB].m_PID)!=5)return -1;
	*iB1=iB; *iW1=iW;
	if(GetDaughters(evt,iW,&d1,&d2)!=2)return -1; d1abs=abs(evt->m_Particles[d1].m_PID); d2abs=abs(evt->m_Particles[d2].m_PID);
	if((d1abs==11)||(d2abs==11)||(d1abs==13)||(d2abs==13)||(d1abs==15)||(d2abs==15)) {
		*iW1q1=-1; *iW1q2=-1; nLept++;
		if((d1abs==11)||(d1abs==13)||(d1abs==15)){ *iW1l=d1; }else{ *iW1l=d2; }
	} else if ((d1abs>=1) && (d1abs<=6) && (d2abs>=1) && (d2abs<=6)) { // quarconic decay
		*iW1q1=d1; *iW1q2=d2;
	} else return -1; // unexpected decay

	if(GetDaughters(evt,noTbar,&d1,&d2)!=2)return -1;
	iW=-1; iB=-1; if(abs(evt->m_Particles[d1].m_PID)==24){iW=d1;iB=d2;}else if(abs(evt->m_Particles[d2].m_PID)==24){iW=d2;iB=d1;}
	if((iW<0)||(iB<0))return -1;
	if(abs(evt->m_Particles[iB].m_PID)!=5)return -1;
	*iB2=iB; *iW2=iW;
	if(GetDaughters(evt,iW,&d1,&d2)!=2)return -1; d1abs=abs(evt->m_Particles[d1].m_PID); d2abs=abs(evt->m_Particles[d2].m_PID);
	if((d1abs==11)||(d2abs==11)||(d1abs==13)||(d2abs==13)||(d1abs==15)||(d2abs==15)) {
		*iW2q1=-1; *iW2q2=-1; nLept++;
		if((d1abs==11)||(d1abs==13)||(d1abs==15)){ *iW2l=d1; }else{ *iW2l=d2; }
	} else if ((d1abs>=1) && (d1abs<=6) && (d2abs>=1) && (d2abs<=6)) { // quarconic decay
		*iW2q1=d1; *iW2q2=d2;
	} else return -1; // unexpected decay
	return nLept;
}
void CHEPEvtLiteLib::LogEventStatistics(CHEPEvtLite* evt, std::ostream& os, bool bFinalOnly) {
	os<<"=== CHEPEvtLiteLib::LogEventStatistics ==="<<std::endl;
	if (!bFinalOnly) { os<<"Statistics for all particles (initial, virtual & final)"<<std::endl; }
	if (evt==NULL) { os<<"ERROR: NULL event"<<std::endl; return; }

	int statPID[EVTL_MAX_PART],statNCount[EVTL_MAX_PART],nStat,i;
	nStat=0; for(i=0; i<EVTL_MAX_PART; i++) { statPID[i]=0; statNCount[i]=0; }
	for(i=0; i<evt->m_nParticles; i++) {
		CHEPParticleLite* part=&evt->m_Particles[i]; part->m_Mark=0;
		if ((bFinalOnly) && (part->m_Status!=CHEPParticleLite::STATUS_FINAL)) continue; // final only
		// selection
		//if (sqrt(part->m_Mom.GetPSqr()) < 0.05) continue;				// |P|>m_PMinMomAbs {GeV}
		//if (ptot < m_PMinMomAbs) continue;							// |P|>m_PMinMomAbs {GeV}
		//if (part->m_Pt < 10) continue;								// |Pt|>m_PMinPT {GeV}
		//if (part->m_Eta > m_PMaxEta) continue;						// Eta < m_PMaxEta
		int id=part->m_PID; //int idPAbs=abs(id);
		int iStat=-1;
		for(int j=0; j<nStat; j++) { if(statPID[j]==id){ iStat=j; break; } }
		if(iStat<0) { if(nStat>=EVTL_MAX_PART)continue; iStat=nStat; statPID[iStat]=id; nStat++; }
		statNCount[iStat]++;
	}
	for(i=0; i<nStat; i++) {
		int id=statPID[i]; const char* sID=CAlHepLib::GetParticleTitle(id);
		os<<i<<". "; if(sID!=NULL){ os<<((id>=0)?"":"-")<<sID; }else{ os<<"#"<<id; }
		os<<" \t ..."<<statNCount[i]<<std::endl;
	}
	os<<"- End of list -"<<std::endl;
}

//debug switch
//#define DEBUG_VER_OLD1

///////////////////////////////////////////// CHEPEvtLiteCut: Basic interface for cuts
#ifdef Decl_CHEPEvtLiteCut
tReal CHEPEvtLiteCut::SmoothCutTest(tReal c, tReal w, tReal x, bool bMin) {
	tReal res = 1.;
	if(bMin){
		if(x<c+w){ if(x>c){ res=1.-pow(((c+w)-x)/w,8)/2; }else{ res=1./ pow(((c+w)-x)/w,8)/2; } }
	}else{
		if(x>c-w){ if(x<c){ res=1.-pow((x-(c-w))/w,8)/2; }else{ res=1./ pow((x-(c-w))/w,8)/2; } }
	}
	return res;
}
#endif Decl_CHEPEvtLiteCut

///////////////////////////////////////////////////////////// CVMPhaseSpaceGen
// #ifndef Decl_CVMPhaseSpaceGen //temp for intellisence
// #define Decl_CVMPhaseSpaceGen
// #endif //#ifndef Decl_CVMPhaseSpaceGen //temp for intellisence
#ifdef Decl_CVMPhaseSpaceGen
///////////////////////////////////////////////////////////// CVMPhaseSpaceGen::CVMHepEvtFunc
CVMPhaseSpaceGen::CVMHepEvtFunc::CVMHepEvtFunc(CVMPhaseSpaceGen* _GenRef, tReal(vmdecl *func)(CHEPEvtLite*)) : CVMGen::CVMGenFunc() {
	m_GenRef=_GenRef;
	m_pFunc=func;
}
tReal CVMPhaseSpaceGen::CVMHepEvtFunc::Call(tReal* params) {
	if ((m_GenRef!=NULL) && (m_pFunc!=NULL)) {
		if (m_GenRef->FillEvent(params)) {
			tReal wgt=m_GenRef->m_PhSpaceWgt;
			//Note: Wgt==0 means e.g. cuts specified at FillEvent-step
			if (wgt>0.) { // Note: Don't call function for Wgt=0 event
				wgt*=((*m_pFunc)(m_GenRef->GetLastEvent()));
			}
			return wgt;
		}
	}
	if(m_GenRef!=NULL)m_GenRef->IncEvtsFailed();
	return 0.;
}
CVMGen::CVMGenFunc* CVMPhaseSpaceGen::CVMHepEvtFunc::Clone() const { return new CVMHepEvtFunc(m_GenRef,m_pFunc); }
bool CVMPhaseSpaceGen::CBeamInfo::Init() {
	// init
	massSqr=mass*mass;
	pPabs=p.GetPAbs(); 

	// self-check
	if(p.m_E<0.) { std::cout<<"ERROR: Negative beam energy. Mom="<<p<<std::endl; return false; }
	if(mass<0.) { std::cout<<"ERROR: Negative beam mass value. m="<<mass<<std::endl; return false; }
	if(mass>p.m_E) { std::cout<<"ERROR: Invalid beam mass value (m>E). m="<<mass<<", E="<<p.m_E<<std::endl; return false; }

	// check mass calculation presicion
	tReal fMassFitFactor=1e-7;
	if(sizeof(tReal)<=8) {
		//elctron at 1TeV requires precision of 10^-3
		if(massSqr>0.){
			fMassFitFactor=10.e-15*(p.m_E*p.m_E/massSqr);		//approx. precision of mass calculation
			if(fMassFitFactor>0.01) { std::cout<<"[WARNING]: Numerical precision is NOT enough to operate with beam vector:"<<p<<", Mass:"<<p.GetMassSigned()<<" Required: "<<mass<<", mass error: "<<fMassFitFactor<<std::endl; }
			fMassFitFactor*=10.;								//a kind of 'eps'
		}else{ fMassFitFactor=1.e-6; }
	}
	if(fabs(p.GetSqr()-massSqr) > ((massSqr>0)?massSqr:1.)*fMassFitFactor) { std::cout<<"ERROR: Initial momentum doesn't fit initial mass. Mom:"<<p<<", Mass:"<<p.GetMassSigned()<<" Mass required: "<<mass<<", expected precision: "<<fMassFitFactor<<std::endl; return false; }

	// check 'scaleQ2'
	if(bUsePartons) {
		if(scaleQ2 <= 0.) { scaleQ2=100.; std::cout<<"WARNING: invalid 'scaleQ2' value is specified. Set to 100GeV"<<std::endl; }
	}
	return true;
}
CVMPhaseSpaceGen::CAlgorithmInfo::CAlgorithmInfo() {
	int i;
	bUseMinMaxLastTheta=false;			// Apply cuts on the last particle angle
	bUseMinMaxQ2=false;					// Apply cuts on Q2 (last final-last initial particles)
	bUsePartFinF0eCuts=false;			// Apply Emin/Emax-cut on X-final particle (e.g. for IR-photon cut)
	bInclude1dQ4=false;					// use 'd(1/Q2)' instead of d(Q^2) in 'bUseMinMaxQ2' mode
	dQdCosTrg=false;					// elsewhere dQdM2
	//
	valMinMaxLastThCos[0]=-1.; valMinMaxLastThCos[1]=1.; valMinMaxLastThCos[2]=2.;
	for(i=0;i<3;i++){ valMinMaxQ2[i]=0; }
	for(i=0;i<CVMPhSpGen_MAXN;i++){ fPartFinF0eMinCuts[i]=-1.;fPartFinF0eMaxCuts[i]=-1.;iPartFinF0eOrder[i]=1; }
	bReqPreciseMasses=false;			// re-calc energy for every particle
	bCheckGenEvts=false;				// self-check
	bFillCouplingsCache=true;
	m_bAddNuclonsToEvent=true;			// add nuclon lines to event (as 'DOC' lines)
}
/*[use default]
CVMPhaseSpaceGen::CAlgorithmInfo& CVMPhaseSpaceGen::CAlgorithmInfo::operator=(const CVMPhaseSpaceGen::CAlgorithmInfo& a) {
	//memcpy((void*)this,(void*)(&a),sizeof(CVMPhaseSpaceGen::CAlgorithmInfo));
	//TODO: check & re-implement this
	return *this;
}
*/
//bool CVMPhaseSpaceGen::CAlgorithmInfo::Init() {
//	return IsConsistent();
//}
bool CVMPhaseSpaceGen::CAlgorithmInfo::IsConsistent(
													bool bZAlignedProcess,
													int Nf,
													tReal s
													)
{
	// Update intervals (ignoring here if they are used or not)
	int i;
	valMinMaxLastThCos[2]=valMinMaxLastThCos[1]-valMinMaxLastThCos[0]; if(valMinMaxLastThCos[2]<0.)valMinMaxLastThCos[2]=0.;
	valMinMaxQ2[2]=valMinMaxQ2[1]-valMinMaxQ2[0]; if(valMinMaxQ2[2]<0.)valMinMaxQ2[2]=0.;

	// Check algorithm
	// 1. Special limits for the last particle polar angle
	if(bUseMinMaxLastTheta) {
		if((valMinMaxLastThCos[0]<-1.) || (valMinMaxLastThCos[1]>1.) || (valMinMaxLastThCos[2]<=0.)) {
			std::cout<<"ERROR. Failed to set special limits for last particle polar angle. Invalid limits: "<<valMinMaxLastThCos[0]<<" : "<<valMinMaxLastThCos[1]<<""<<std::endl;
			return false;
		}
		if(!bZAlignedProcess) {
			//TODO: allow this (reconstruct vector with cosTetha from any momentum)
			std::cout<<"ERROR. Failed to set special limits for polar angle of last particle. Z-aligned initial momenta required."<<std::endl;
			return false;
		}
	}

	// 2. Special limits for Q2
	if(bUseMinMaxQ2) {
		if (Nf <= 2) { std::cout<<"ERROR: Q2-cuts are implemented for 3+ final particles only!"; return false; }
		if (bUseMinMaxLastTheta) { std::cout<<"ERROR: Can't set cuts for both Theta[last] and Q2!"; return false; }
		if(valMinMaxQ2[1]>s) { //fix error, allow further steps
			std::cout<<"ERROR. Q2-max value ("<<valMinMaxQ2[1]<<") > 's' ("<<s<<"). Q2-max set to "<<s<<std::endl;
			valMinMaxQ2[1]=s; valMinMaxQ2[2]=valMinMaxQ2[1]-valMinMaxQ2[0];
		}
		if((valMinMaxQ2[0]<0.) || (valMinMaxQ2[1]>s) || (valMinMaxQ2[2]<=0.)) {
			std::cout<<"ERROR. Failed to set special limits for Q2. Invalid limits: "<<valMinMaxQ2[0]<<" : "<<valMinMaxQ2[1]<<", delta: "<<valMinMaxQ2[2]<<", s="<<s<<std::endl;
			return false;
		}
		// check 'bInclude1dQ4' case
		if(bInclude1dQ4) {
			if((valMinMaxQ2[0]==0.) || (valMinMaxQ2[1]==0.)) {
				std::cout<<"ERROR. Failed to set special limits for 1/Q^2. Zero spotted: "<<valMinMaxQ2[0]<<" : "<<valMinMaxQ2[1]<<", delta: "<<valMinMaxQ2[2]<<std::endl;
				return false;
			}
		}
		if (!bZAlignedProcess) {
			//TODO: allow this (reconstruct vector with Q2 from any momentum)
			std::cout<<"ERROR. Failed to set special limits for Q2. Z-aligned initial momenta required."<<std::endl;
			return false;
		}
	}

	// 3. Apply Emin-cut on second final particle (e.g. for IR-photon cut)
	if(bUsePartFinF0eCuts){
		for(i=0;i<Nf;i++){
			tReal eMinI=fPartFinF0eMinCuts[i];
			tReal eMaxI=fPartFinF0eMaxCuts[i];
			if(eMinI>=0.){ //Note: eMinI==0. may be used to select 'dM-dE' sequence without E-cuts
				//m_bUsePartFinF0eCuts=true;
				iPartFinF0eOrder[i] = ((iPartFinF0eOrder[i]>=0)?1:-1); //m_iPartFinF0eOrder: 1(dE) or -1(dE^-1)
				//[TODO: Allow smart generation?] ==> if((eMinI>0.)&&(eMinI*eMinI < m_s*1.e-10))m_iPartFinF0eOrder[i]=-1;
			}
			if(eMaxI>=0.){
				//m_bUsePartFinF0eCuts=true;
				//[Allow smart generation?] ==> if((eMinI>0.)&&(eMinI <= eMaxI*1.e-3))m_iPartFinF0eOrder[i]=-1;
			}
		}
	}

	//still unchecked:
	//dQdCosTrg=false;					// elsewhere dQdM2
	return true;
}
CVMPhaseSpaceGen::CVMFuncMultiChannel::CVMFuncMultiChannel() : CVMGen::CVMGenFunc() {}
CVMPhaseSpaceGen::CVMPhaseSpaceGen() {
	int i;

	// generator object
	m_Gen=NULL;
	m_bGenIsExt=false;					// m_Gen is external ref, do not delete

	// input values
	m_Nf=0;								// number of finals
	m_nDim=0;							// m_Gen's dimension
	for(i=0;i<CVMPhSpGen_MAXN;i++){
		m_MassFin[i]=0.;				// masses of final particles
		m_idFin[i]=0;					// IDs of  final particles
	}
	m_bMultiChannel=false;				// == m_beam[0].bUsePartons || m_beam[0].bUsePartons
	m_func_multi=NULL;					// [external link] used (and inited) in 'm_bMultiChannel' mode only

	// internal info
	m_bAzimutalSym=false;				// Azimutal symmetry of the process
	m_bInitOk=false;					// if Init() passed successfully
	for(i=0;i<CVMGen_MAXDIM;i++){
		m_lastParams[i]=-1.;			// last parameters passed to FillEvent()
	}

	// output values: event
	// Note: event is created in 'Init' unless 'SetEventExternalRef(NULL)' is called before that
	m_GenEvt=NULL;						// event generated (incl. 'm_GenEvt->wgt')
	m_bEvtExternalRef=false;			// if 'm_GenEvt' an external object

	// internal/debug output values
	m_EvtWgtDebug=0.;					// Includes 'overweight', to be used for debug ONLY.
	m_PhSpaceWgt=0.;					// Filled by FillEvent() for internal use

	// generation statictics
	m_nEvtsFailed=0;					// Error counter
	m_MaxPhotonMass=0.;
	m_MaxParticleMassError=0.;			// = M/Mcorrect-1
	m_nFillEvtCalls=0;
	m_nEmptyEvents=0;					// number of events failed to be created (e.g. due to internal cuts)

	// calculation cache
	// note: cache is uniform for all parton-channels (m_{parton} === 0)
	m_s=0.;								// |pIn1+pIn2|^2
	m_sSqrt=0.;							// sqrt(m_s)
	for(i=0;i<CVMPhSpGen_MAXN;i++){ m_MassFinSqr[i]=0.; m_M2kMin[i]=0.; }
	m_M1M2Diff2=0.;						// (m0-m1)^2 cache

	// debug info
	m_bLogEvents=false;
	m_nLogNFirstEvents=0; m_iLogEventsPrecision=-1;
	m_nHaltAfter=-1; for(i=0;i<10;i++){ m_nDebugCounter[i]=0; }
}
CVMPhaseSpaceGen::~CVMPhaseSpaceGen() {
	if(m_Gen!=NULL){ if(m_bGenIsExt){ if(m_func_multi!=NULL)m_Gen->ReleaseExtFunction(); }else{ delete m_Gen; } m_Gen=NULL; }
	if(m_func_multi!=NULL){ delete m_func_multi; m_func_multi=NULL; }
	if(m_GenEvt!=NULL){ if(!m_bEvtExternalRef)delete m_GenEvt; m_GenEvt=NULL; }
}
void CVMPhaseSpaceGen::SetEventRefExternal(CHEPEvtLite* evt) { //set external event ref., may be NULL
	if(!m_bEvtExternalRef){
		m_bEvtExternalRef=true;
		if(evt==m_GenEvt)evt=NULL; //one may not set inernal object as external
		CHEPEvtLite* evtPrev=m_GenEvt; m_GenEvt=evt; if(evtPrev!=NULL)delete evtPrev;
	}else{ m_GenEvt=evt; }
}
void CVMPhaseSpaceGen::SetEventRefInternal() { //set internal event ref.
	if(m_bEvtExternalRef){
		m_GenEvt=new CHEPEvtLite(); m_bEvtExternalRef=false;
	}else{
		if(m_GenEvt==NULL)m_GenEvt=new CHEPEvtLite();
	}
}
bool CVMPhaseSpaceGen::Init(
							int Nf,									//number of final particles
							tReal* _MassFin,						//final particle masses
							long* _idFin,							//final particle IDs (may be NULL)
							CBeamInfo& beam1, CBeamInfo& beam2,
							const CAlgorithmInfo& _alg,				//
							tReal(vmdecl *func)(CHEPEvtLite*),		//use either 'func' or 'gen_func' or 'multi_func'
							const CVMGen::CVMGenFunc* gen_func,		//use either 'func' or 'gen_func' or 'multi_func'
							CVMFuncMultiChannel* func_multi,		//use either 'func' or 'gen_func' or 'multi_func'
							CVMGen* _GenExt							//external MC core (if required)
							)
//
// Notes
// =====
//	- All initialization in 'm_GenEvt' MUST be done in GenEvent() because evt-object may be different there
//	- input 'beam1' and 'beam2' MUST be same to internal 'm_beam' values -- they MAY be used in process class
//		==> modify/initialize input values and save local copy afterwards
//
{
	if(m_bInitOk) { std::cerr<<"ERROR: Unexpected CVMPhaseSpaceGen::Init() call. May be called once only."<<std::endl; return false; }
	int i;
	m_bInitOk=false;

	// Set vectors & masses + check values
	m_Nf=Nf;							// number of finals
	if((m_Nf<2)||(m_Nf>CVMPhSpGen_MAXN)){ //we may not continue to avoid segmentation fault if n>array_sizes
		if(m_Nf>CVMPhSpGen_MAXN){ std::cout<<"Too many final particles: "<<m_Nf<<", Allowed: "<<CVMPhSpGen_MAXN<<std::endl;}else{ std::cout<<"Number of particles may not be "<<m_Nf<<std::endl; } return false;
	}
	// Note: we must initialize input values (that may be used externally!)
	if(!beam1.Init()) return false; if(!beam2.Init()) return false;
	m_beam[0]=beam1; m_beam[1]=beam2;
	for(i=0;i<Nf;i++){ m_idFin[i]=_idFin[i]; tReal m=_MassFin[i]; m_MassFin[i]=m; if(m<0.){ std::cout<<"ERROR: Negative final particle mass: "<<m<<std::endl; return false; } }

	// Channels list init
	m_bMultiChannel = ((m_beam[0].bUsePartons) || (m_beam[1].bUsePartons));
	if(m_bMultiChannel) { channels.clear(); }

	// Cache kinematic variables
	if(!UpdateCalculationCache(true,true)){ return false; }

	// Remove extra dimension if Px=Py=0
	m_bAzimutalSym=false;				// Azimutal symmetry of the process
	if ((m_beam[0].p.m_Px==0.) && (m_beam[0].p.m_Py==0.) &&
		(m_beam[1].p.m_Px==0.) && (m_beam[1].p.m_Py==0.)) m_bAzimutalSym=true;

	// Init & check algorithm
	alg=_alg;

	if(!alg.IsConsistent(m_bAzimutalSym, m_Nf, m_s)){ return false; }
	if((m_bMultiChannel) && (alg.bFillCouplingsCache)) {
		//TODO: fix Q-definition uncertainty (q in FillEvent != q in Evt) and allow this mode
		std::cout<<"WARNING: 'bFillCouplingsCache' mode is currently forbidden for parton-based process"<<std::endl;
		//return false;
		alg.bFillCouplingsCache=false;
	}

	// Save particle IDs
	for(i=0; i<m_Nf; i++){ m_idFin[i]=((_idFin!=NULL)?(_idFin[i]):0); }

	// Init particle positions
	// Examples:
	//		p1, p2, q1, q2
	//		p1, e2, q1             <--- 'first' parton after 'second' lepton!
	//		e1, e2
	// Note: the same order must be presumed in implementation class (see ALHEP->CCodeProcessAdv::AddSubProcess)
	align.posFin0=2;
	for(i=0; i<2; i++){
		align.posBeam[i]=i;
		if(alg.m_bAddNuclonsToEvent && m_beam[i].bUsePartons) {
			align.posIn[i]=align.posFin0; align.posFin0++;
		}else{
			align.posIn[i]=align.posBeam[i];
		}
	}

	// Check function existance and unity
	int nFuncFefined=0; if(func!=NULL)nFuncFefined++; if(gen_func!=NULL)nFuncFefined++; if(func_multi!=NULL)nFuncFefined++;
	if(nFuncFefined==0) { std::cout<<"PH::Init error: NULL process function"<<std::endl; return false; }
	if(nFuncFefined!=1) { std::cout<<"PH::Init error: the only ONE function may be defined (!=NULL)"<<std::endl; return false; }
	if(m_bMultiChannel && (func_multi==NULL)){ std::cout<<"PH::Init error: 'func_multi' function must be defined in multi-channel mode"<<std::endl; return false; }
	if((!m_bMultiChannel) && (func_multi!=NULL)){ std::cout<<"PH::Init ERROR: multiple channel function declared for parton-free process!"<<std::endl; return false; }

	// Create local event object
	if((!m_bEvtExternalRef) && (m_GenEvt==NULL)){
		m_GenEvt=new CHEPEvtLite(); if(m_GenEvt==NULL){ std::cout<<"CVMPhaseSpaceGen::Init ERROR: Can't create internal event object."<<std::endl; return false; }
	}

	// Init generator
	if(_GenExt!=NULL){ m_Gen=_GenExt; m_bGenIsExt=true; }else{ m_Gen=new CVMGenLite(); m_bGenIsExt=false; }
	if(m_Gen==NULL)return false; //internal error
	m_nDim=3*m_Nf-4; if(m_bAzimutalSym)m_nDim--;
	if(m_beam[0].bUsePartons)m_nDim++; if(m_beam[1].bUsePartons)m_nDim++; //for PDFs
	if(m_nDim>=CVMGen_MAXDIM){ std::cout<<"ERROR. Too many dimensions: "<<m_nDim<<". Max. allowed: "<<CVMGen_MAXDIM<<". N final partices: "<<m_Nf<<std::endl; return false; }
	m_PhSpaceWgt=0.; m_EvtWgtDebug=0.; m_nEvtsFailed=0;

	tReal _Min[CVMGen_MAXDIM]; tReal _Max[CVMGen_MAXDIM]; for(i=0;i<m_nDim;i++) { _Min[i]=0.; _Max[i]=1.; }
	if(func_multi!=NULL) {
		if(!m_bMultiChannel){ std::cout<<"PH::Init Internal ERROR"<<std::endl; return false; } //error already traced
		m_func_multi=func_multi;
		m_bInitOk=m_Gen->Init_ExternalFunc(m_nDim,_Min,_Max,m_func_multi);
	}else if(func!=NULL) {
		if(m_bMultiChannel){ std::cout<<"PH::Init Internal ERROR"<<std::endl; return false; } //error already traced
		CVMHepEvtFunc pFunc(this,func); // to be cloned in m_Gen->Init()
		m_bInitOk=m_Gen->Init(m_nDim,_Min,_Max,&pFunc); // generator uses [0,1) interval
	}else if(gen_func!=NULL){
		if(m_bMultiChannel){ std::cout<<"PH::Init Internal ERROR"<<std::endl; return false; } //error already traced
		m_bInitOk=m_Gen->Init(m_nDim,_Min,_Max,gen_func);
	}else{ std::cout<<"ERROR. Both 'gen_func==NULL' and 'func==NULL'"<<std::endl; return false; }
	if(!m_bInitOk){
		std::cout<<"ERROR initializing MC core"<<std::endl;
		std::cout<<"Failed to initialize phase space generator"<<std::endl;
		return false;
	}

	// Log
	std::cout<<"Initialized generator for "<<m_Nf<<" final particles"<<std::endl;
	std::cout<<"+------------------------ Settings -------------------------+"<<std::endl;
	std::cout<<"| sqrt(s)= "<<m_sSqrt<<std::endl;
	std::cout<<"| s = "<<m_s<<std::endl; // Nt.: 'm_s' will be updated for parton-based process in every 'FillEvent'
	std::cout<<"| z-aligned = "<<(m_bAzimutalSym?"true":"false")<<std::endl;
	std::cout<<"| UseMinMaxLastTheta = "<<(alg.bUseMinMaxLastTheta?"true":"false")<<std::endl;
	std::cout<<"| UseMinMaxQ2 = "<<(alg.bUseMinMaxQ2?"true":"false")<<std::endl;
	std::cout<<"| UsePartFinF0eCuts = "<<(alg.bUsePartFinF0eCuts?"true":"false")<<std::endl;
	if(alg.bUsePartFinF0eCuts){
		for(i=0;i<m_Nf;i++){
			if(alg.fPartFinF0eMinCuts[i]>=0.){ std::cout<<"| Final particle #"<<i<<": "<<((alg.iPartFinF0eOrder[i]>=0)?"dE":"dE^{-1}")<<" with eMin = "<<alg.fPartFinF0eMinCuts[i]<<std::endl; }
			if(alg.fPartFinF0eMaxCuts[i]>=0.){ std::cout<<"| Final particle #"<<i<<": with eMax = "<<alg.fPartFinF0eMaxCuts[i]<<std::endl; }
		}
	}
	std::cout<<"+-----------------------------------------------------------+"<<std::endl;
	return m_bInitOk;
}
bool CVMPhaseSpaceGen::AddChannel(int id, long idParton1, long idParton2) {
	//check values
	if(!m_bMultiChannel) { std::cout<<"Invalid CVMPhaseSpaceGen::AddChannel call for NON-parton-based process"<<std::endl; return false; }
	if((idParton1==0) && (idParton2==0))return false; //empty channel?
	if((m_beam[0].bUsePartons) && (idParton1==0)) { std::cout<<"ERROR in CVMPhaseSpaceGen::AddChannel: parton for beam 1 not specified"<<std::endl; return false; }
	if((m_beam[1].bUsePartons) && (idParton2==0)) { std::cout<<"ERROR in CVMPhaseSpaceGen::AddChannel: parton for beam 1 not specified"<<std::endl; return false; }
	//TODO1: check for (1<=|idPartonX|<=8) || (idPartonX==21)?
	//TODO2: check for channel doubles?
	channels.push_back(CChannelInfo(id,idParton1,idParton2));
	std::cout<<"| Channel added: "<<idParton1<<", "<<idParton2<<std::endl;
	return true;
}
bool CVMPhaseSpaceGen::UpdateCalculationCache(bool bInitChngd, bool bFinChngd) {
	if(bInitChngd){
		CVM4Vec pSum=m_beam[0].p; pSum+=m_beam[1].p; m_s=pSum.GetSqr();
		if(m_s<0.){
			if(m_s < -pSum.m_E*pSum.m_E*100.*epsReal){ std::cout<<"[ERROR] Negative initial invariant mass. Sum momenta="<<pSum<<std::endl; return false; }
			m_s=0.; //fix eps-size error
		}
		m_sSqrt=sqrt(fabs(m_s));
	}

	if(bFinChngd){
		// final masses
		int i;
		for(i=0;i<m_Nf;i++){ tReal m=m_MassFin[i]; m_MassFinSqr[i]=m*m; }
		tReal dTmp=0.; for(i=0;i<m_Nf;i++){ dTmp+=m_MassFin[i]; m_M2kMin[i]=dTmp*dTmp; } //m_M2kMin[0]=m0^2 is actually unused
		m_M1M2Diff2=(m_MassFin[0]-m_MassFin[1])*(m_MassFin[0]-m_MassFin[1]);
		if(m_s<=m_M2kMin[m_Nf-1]) { std::cout<<"[ERROR] sqrt(s) (=="<<m_sSqrt<<") is less than sum of masses (=="<<sqrt(m_M2kMin[m_Nf-1])<<")"<<std::endl; return false; }
	}
	return true;
}
void CVMPhaseSpaceGen::LogStatistics() { // push internal info to std::cout
	int i;
	std::cout<<"+-------------- CVMPhaseSpaceGen::LogStatistics ------------+"<<std::endl;
	std::cout<<"| Total FillEvent calls: "<<m_nFillEvtCalls<<std::endl;
	std::cout<<"| Total events failed:   "<<m_nEvtsFailed<<" ("<<((float)m_nEvtsFailed*100.f)/((m_nFillEvtCalls>0)?m_nFillEvtCalls:1)<<" %)"<<std::endl;
	std::cout<<"| Total empty events:    "<<m_nEmptyEvents<<" ("<<((float)m_nEmptyEvents*100.f)/((m_nFillEvtCalls>0)?m_nFillEvtCalls:1)<<" %)"<<std::endl;
	std::cout<<"| Debug counters: "; for(i=0;i<10;i++){ std::cout<<((i>0)?", ":"")<<m_nDebugCounter[i]; } std::cout<<std::endl;
	std::cout<<"+-----------------------------------------------------------+"<<std::endl;
}
tReal CVMPhaseSpaceGen::TestPointValue(double* params) { // for tests only
	CVMGen::CVMGenFunc* f=m_Gen->GetFuncRef();
	if (f==NULL) { std::cout<<"NULL function"<<std::endl; return 0.; }
	std::cout<<"=========================================== TEST"<<std::endl;
	tReal rParams[CVMGen_MAXDIM]; for(int i=0;i<m_nDim;i++){ rParams[i]=(tReal)params[i]; }
	tReal res=f->Call(rParams);
	std::cout<<"=========================================== TEST RESULT: "<<res<<std::endl;
	return res;
}
tReal CVMPhaseSpaceGen::Integrate(tReal _Precision) {
	if(!m_bInitOk)return 0.;
	return m_Gen->Integrate(DBLtR(_Precision));
}

/*
*	Notes:
*		1. Function MAY be NOT-called during m_Gen->GenEventWgt() (e.g. if all results are cached in generator)
*		2. Function MAY be called SEVERAL TIMES during m_Gen->GenEventWgt();
*		3. Generated variables MAY differ from the last function call parameters;
*	==>
*		Event MUST be filled again with actual variables
*		But this call is USUALLY ambigious ==> TODO: Inform CVMPhaseSpaceGen how Generator makes GenEventWgt();
*			Upd:
*				- check 'AreLastParams()' added inside 'FillEvent'
*/
bool CVMPhaseSpaceGen::GenEvent() {
	if(!m_bInitOk)return false;
	bool genOK=m_Gen->GenEvent();

	if(m_bMultiChannel) {
		int idCh=-1;
		if(genOK){ if(m_func_multi!=NULL){ idCh=m_func_multi->SelectChannel(m_Gen->m_GenValues); } genOK=(idCh>=0); }
		if(genOK){ genOK=FillEvent(m_Gen->m_GenValues); }
		if(genOK){ FillChannelInfo(idCh); }
	}else{
		if(genOK){ genOK=FillEvent(m_Gen->m_GenValues); }
	}

	if(genOK){
		if(m_GenEvt!=NULL){m_GenEvt->wgt=m_Gen->m_Weight;m_GenEvt->wgtErr=m_Gen->m_WgtError;} m_EvtWgtDebug=m_Gen->m_WeightDebug;
	}else{
		if(m_GenEvt!=NULL){m_GenEvt->wgt=0.;m_GenEvt->wgtErr=0.;} m_EvtWgtDebug=0.;
	}
	return genOK;
}
bool CVMPhaseSpaceGen::GenEventWgt() {
	if(!m_bInitOk)return false;
	bool genOK=m_Gen->GenEventWgt();

	if(m_bMultiChannel) {
		int idCh=-1;
		if(genOK){ if(m_func_multi!=NULL){ idCh=m_func_multi->SelectChannel(m_Gen->m_GenValues); } genOK=(idCh>=0); }
		if(genOK){ genOK=FillEvent(m_Gen->m_GenValues); }
		if(genOK){ FillChannelInfo(idCh); }
	}else{
		if(genOK){ genOK=FillEvent(m_Gen->m_GenValues); }
	}

	if(genOK){
		if(m_GenEvt!=NULL){m_GenEvt->wgt=m_Gen->m_Weight;m_GenEvt->wgtErr=m_Gen->m_WgtError;} m_EvtWgtDebug=m_Gen->m_WeightDebug; /*old: =m_Gen->m_Weight*/
	}else{
		if(m_GenEvt!=NULL){m_GenEvt->wgt=0.;m_GenEvt->wgtErr=0.;} m_EvtWgtDebug=0.;
	}
	return genOK;
}
bool CVMPhaseSpaceGen::GenEvtDebug() { //Fast & incorrect events for viewing the integration grid.
	if(!m_bInitOk)return false;
	bool genOK=false; if(m_Gen->ClassName()==CVMGenLite::CLASS_NAME){ genOK=((CVMGenLite*)m_Gen)->GenEvtDebug(); }

	if(m_bMultiChannel) {
		int idCh=-1;
		if(genOK){ if(m_func_multi!=NULL){ idCh=m_func_multi->SelectChannel(m_Gen->m_GenValues); } genOK=(idCh>=0); }
		if(genOK){ genOK=FillEvent(m_Gen->m_GenValues); }
		if(genOK){ FillChannelInfo(idCh); }
	}else{
		if(genOK){ genOK=FillEvent(m_Gen->m_GenValues); }
	}

	if(genOK){
		if(m_GenEvt!=NULL){m_GenEvt->wgt=m_Gen->m_Weight;m_GenEvt->wgtErr=m_Gen->m_WgtError;} m_EvtWgtDebug=m_Gen->m_WeightDebug; /*old: =m_Gen->m_Weight*/
	}else{
		if(m_GenEvt!=NULL){m_GenEvt->wgt=0.;m_GenEvt->wgtErr=0.;} m_EvtWgtDebug=0.;
	}
	return genOK;
}
//[inlined] bool CVMPhaseSpaceGen::SaveLastParams(tReal* params) { // save parameters as 'last' (called by 'FillEvent' only)
//	for(i=0;i<m_nDim;i++){ m_lastParams[i]=params[i]; }
//}
//[inlined] bool CVMPhaseSpaceGen::ClearLastParams() { // clear 'last' parameters (e.g. if 'FillEvent' failed)
//	m_lastParams[0]=-1.; //that's enough
//}
bool CVMPhaseSpaceGen::AreLastParams(tReal* params) { // check parameters to previous 'FillEvent' call
	if(params==NULL)return false;
	for(int i=0;i<m_nDim;i++){ if(m_lastParams[i]!=params[i]){ return false; } }
	return true;
}
bool CVMPhaseSpaceGen::FillChannelInfo(int idChannel) { // fills parton IDs in multichannel mode
	int idParton[]={0,0};		// used in multichannel mode only
	if(idChannel>=0){
		std::vector<CChannelInfo>::iterator it; for(it=channels.begin(); it!=channels.end(); ++it){
			if(it->idChannel==idChannel) { idParton[0]=it->idParton1; idParton[1]=it->idParton2; break; }
		}
	}
	for(int i=0; i<2; i++) {
		if(m_beam[i].bUsePartons){ m_GenEvt->m_Particles[align.posIn[i]].m_PID=idParton[i]; }
	}
	return true;
}
bool CVMPhaseSpaceGen::FillEvent(tReal* params, int idChannel) {	// fills m_GenEvt
//
// Notes:
//	1. 'chnl==-1' is allowed for parton-based process, if PDF value is multiplied by user (manually)
//	2. Q-vector below (last final-last parton) is different to evt-definition (last - second)
//		- leave 'as is' --> 'alg.bFillCouplingsCache' should not be used for parton process
//
//	3. PDFs are applied if FillEvent(params,SPECIFIC_CHANNEL) is called.
//		==> do not apply it twice in process implementation procedure!
//	4. Normally FillEvent(,-1) is invoked
//		==> and PDFs are applied in process implementation class
//
// TODO:
//		Ref to B&K, Weinzierl, V.Mossolov or noone ?
//
	if(AreLastParams(params)){ return true; }					// TODO: skip this check in debug version!
	ClearLastParams();
	if(m_nFillEvtCalls<LONG_MAX)m_nFillEvtCalls++;
	
	if(m_GenEvt==NULL){
		if(m_nFillEvtCalls<10){ std::cout<<"FillEvent failed: m_GenEvt==NULL"<<std::endl; }
		return false;
	}

	if((!m_bInitOk) || (params==NULL))return false;

	int dim=0; //no. of next parameter to use
	m_PhSpaceWgt=1.;
	int i;

	bool bLog=((m_bLogEvents) || (m_nFillEvtCalls<m_nLogNFirstEvents));
	//bLog=(m_nFillEvtCalls==173);
	//bLog=(m_nFillEvtCalls==1058800);

	// Save initial vectors, N[particles] & particle IDs
	// beams
	int idParton[]={0,0};		// used in multichannel mode only
	if((m_bMultiChannel)&&(idChannel>=0)){
		// find channel info (slow and not recommended!)
		if(m_nFillEvtCalls<200){
			std::cout<<"Warning: CVMPhaseSpaceGen::FillEvent(,SPECIFIC_CHANNEL) call is invoked!"<<std::endl;
		}
		std::vector<CChannelInfo>::iterator it; for(it=channels.begin();it!=channels.end();++it){ if(it->idChannel==idChannel) { idParton[0]=it->idParton1; idParton[1]=it->idParton2; break; } }
	}
	bool bUpd_Shat=false;
	for(i=0; i<2; i++) {
		if(m_beam[i].bUsePartons) {
			if(alg.m_bAddNuclonsToEvent){
				CHEPParticleLite* partB=&(m_GenEvt->m_Particles[align.posBeam[i]]);
				partB->m_Mom=m_beam[i].p; partB->m_PID=m_beam[i].id; partB->m_Mother1_f0=-1; partB->m_Mother2_f0=-1;
				partB->m_Status=CHEPParticleLite::STATUS_DOC;
			}
			tReal x=params[dim];dim++;
			CHEPParticleLite* part = &(m_GenEvt->m_Particles[align.posIn[i]]);
			part->m_Mom=m_beam[i].p; part->m_Mom.m_E=m_beam[i].pPabs; part->m_Mom*=x; bUpd_Shat=true;
			part->m_PID=idParton[i];
			part->m_Status=CHEPParticleLite::STATUS_INITIAL; part->m_Mother1_f0=align.posBeam[0]; part->m_Mother2_f0=align.posBeam[1];
			//
			// Important notes:
			//		1. PDFs are applied if FillEvent(params,SPECIFIC_CHANNEL) is called.
			//			==> do not apply it twice in process implementation procedure!
			//		2. Normally FillEvent(,-1) is invoked
			//			==> and PDFs are applied in process implementation class
			//	
			if((idParton[i]!=0)&&(m_beam[i].pdf!=NULL)) {
				int pid_PDF=idParton[i]; if(pid_PDF==PDG_KF_GLUON)pid_PDF=0;
				m_PhSpaceWgt*=(tReal)(m_beam[i].pdf->GetDensity(pid_PDF,x,m_beam[i].scaleQ2));
			}
		}else{
			CHEPParticleLite* part=&(m_GenEvt->m_Particles[align.posIn[i]]);
			part->m_Mom=m_beam[i].p; part->m_PID=m_beam[i].id; part->m_Mother1_f0=-1; part->m_Mother2_f0=-1;
			part->m_Status=CHEPParticleLite::STATUS_INITIAL;
			if(align.posBeam[i] != align.posIn[i]){ std::cerr<<"Internal error in ph::FillEvent()"<<std::endl; return false; }
		}
	}
	// finals
	m_GenEvt->m_nParticles=align.posFin0+m_Nf;
	for(i=0; i<m_Nf; i++){
		CHEPParticleLite* part=&(m_GenEvt->m_Particles[align.posFin0+i]);
		part->m_PID=m_idFin[i]; part->m_Status=CHEPParticleLite::STATUS_FINAL;
		part->m_Mother1_f0=align.posIn[0]; part->m_Mother2_f0=align.posIn[1];
	}

	bool bZeroEvt=false;

	if(bUpd_Shat) {
		CVM4Vec pSum = m_GenEvt->m_Particles[align.posIn[0]].m_Mom; pSum+=m_GenEvt->m_Particles[align.posIn[1]].m_Mom;
		m_s=pSum.GetSqr();
		if(m_s<0.){ m_s=0.; } //fix eps-size error (we presume no internal error here)
		if(m_s<=m_M2kMin[m_Nf-1]) {
			m_sSqrt=0.; bZeroEvt=true; // \hat{s} < Sum{masses} (== sqrt(m_M2kMin[m_Nf-1])
		}else{
			m_sSqrt=sqrt(m_s);
		}
	}

	if(!bZeroEvt){
		tReal _Pi=CVMConst::Pi();
		//tReal _Pi=3.1415926535897932385;
		tReal _2Pi=2.*_Pi; tReal _1d4PiSqr=1./(_Pi*_Pi*16.);
		CVM4Vec qPrev; tReal _M2kPrev,_MkPrev;
		CVM4Vec qk=m_GenEvt->m_Particles[align.posIn[0]].m_Mom; qk+=m_GenEvt->m_Particles[align.posIn[1]].m_Mom; tReal _M2k=m_s; tReal _Mk=m_sSqrt;
		m_PhSpaceWgt*=_2Pi; // From common dG_n -> dM2*dG_2*dM2*dG_2... dG_2 decomposition

		//Check parameters (z-alignment for initial vector & others)
		if((qPrev.m_Px!=0.) || (qPrev.m_Py!=0.)) {
			if(m_bAzimutalSym) { std::cout<<"[err]Internal error with 'bAzimutalSym' definition"<<std::endl; return false; }
			if(alg.bUseMinMaxLastTheta) { std::cout<<"[err]Internal error with 'bUseMinMaxLastTheta' definition"<<std::endl; return false; }
			if(alg.bUseMinMaxQ2) { std::cout<<"[err]Internal error with 'bUseMinMaxQ2' definition"<<std::endl; return false; }
		}
		if (alg.bUseMinMaxQ2 && (m_Nf <= 2)) { std::cout<<"[err]Internal error with 'bUseMinMaxQ2' definition"<<std::endl; return false; }

		//Local switches
		bool bSmartLowQRegion=(alg.bFillCouplingsCache && alg.bUseMinMaxQ2); //Calculate k.k couplings via q.k. Use with

		bool bRestoreStdOutPrec=false; int dbgPrecOld;
		if((bLog) && (m_iLogEventsPrecision>0)){ dbgPrecOld=std::cout.precision(m_iLogEventsPrecision); bRestoreStdOutPrec=true; }
		if (bLog) { std::cout<<"[evt].................................................... Creating Event #"<<m_nFillEvtCalls<<std::endl; }
		int pos0=align.posIn[0]; int pos1=align.posIn[1]; int posF=align.posFin0;	//for faster access & shorter code

		if (alg.bFillCouplingsCache) { // Don't move this to Init() function
			if (m_Nf+2 > EVTL_MAX_CPL_PART) { std::cout<<"[err]ERROR: Too many particles. Couplings cache is unavailable"<<std::endl; alg.bFillCouplingsCache=false; }
			m_GenEvt->ClearCouplingsCache(); //Call after 'm_GenEvt->m_nParticles' is set
			m_GenEvt->SetCpl(pos0,pos1,m_GenEvt->m_Particles[pos0].m_Mom.GetCpl(m_GenEvt->m_Particles[pos1].m_Mom));
		}

		// Vec 'Q': vecQ=pFinLast-pInLast;
		tReal Q2=-1.; CVM4Vec vecQ; bool bVecQCreated=false; //to save/cache value
		for(int k=m_Nf-1; k>=1; k--) { // 'k': number of final particle (first: 0)
			qPrev=qk; _M2kPrev=_M2k; _MkPrev=_Mk; //std::cout<<"k="<<k<<", qPrev="<<qPrev<<", _M2kPrev="<<_M2kPrev<<"("<<qPrev.GetSqr()<<"), _MkPrev="<<_MkPrev<<std::endl;
			m_PhSpaceWgt/=_2Pi; // From common dG_n -> {dM2*dG_2}*{dM2*dG_2}... dG_2 decomposition

			// dPhi (Azimutal angle)
			tReal fi;
			if((k==m_Nf-1)&&(m_bAzimutalSym)){
				fi=0.; // Note: generate the free azimutal angle AFTER all the parameters are applied!
			} else {
				fi=_2Pi*params[dim];dim++; m_PhSpaceWgt*=_2Pi;
			}

			if ((k==(m_Nf-1)) && (alg.bUseMinMaxQ2)) {
				// 1. dQ^{+-2)*dCosT
				// 2. dQ^(+-2)*dM^2

				//'Target' system: pIn1={xE1,xP1}, pIn2={m2,0}
				//TODO: check if it's correct if (!m_bAzimutalSym)
				tReal mIn1=m_beam[0].mass; tReal mIn1Sqr=m_beam[0].massSqr;
				tReal mIn2=m_beam[1].mass; tReal mIn2Sqr=m_beam[1].massSqr;
				tReal mk=m_MassFin[k]; tReal mkSqr=m_MassFinSqr[k];
				tReal xE1=(_M2kPrev-mIn1Sqr-mIn2Sqr)/(2.*mIn2);
				tReal xP1=sqrt((tReal)((xE1-mIn1)*(xE1+mIn1))); //TODO: insert sign here!

				if (bLog) {
					std::cout<<"[evt]Initial mass #1: "<<mIn1<<std::endl; //0.00051099890000000001
					std::cout<<"[evt]Initial mass #1 sqr: "<<mIn1Sqr<<std::endl; //2.6111987580121002e-007
					std::cout<<"[evt]Initial mass #2: "<<mIn2<<std::endl; //0.93827201299999996
					std::cout<<"[evt]Initial mass #2 sqr: "<<mIn2Sqr<<std::endl; //0.88035437037907205
					std::cout<<"[evt][Lab.syst.] Initial mass#2 error: "<<fabs((mIn2-m_GenEvt->m_Particles[pos1].m_Mom.GetMassSigned()))/mIn2<<std::endl; //asdf
					std::cout<<"[evt]Final mass #"<<k<<": "<<mk<<std::endl; //0.93827201299999996
					std::cout<<"[evt]Final mass #"<<k<<" sqr: "<<mkSqr<<std::endl; //0.88035437037907205
					std::cout<<"[evt][Target syst.] E#1: "<<xE1<<std::endl; //54125.015013454336
					std::cout<<"[evt][Target syst.] P#1:"<<xP1<<std::endl; //54125.015013454336
				}

				// dQ^2. Q^2=2*m2*Ep-m2^2-mp^2;
				if(k<=1){ std::cout<<"[err]Internal error with 'bUseMinMaxQ2' definition"<<std::endl; return false; }
				if(alg.bInclude1dQ4) { //dQ^-2
					tReal _1dQ2min=((tReal)1.)/alg.valMinMaxQ2[1]; tReal _1dQ2max=((tReal)1.)/alg.valMinMaxQ2[0];
					tReal _1dQ2=_1dQ2min+(_1dQ2max-_1dQ2min)*params[dim]; dim++; m_PhSpaceWgt*=(_1dQ2max-_1dQ2min);
					Q2=((tReal)1.)/_1dQ2;
					m_PhSpaceWgt*=(Q2*Q2);
				} else { //dQ^2
					Q2=alg.valMinMaxQ2[0]+alg.valMinMaxQ2[2]*params[dim]; dim++; m_PhSpaceWgt*=alg.valMinMaxQ2[2]; //Q2_max-Q2_min
				}
				tReal xEpMinusMp=(Q2+((mIn2==mk)?0.:((mIn2-mk)*(mIn2-mk))))/(2.*mIn2);
				tReal xEp=(Q2+mIn2Sqr+mkSqr)/(2.*mIn2);
				if(bLog)std::cout<<"[evt]Q2 generated: "<<Q2<<std::endl; //54125.015013454336
				if(bLog)std::cout<<"[evt][Target syst.] E#"<<k<<": "<<xEp<<std::endl; //0.93827201299999996
				if(bLog)std::cout<<"[evt][Target syst.] E-M#"<<k<<": "<<xEpMinusMp<<" (Calc: "<<xEp-mk<<")"<<std::endl; //4.5784795498423367e-028,0

				if(xEp<mk){ m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[0]++; break; }
				tReal xPp=sqrt(xEpMinusMp*(xEp+mk)); //Nt. 'mk'==='Mp' (by definition)
				if(bLog)std::cout<<"[evt][Target syst.] P"<<k<<": "<<xPp<<std::endl; //2.9311633266366792e-014
				m_PhSpaceWgt*=_1d4PiSqr;

				tReal xQ0=xEpMinusMp+((mIn2==mk)?0:(mk-mIn2)); // 0-component of Q-vector, =(Ep-mIn2)
				//Note: Here vector 'Eq,Pq' is the 'rest', but NOT (p2-p1)
				tReal xEqMinusE1=-xQ0; //=mIn2-xEp=-Q_{0}
				if(bLog)std::cout<<"[evt][Target syst.] E(rest)-E1: "<<xEqMinusE1<<std::endl; //-4.5784795498423367e-028
				tReal xEq=xE1+xEqMinusE1; if(xEq<0.){ m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[1]++; break; } //TODO: complete here
				if(bLog)std::cout<<"[evt][Target syst.] E(rest): "<<xEq<<std::endl; //54125.015013454336

				tReal cost,sint;
				if(alg.dQdCosTrg) { // Note '!dQdCosTrg' is much better choice now!
					m_PhSpaceWgt*=xPp/mIn2Sqr; // TODO: check 1/mIn2 or 1/mIn2Sqr
					// dcosT(Target system)
					tReal cosTmin=-1; tReal cosTmax=1;
					tReal _MkMax=_MkPrev-mk; tReal _M2kMax=((_MkMax>=0.)?(_MkMax*_MkMax):0.);
					tReal _M2kMin=m_M2kMin[k-1]; if(_M2kMax<_M2kMin)_M2kMax=_M2kMin;
					//ToDo: adjust tmpXfor small Q2
					tReal tmpX=xE1*xE1-mIn1Sqr+xEp*xEp-mkSqr-xEq*xEq;
					tReal cosTmin2=(tmpX+((xP1>=0.)?_M2kMin:_M2kMax))/(2.*xP1*xPp);
					tReal cosTmax2=(tmpX+((xP1>=0.)?_M2kMax:_M2kMin))/(2.*xP1*xPp);
					if (cosTmax2<cosTmin2){ std::cout<<"[err]Internal error in FillEvent"<<std::endl; return false; }
					if (cosTmin<cosTmin2) { cosTmin=cosTmin2; if(cosTmin>1.)cosTmin=1.; }
					if (cosTmax>cosTmax2) { cosTmax=cosTmax2; if(cosTmax<-1.)cosTmax=-1.; }
					//TODO: involve (1-cost) or (1+cost) parameter and switch (rnd) and (1.-rnd) more smart!
					cost=cosTmin+(cosTmax-cosTmin)*params[dim]; dim++;
					m_PhSpaceWgt*=(cosTmax-cosTmin);
					sint=sqrt(((tReal)1.-cost)*((tReal)1.+cost)); if(sint>1.)sint=1.;
					if (bLog) {
						std::cout<<"[evt][Target syst.] CosTheta limits: "<<cosTmin<<" : "<<cosTmax<<std::endl;
						std::cout<<"[evt] Sum{rest}^2 limits: "<<_M2kMin<<" : "<<_M2kMax<<std::endl;
					}
					_M2k=2.*xP1*xPp*cost-tmpX; _Mk=sqrt(fabs(_M2k));
					if ((_M2k<_M2kMin) || (_M2k>_M2kMax)) { m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[2]++; break; }
				} else {
					// dM^2 (+ reconstruction in Target system)
					m_PhSpaceWgt*=1./fabs(2.*xP1*mIn2); // TODO: check 1/mIn2 or 1/mIn2Sqr
					tReal cosTmin=-1; tReal cosTmax=1;
					tReal _MkMax=_MkPrev-mk; tReal _M2kMax=((_MkMax>=0.)?(_MkMax*_MkMax):0.);
					tReal _M2kMin=m_M2kMin[k-1];
					if(_M2kMax<=_M2kMin){ m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[2]++; break; }//_M2kMax=_M2kMin;
					//tReal P12Pp2minusEq2=-mIn1Sqr+xPp*xPp-xEpMinusMp*xEpMinusMp+2.*xEpMinusMp*xE1;
					//if(mIn2!=mk) P12Pp2minusEq2+=(mIn2-mk)*(2.*(xEpMinusMp-xE1)-(mIn2-mk));
					tReal P12Pp2minusEq2=-mIn1Sqr+xPp*xPp-xEqMinusE1*xEqMinusE1-2.*xEqMinusE1*xE1; // same to prev. 2 lines
					if(bLog)std::cout<<"[evt][Target syst.] P12Pp2minusEq2="<<P12Pp2minusEq2<<" (Calc: "<<xPp*xPp+xE1*xE1-mIn1Sqr-xEq*xEq<<")"<<std::endl;
					tReal _M2kMin2=-fabs(2.*xP1*xPp)-P12Pp2minusEq2; tReal _M2kMax2=fabs(2.*xP1*xPp)-P12Pp2minusEq2;
					if(bLog)std::cout<<"[evt] Sum{rest}^2 limits (partial): "<<_M2kMin2<<" : "<<_M2kMax2<<std::endl;
					if (_M2kMin<_M2kMin2) { _M2kMin=_M2kMin2; if(_M2kMin>=_M2kMax){ m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[2]++; break; } }
					if (_M2kMax>_M2kMax2) { _M2kMax=_M2kMax2; if(_M2kMax<=_M2kMin){ m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[2]++; break; } }
					if(bLog)std::cout<<"[evt] Sum{rest}^2 limits : "<<_M2kMin<<" : "<<_M2kMax<<std::endl; //4.7366417302190307e-007,4.8001014338434693e-007
					_M2k=_M2kMin+(_M2kMax-_M2kMin)*params[dim]; dim++; _Mk=sqrt(fabs(_M2k));
					m_PhSpaceWgt*=(_M2kMax-_M2kMin);
					// TODO: find 1/fabs(2.*xP1*xPp) origin!!
					//TODO: involve (1-cost) or (1+cost) parameter
					cost=(P12Pp2minusEq2+_M2k)/(2.*xP1*xPp);
					//if((cost<-1.) || (cost>1.)) { m_PhSpaceWgt=0.; bZeroEvt=true; break; }
					if(cost<-1.)cost=-1.; if(cost>1.)cost=1.; //check for errors cost>>1 or cost<<-1
					sint=sqrt(((tReal)1.-cost)*((tReal)1.+cost)); if(sint>(tReal)1.)sint=(tReal)1.;
				}
				if(bLog)std::cout<<"[evt][Target syst.] cosT#"<<k<<": "<<cost<<", sinT#"<<k<<": "<<sint<<std::endl;
				if(bLog)std::cout<<"[evt] Sum{rest}^2: "<<_M2k<<", Inv.Mass{rest}: "<<_Mk<<std::endl;
				if(bLog)std::cout<<"[evt] Weight Factor (intermediate): "<<m_PhSpaceWgt<<std::endl;

				if((xEp==0.)&&(xPp==0.)){ tReal a=((_Mk>0.)?(_Mk):(_MkPrev));xEp=a*epsReal;xPp=a*epsReal; } //Fix 0-photon momentum (if rnd==1.|0.)
				qk.Set(xEq,-xPp*sint*cos(fi),-xPp*sint*sin(fi),xP1-xPp*cost);
				if(bLog)std::cout<<"[evt][Target syst.] Sum{rest}: "<<qk<<std::endl;
				if(bLog)std::cout<<"[evt][Target syst.] Sum{rest}^2 check: "<<qk.GetSqr()<<std::endl;
				m_GenEvt->m_Particles[posF+k].m_Mom.Set(xEp,-qk.m_Px,-qk.m_Py,xPp*cost);
				if(bLog)std::cout<<"[evt][Target syst.] Vec#"<<k<<": "<<m_GenEvt->m_Particles[posF+k].m_Mom<<std::endl;
				if(bLog)std::cout<<"[evt] Mass#"<<k<<" check: "<<m_GenEvt->m_Particles[posF+k].m_Mom.GetSqr()<<std::endl;
				// cache q-vector
				vecQ.Set(xQ0,-qk.m_Px,-qk.m_Py,xPp*cost); // ={Ep-MIn2,Pp}

				if (alg.bFillCouplingsCache) {
					//in target system
					m_GenEvt->SetCpl(pos0,posF+k,m_GenEvt->m_Particles[posF+k].m_Mom.m_E*xE1-m_GenEvt->m_Particles[posF+k].m_Mom.m_Pz*xP1);
					m_GenEvt->SetCpl(pos1,posF+k,m_GenEvt->m_Particles[posF+k].m_Mom.m_E*mIn2);
					m_GenEvt->SetQCpl(pos0,xQ0*xE1-xPp*cost*xP1);
					m_GenEvt->SetQCpl(pos1,xQ0*mIn2); //==Q2/2 if (mIn2==mk)
					m_GenEvt->SetQCpl(posF+k,(-Q2+((mIn2==mk)?0.:(mkSqr-mIn2Sqr)))/2.); //= xQ0*xEp-xPp*xPp = xQ0*mIn2-Q2
				}
				if(bLog)std::cout<<"[evt][Target syst.] Inv.Mass{rest} check: "<<qk.GetMassSigned()<<" (diff: "<<(_Mk-fabs(qk.GetMassSigned()))/_Mk<<" )"<<std::endl;

				//Note: Q2 loses ~6 digits at every Lorentz-transformation (at Q2<=10^-13).

				//TODO: use Q-cpls instead of User-cpls in Born-calc! q.k1 is more correct in q-cpls! //asdf
				//Done? (I don't see user couplings below any more, by also no 'GetQCpl'-calls)

				// To lab. system
				if (false) { //if(fi==0.){
					//
					// cache Lorentz transform (not tested well)
					//
					if(bLog)std::cout<<"[evt] Target->Lab.syst shift vec: "<<m_GenEvt->m_Particles[pos1].m_Mom<<std::endl;
					// 920,0,0,-919.99952
					tReal shE=m_GenEvt->m_Particles[pos1].m_Mom.m_E; tReal shPx=m_GenEvt->m_Particles[pos1].m_Mom.m_Px;
					tReal shPy=m_GenEvt->m_Particles[pos1].m_Mom.m_Py; tReal shPz=m_GenEvt->m_Particles[pos1].m_Mom.m_Pz; tReal shM=mIn2;
					tReal resE,resPx,resPy,resPz,r;
					tReal tmpE,tmpPx,tmpPy,tmpPz;
					//LorMToK:
					//resE=(tmpE*shE + tmpPx*shPx + tmpPy*shPy + tmpPz*shPz)/shM;
					//r=(tmpE+resE)/(shE+shM); resPx=tmpPx+shPx*r; resPy=tmpPy+shPy*r; resPz=tmpPz+shPz*r;

					tmpE=m_GenEvt->m_Particles[posF+k].m_Mom.m_E; tmpPx=m_GenEvt->m_Particles[posF+k].m_Mom.m_Px;
					tmpPy=m_GenEvt->m_Particles[posF+k].m_Mom.m_Py; tmpPz=m_GenEvt->m_Particles[posF+k].m_Mom.m_Pz;
					resE=(tmpE*shE + tmpPx*shPx + tmpPy*shPy + tmpPz*shPz)/shM;
					r=(tmpE+resE)/(shE+shM); resPx=tmpPx+shPx*r; resPy=tmpPy+shPy*r; resPz=tmpPz+shPz*r;
					if (bLog) { std::cout<<"resE: "<<resE<<std::endl; }
					if (bLog) { std::cout<<"r: "<<r<<std::endl; }
					m_GenEvt->m_Particles[posF+k].m_Mom.m_E=resE; m_GenEvt->m_Particles[posF+k].m_Mom.m_Px=resPx;
					m_GenEvt->m_Particles[posF+k].m_Mom.m_Py=resPy; m_GenEvt->m_Particles[posF+k].m_Mom.m_Pz=resPz;

					tmpE=qk.m_E; tmpPx=qk.m_Px; tmpPy=qk.m_Py; tmpPz=qk.m_Pz;
					resE=(tmpE*shE + tmpPx*shPx + tmpPy*shPy + tmpPz*shPz)/shM;
					r=(tmpE+resE)/(shE+shM); resPx=tmpPx+shPx*r; resPy=tmpPy+shPy*r; resPz=tmpPz+shPz*r;
					if (bLog) { std::cout<<"resE: "<<resE<<std::endl; }
					if (bLog) { std::cout<<"r: "<<r<<std::endl; }
					qk.m_E=resE; qk.m_Px=resPx; qk.m_Py=resPy; qk.m_Pz=resPz;
				} else { // common procedure
					m_GenEvt->m_Particles[posF+k].m_Mom.LorMToK(mIn2,m_GenEvt->m_Particles[pos1].m_Mom);
					qk.LorMToK(mIn2,m_GenEvt->m_Particles[pos1].m_Mom);
				}
				vecQ.LorMToK(mIn2,m_GenEvt->m_Particles[pos1].m_Mom); bVecQCreated=true;
				if(bLog)std::cout<<"[evt][Lab.syst.] Q2 calc. error: "<<fabs(vecQ.GetSqr()+Q2)/Q2<<std::endl;

				if(alg.bReqPreciseMasses) { // make vector square precise
					m_GenEvt->m_Particles[posF+k].m_Mom.m_E=sqrt(mkSqr+m_GenEvt->m_Particles[posF+k].m_Mom.GetPSqr());
					qk.m_E=sqrt(_M2k+qk.GetPSqr());
				}
			} else {
				// 3. dM^2*...

				// dM^2_i
				tReal lambdaSqrt=0.;
				if(k>1) {
					tReal _MkMax=_MkPrev-m_MassFin[k]; if(_MkMax<0.)_MkMax=0.; tReal _M2kMax=_MkMax*_MkMax;
					tReal _M2kMaxMinDiff=_M2kMax-m_M2kMin[k-1]; if(_M2kMaxMinDiff<0.)_M2kMaxMinDiff=0.;
					m_PhSpaceWgt*=_M2kMaxMinDiff;
					_M2k=m_M2kMin[k-1]+_M2kMaxMinDiff*params[dim]; dim++; if(_M2k>_M2kMax)_M2k=_M2kMax; _Mk=sqrt(_M2k);
					lambdaSqrt=sqrt((_M2kMax-_M2k)*(4.*m_MassFin[k]*_MkPrev+_M2kMax-_M2k)); //lambdaSqrt checked
					if(bLog)std::cout<<"[evt] Sum{rest}^2 limits: "<<m_M2kMin[k-1]<<" : "<<_M2kMax<<std::endl;
					if(bLog)std::cout<<"[evt] Sum{rest}^2: "<<_M2k<<", Inv.Mass{rest}: "<<_Mk<<std::endl;
				} else {
					_M2k=m_MassFinSqr[0]; _Mk=m_MassFin[0];
					lambdaSqrt=sqrt(((_M2kPrev>m_M2kMin[1])?(_M2kPrev-m_M2kMin[1]):0.)*(_M2kPrev-m_M1M2Diff2)); //lambdaSqrt checked
					//if(bLog)std::cout<<"[evt] mom^2{last}^2: "<<_M2k<<", Inv.Mass{last}: "<<_Mk<<std::endl;
				}

				// Select reco mode
				int iRecoMode=0;										// dM^2*dCosT[c.m.s.]
				if(alg.bUseMinMaxLastTheta && (k==(m_Nf-1)))iRecoMode=1;	// dM^2*dCosT[lab.]
				else if(
					(alg.bUsePartFinF0eCuts && (alg.fPartFinF0eMinCuts[k]>=0.)) //Note: 'eMin' is mandatory for 'eMax' usage!!!
					&&
					((qPrev.m_Pz!=0.)||(qPrev.m_Px!=0.)||(qPrev.m_Py!=0.))
					)iRecoMode=2;										// dM^2*dE[lab.] (PAbs(qPrev)>0 is required)
				if((qPrev.m_E==0.)
					||(_MkPrev<=0.)
					||(_M2kPrev<=0.))iRecoMode=9;						// Ersatz case
				//if(bLog){ std::cout<<"[evt] alg.bUsePartFinF0eCuts="<<alg.bUsePartFinF0eCuts<<std::endl; }
				//if(bLog){ std::cout<<"[evt] alg.fPartFinF0eMinCuts[k]="<<alg.fPartFinF0eMinCuts[k]<<std::endl; }
				//if(bLog){ std::cout<<"[evt] qPrev="<<qPrev<<std::endl; }
				if(bLog){ std::cout<<"[evt] iRecoMode: "<<iRecoMode<<std::endl; }

				switch(iRecoMode){
				case 9: {
					// Ersatz case: (E0==0) || (s<s_min) || (_M2kPrev==0)
					// Nt. 'Unphysical case' may be allowed if discrepancy is ~'eps'

					tReal r=params[dim]; dim++; //'Parameter' must be used
					std::cout<<"[WARNING][CVMPhaseSpaceGen::FillEvent] Ersatz case is involved"<<std::endl;
					if(qPrev.m_E==0.){
						m_GenEvt->m_Particles[posF+k].m_Mom.Set(m_MassFin[k],0.,0.,0.); qk.Set(_Mk,0.,0.,0.);
						std::cout<<"[Error] Ersatz case(1) is not implemented"<<std::endl; exit(0); //TODO: update 'm_PhSpaceWgt'
					}else if((_Mk==0.) && (m_MassFin[k]==0.)){
						m_GenEvt->m_Particles[posF+k].m_Mom=qPrev; m_GenEvt->m_Particles[posF+k].m_Mom*=r; //<== Po=r*Pp+(1-r)Pq
						qk=qPrev; qk-=m_GenEvt->m_Particles[posF+k].m_Mom;
						std::cout<<"[Error] Ersatz case(2) is not implemented"<<std::endl; exit(0); //TODO: update 'm_PhSpaceWgt'
					}else{
						m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[9]++; //Unphysical case
					}
						} break;
				case 1: {
					// [...dM^2]*dCosT, Lab system (for last particle only)
					// dCos(Theta) (Polar angle). Note: Theta angle is {Oz,Pk}

					// June'10'2010: ERROR found, usage is temporary forbidden
					std::cout<<"[ERROR] dM^2*dCosT event reconstruction is FORBIDDEN until proper debug"<<std::endl; exit(0);

					tReal Eprev=qPrev.m_E; tReal Pprev=fabs(qPrev.m_Pz); tReal m2=m_MassFinSqr[k];
					tReal costMin=alg.valMinMaxLastThCos[0]; tReal costMax=alg.valMinMaxLastThCos[1];
					if((Pprev>0.) && (m2>0.)) { // Note: Theta-Max depends on M_q^2
						tReal sin2PkMax=lambdaSqrt*lambdaSqrt/(4.*Pprev*Pprev*m2); //TODO: check if it's sin2 but not sin
						if (sin2PkMax<1.) {
							// Note angle(Pk) is angle(Pin,k) and may be either Theta or (Pi-Theta)
							tReal cosPkX=sqrt((tReal)1.-sin2PkMax);
							if (qPrev.m_Pz>=0.) { //cost=cosPk
								if(costMin<cosPkX){ costMin=cosPkX; if(costMax<costMin){ costMax=costMin; m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[3]++; break; } }
							} else { // cost=-cosPk
								if(costMax>(-cosPkX)){ costMax=-cosPkX; if(costMin>costMax){ costMin=costMax; m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[3]++; break; } }
							}
						}
					}
					//TODO: involve (1-cost) or (1+cost) parameter and switch (rnd) and (1.-rnd) more smart!
					tReal cost=costMin+(costMax-costMin)*params[dim]; dim++;
					m_PhSpaceWgt*=(costMax-costMin);
					//Old: tReal cost=params[dim]; dim++;
					if(cost>1.)cost=1.;if(cost<-1.)cost=-1.; tReal sint=sqrt((1.+cost)*(1.-cost));if(sint>1.)sint=1.;
					// cost: angle generated along Z-axis
					tReal cosPk=((qPrev.m_Pz>=0.)?cost:(-cost)); // cosPk: angle between p and qPrev

					tReal EThChk=Eprev*Eprev-Pprev*Pprev*cosPk*cosPk; tReal PhiSChk=_M2kPrev-_M2k+m2;
					tReal Ep=(PhiSChk*Eprev+Pprev*cosPk*sqrt(PhiSChk*PhiSChk-4.*m2*EThChk))/(2.*EThChk);
					tReal Pp=(PhiSChk*Pprev*cosPk+Eprev*sqrt(PhiSChk*PhiSChk-4.*m2*EThChk))/(2.*EThChk); tReal PpSqr=Pp*Pp;
					if(Pp<0.){m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[3]++; break;} //Error: Cost-cuts are probably WRONG!

					//tReal Ep=(Eprev*(_M2kPrev-_M2k+m2)+Pprev*cosPk*sqrt(lambdaSqrt*lambdaSqrt-4.*Pprev*Pprev*m2*sint*sint))/(_M2kPrev+Pprev*Pprev*sint*sint)/2.;
					//tReal PpSqr=Ep*Ep-m2; if(PpSqr<0.)PpSqr=0.; tReal Pp=sqrt(PpSqr);

					if((Ep==0.) && (Pp==0.)) { Ep=Eprev*epsReal; Pp=Pprev*epsReal; } //Fix 0-photon momentum (if rnd==1.)
					m_PhSpaceWgt*=_1d4PiSqr*PpSqr/(Eprev*Pp-Pprev*Ep*cosPk); //(CosMax-CosMin) is included above (or as genetator weight)
					// Reconstruct momenta
					m_GenEvt->m_Particles[posF+k].m_Mom.Set(Ep,Pp*sint*cos(fi),Pp*sint*sin(fi),Pp*cost); //std::cout<<"P"<<k+2<<" init="<<m_GenEvt->m_Particles[posF+k].m_Mom<<std::endl;
					qk=qPrev; qk-=m_GenEvt->m_Particles[posF+k].m_Mom; // Q_{k-1}=Q_k-pFin_k
						} break;
				case 2: {
					// [...dM^2]*dE, Lab. system
					// P0 must be >0 here! (elsewere eMin==eMax and on must generate cosT)

					// Nt. s=_M2kPrev; sqrt(s)=_MkPrev; \vec{P}_0=qPrev
					tReal E0=qPrev.m_E; tReal m2=m_MassFinSqr[k]; tReal m=m_MassFin[k];
					if(E0<=0.){ m_PhSpaceWgt=0.; bZeroEvt=true; break; } //Safety check
					tReal P0;
					if((qPrev.m_Px==0.)&&(qPrev.m_Py==0.)){ P0=fabs(qPrev.m_Pz); } //==if((k==(m_Nf-1))&&m_bAzimutalSym){ P0=fabs(qPrev.m_Pz); }
					else{ P0=sqrt(fabs((E0-_MkPrev)*(E0+_MkPrev))); /*P0=qPrev.GetPAbs();*/ if(P0>E0)P0=E0; }
					if(P0<=0.){ std::cout<<"[Error] P0==0 is not allowed here"; break; } //Safety check
					tReal D0=_M2kPrev/(E0+P0); //=E0-P0
					tReal phiS=_M2kPrev+m2-_M2k; //'+/-' signs checked

					tReal eMin=(E0*phiS-P0*lambdaSqrt)/(2.*_M2kPrev); if(eMin<m)eMin=m; //TODO: check if |eMin<m|>eps here
					tReal eMax=(E0*phiS+P0*lambdaSqrt)/(2.*_M2kPrev);
					//[debug]if(bLog){ std::cout<<"[evt] eMin="<<eMin<<", eMax="<<eMax<<std::endl; }
					//[debug]if(bLog){ std::cout<<"[evt] E0="<<E0<<", phiS="<<phiS<<", P0="<<P0<<", lambdaSqrt="<<lambdaSqrt<<std::endl; }
					//[debug]if(bLog){ std::cout<<"[evt] _M2kPrev="<<_M2kPrev<<std::endl; }
					tReal eMinCut=((alg.bUsePartFinF0eCuts)?(alg.fPartFinF0eMinCuts[k]):-1.);
					if(eMin<eMinCut)eMin=eMinCut;
					tReal eMaxCut=((alg.bUsePartFinF0eCuts)?(alg.fPartFinF0eMaxCuts[k]):-1.);
					if((eMaxCut>=0.)&&(eMax>eMaxCut))eMax=eMaxCut;
					if(eMax<eMin){ m_PhSpaceWgt=0.; bZeroEvt=true; m_nDebugCounter[((k==1)?4:5)]++; break; } // It's NOT an error but a cuts on 'eMin/eMax'

					tReal Ep;
					bool bModeDEminus1=(alg.iPartFinF0eOrder[k]<0); // <------------------------- mode selection
					if(bModeDEminus1){
						//[mode] gen d{E^-1}, dE= -E^2*d{1/E}
						//Nt. this mode is tested, works, but appear ~30% slower in test task
						Ep=1./(1./eMax+(1./eMin-1./eMax)*params[dim]);dim++; if(Ep>eMax)Ep=eMax;
						m_PhSpaceWgt*=Ep*Ep*(1./eMin-1./eMax);
					}else{
						//[mode] gen dE
						Ep=eMin+(eMax-eMin)*params[dim];dim++; if(Ep>eMax)Ep=eMax;
						m_PhSpaceWgt*=(eMax-eMin);
					}
					tReal Pp,Dp; if(m>0.){ Pp=sqrt(fabs((Ep-m)*(Ep+m))); Dp=m2/(Pp+Ep); }else{ Pp=Ep; Dp=0.; } //Dp=Ep-Pp
					//[is 'Pp==0 && m==0' allowed?] if((m==0.)&&(Pp==0.)) { Pp=_MkPrev*epsReal; } //Fix 0-photon momentum (if rnd==1./0.)

					tReal PcosT=(2.*Ep*E0-phiS); tReal PsinT; tReal PpDcosT; //temporary assignment for 'PcosT' here
					if(fabs(D0)>(E0*0.001)){ //Not the ultrarel. region, [Old:] if((fabs(PcosT)>phiS*0.001)||(Pp==0.)){
						PcosT/=(2.*P0); if(PcosT>Pp)PcosT=Pp; if(PcosT<-Pp)PcosT=-Pp;
						PsinT=sqrt(fabs((Pp-PcosT)*(Pp+PcosT))); if(PsinT>Pp)PsinT=Pp;
						PpDcosT=Pp-PcosT;
					}else{
						tReal DcosT=phiS/(2.*Pp*P0)-D0/P0-Dp/Pp-D0*Dp/P0/Pp;
						PcosT=Pp*(1.-DcosT); if(PcosT>Pp)PcosT=Pp; if(PcosT<-Pp)PcosT=-Pp;
						PsinT=Pp*sqrt(fabs(DcosT*(2.-DcosT))); if(PsinT>Pp)PsinT=Pp;
						PpDcosT=Pp*DcosT; //==Pp-PcosT
					}

					m_PhSpaceWgt*=_1d4PiSqr/P0; // Weight

					// Reconstruct momenta
					m_GenEvt->m_Particles[posF+k].m_Mom.Set(Ep,PsinT*cos(fi),PsinT*sin(fi),PcosT);
					m_GenEvt->m_Particles[posF+k].m_Mom.RotateZYYX(qPrev,P0); //Note: case qPrev={0,0,-P0} is properly kept here
					qk=qPrev; qk-=m_GenEvt->m_Particles[posF+k].m_Mom;

					if (alg.bFillCouplingsCache) {
						if((m_Nf==3) && (k==1) && bVecQCreated && bSmartLowQRegion) {
							//Calculate Pp.Pprev, Pq.Pprev=s-Pp.Pprev in P0z-system (used later)
							//[Common formula] tReal cplPoPp=E0*Ep-P0*PcosT; //Common formula
							tReal cplPoPp=P0*(PpDcosT+Dp)+D0*Ep; //Better for forward region
							tReal cplPoPq=_M2kPrev-cplPoPp;

							//Calculate q.2, q.3 in a standard way ('cplQ.' vars are also used later)
							tReal cplQ2=vecQ.GetCpl(qk); tReal cplQ3=vecQ.GetCpl(m_GenEvt->m_Particles[posF+k].m_Mom);
							m_GenEvt->SetQCpl(posF,cplQ2); m_GenEvt->SetQCpl(posF+1,cplQ3);

							//Get previously filled 2.3
							tReal cpl23=m_GenEvt->GetCplSpecial(posF,posF+1); //=(_M2kPrev-m_MassFinSqr[1]-m_MassFinSqr[0])/2.;

							//Calculate 0.2 & 0.3 (uses previous)
							tReal cpl02=cplQ2+m_MassFinSqr[0]+cpl23; m_GenEvt->SetCpl(pos0,posF,cpl02); //=[0+2+3,2]+[2,2]+[2,3]
							tReal cpl03=cplQ3+m_MassFinSqr[1]+cpl23; m_GenEvt->SetCpl(pos0,posF+1,cpl03); //=[0+2+3,3]+[3,3]+[2,3]

							//Calculate 2.4 & 3.4 (uses previous)
							m_GenEvt->SetCpl(posF+1,posF+2,cpl03+m_GenEvt->GetCplSpecial(pos1,posF+1)-cplPoPp);
							//TODO: check WHY NOT USING 'cpl02' calculated above?
							//m_GenEvt->SetCpl(posF,posF+2,qk.GetCpl(m_GenEvt->m_Particles[pos0].m_Mom)+qk.GetCpl(m_GenEvt->m_Particles[pos1].m_Mom)-cplPoPq);
							m_GenEvt->SetCpl(posF,posF+2,cpl02+qk.GetCpl(m_GenEvt->m_Particles[pos1].m_Mom)-cplPoPq);
						}
					}

					if(bLog){
						std::cout<<"[evt][P-rest||Oz syst.] cosT= "<<PcosT/((Pp>0.)?Pp:1.)<<std::endl;
					}
						} break;
				case 0: {
					// [...dM^2]*dCosT, Center-of-mass system (classical algorithm)
					if(bLog){
						std::cout<<"[evt][Rest syst.] Working in c.m. system of 1.."<<k+1<<" final particles."<<std::endl;
						std::cout<<"[evt][Rest syst.] System shift vector: "<<qPrev<<std::endl;
						std::cout<<"[evt][Rest syst.] System shift vector inv. mass error: "<<(qPrev.GetMassSigned()-_MkPrev)/((_MkPrev>0.)?_MkPrev:1.)<<")"<<std::endl;
					}

					//TODO: involve (1-cost) or (1+cost) parameter and switch (rnd) and (1.-rnd) more smart!
					tReal cost=2.*params[dim]-1.;dim++; if(cost>1.)cost=1.;if(cost<-1.)cost=-1.;
					tReal sint=sqrt((1.+cost)*(1.-cost));if(sint>1.)sint=1.;
					tReal pm=0.5*lambdaSqrt/_MkPrev;
					//dig[sint*sint+cost*cost]=65

					// Weight
					m_PhSpaceWgt*=_1d4PiSqr*lambdaSqrt/_M2kPrev; //=lambda/(M2*2*(4Pi)^2)*(cosTMax-cosTMin)
					// Reconstruct momenta
					if((pm==0.)&&(m_MassFin[k]==0.)) { pm=_MkPrev*epsReal; } //Fix 0-photon momentum (if rnd==1./0.)
					qk.Set(sqrt(pm*pm+_M2k),pm*sint*cos(fi),pm*sint*sin(fi),pm*cost);
					// dig[qk's mass^2]=64 (later: =48)
					m_GenEvt->m_Particles[posF+k].m_Mom.Set(sqrt(pm*pm+m_MassFinSqr[k]),-qk.m_Px,-qk.m_Py,-qk.m_Pz); //std::cout<<"P"<<k+2<<" init="<<m_GenEvt->m_Particles[posF+k].m_Mom<<std::endl;
					// dig[m_GenEvt->m_Particles[posF+k].m_Mom's mass^2]=61 (later: =51)

					tReal tmpCplF1Rest,tmpCplF0Rest; //used if (m_bFillCouplingsCache)
					if (alg.bFillCouplingsCache) {
						if((m_Nf==3) && (k==1)) {
							tmpCplF1Rest=m_GenEvt->m_Particles[posF+k].m_Mom.m_E*_MkPrev; tmpCplF0Rest=qk.m_E*_MkPrev; //f*(i1+i2-f1)
						}
					}

					if(bLog){
						std::cout<<"[evt][Rest syst.] Vec#"<<k<<": "<<m_GenEvt->m_Particles[posF+k].m_Mom<<std::endl;
						std::cout<<"[evt][Rest syst.] Mass^2#"<<k<<" error: "<<(m_GenEvt->m_Particles[posF+k].m_Mom.GetSqr()-m_MassFinSqr[k])/((m_MassFinSqr[k]>0.)?m_MassFinSqr[k]:1.)<<std::endl;
						std::cout<<"[evt][Rest syst.] Sum{rest}: "<<qk<<std::endl;
						std::cout<<"[evt][Rest syst.] Sum{rest}^2 error: "<<(qk.GetSqr()-_M2k)/((_M2k>0.)?_M2k:1.)<<")"<<std::endl;
					}

					m_GenEvt->m_Particles[posF+k].m_Mom.LorMToK(_MkPrev,qPrev); //std::cout<<"P"<<k+2<<" fin ="<<m_GenEvt->m_Particles[posF+k].m_Mom<<std::endl;
					// dig[m_GenEvt->m_Particles[posF+k].m_Mom's mass^2]=51 (before Lor: =51)
					qk.LorMToK(_MkPrev,qPrev);
					//[Allowed instead of last line?]: qk=qPrev; qk-=m_GenEvt->m_Particles[posF+k].m_Mom;
					// dig[qk's mass^2]=48 (before Lor: =64)

					if (alg.bFillCouplingsCache) {
						if((m_Nf==3) && (k==1) && bVecQCreated && bSmartLowQRegion) {
							//Calculate q.2, q.3 in a standard way ('cplQ.' vars are also used later)
							tReal cplQ2=vecQ.GetCpl(qk); tReal cplQ3=vecQ.GetCpl(m_GenEvt->m_Particles[posF+k].m_Mom);
							m_GenEvt->SetQCpl(posF,cplQ2); m_GenEvt->SetQCpl(posF+1,cplQ3);

							//Get previously filled 2.3
							tReal cpl23=m_GenEvt->GetCplSpecial(posF,posF+1); //=(_M2kPrev-m_MassFinSqr[1]-m_MassFinSqr[0])/2.;

							//Calculate 0.2 & 0.3 (uses previous)
							tReal cpl02=cplQ2+m_MassFinSqr[0]+cpl23; m_GenEvt->SetCpl(pos0,posF,cpl02); //=[0+2+3,2]+[2,2]+[2,3]
							tReal cpl03=cplQ3+m_MassFinSqr[1]+cpl23; m_GenEvt->SetCpl(pos0,posF+1,cpl03); //=[0+2+3,3]+[3,3]+[2,3]

							//Calculate 2.4 & 3.4 (uses previous)
							m_GenEvt->SetCpl(posF+1,posF+2,cpl03+m_GenEvt->GetCplSpecial(pos1,posF+1)-tmpCplF1Rest);
							//[Old]m_GenEvt->SetCpl(posF+1,posF+2,m_GenEvt->GetCplSpec(pos0,posF+1)+m_GenEvt->GetCplSpec(pos1,posF+1)-tmpCplF1Rest);
							//TODO: check WHY NOT USING 'cpl02' calculated above?
							m_GenEvt->SetCpl(posF,posF+2,qk.GetCpl(m_GenEvt->m_Particles[pos0].m_Mom)+qk.GetCpl(m_GenEvt->m_Particles[pos1].m_Mom)-tmpCplF0Rest);
							//[Old]m_GenEvt->SetCpl(posF,posF+2,qk.GetCpl(m_GenEvt->m_Particles[pos0].m_Mom)+qk.GetCpl(m_GenEvt->m_Particles[pos1].m_Mom)-tmpCplF0Rest);
						}
					}
						} break;
				default: std::cout<<"[ERROR] in event reconstruction"<<std::endl; exit(0);
				}
			}

			if (alg.bFillCouplingsCache) {
				if(k==2) { // 2 particles left to be generated
					m_GenEvt->SetCpl(posF,posF+1,(_M2k-m_MassFinSqr[1]-m_MassFinSqr[0])/2.);
				}
			}

			if(bLog)std::cout<<"[evt][Lab.syst.] Vec#"<<k<<": "<<m_GenEvt->m_Particles[posF+k].m_Mom<<std::endl;
			if(bLog)std::cout<<"[evt][Lab.syst.] Mass^2#"<<k<<" error: "<<(m_GenEvt->m_Particles[posF+k].m_Mom.GetSqr()-m_MassFinSqr[k])/((m_MassFinSqr[k]>0.)?m_MassFinSqr[k]:1.)<<std::endl;
			if(bLog && (k>1))std::cout<<"[evt][Lab.syst.] Sum{"<<1<<".."<<k<<"}: "<<qk<<std::endl;
			if(bLog && (k>1))std::cout<<"[evt][Lab.syst.] Sum{"<<1<<".."<<k<<"}^2 error: "<<(qk.GetSqr()-_M2k)/((_M2k>0.)?_M2k:1.)<<")"<<std::endl;
		}
		
		if(!bZeroEvt) {
			m_GenEvt->m_Particles[posF].m_Mom=qk; //std::cout<<"P"<<2<<" fin ="<<m_GenEvt->m_Particles[posF].m_Mom<<std::endl;
			if(bLog)std::cout<<"[evt][Lab.syst.] #1:"<<m_GenEvt->m_Particles[posF].m_Mom<<std::endl;
			if(bLog)std::cout<<"[evt][Lab.syst.] Mass^2#1 error: "<<(m_GenEvt->m_Particles[posF].m_Mom.GetSqr()-m_MassFinSqr[0])/((m_MassFinSqr[0]>0.)?m_MassFinSqr[0]:1.)<<std::endl;

			// Generate the free azimutal angle after all the parameters are applied
			if(m_bAzimutalSym) { // The initial momenta pT must be already checked
				m_PhSpaceWgt*=_2Pi;
				tReal fi=_2Pi*m_Gen->GetRndRef()->Rnd(); tReal cosFi=cos(fi); tReal sinFi=sin(fi); //cache cos/sin values

				//Note: 'double' type may be used for 'cos()' here: //asdf
				//double fi8=DBLtR(_2Pi*m_Gen->GetRndRef()->Rnd()); double cosFi8=cos(fi8); tReal cosFi=cosFi8; tReal sinFi=sqrt(1-cosFi*cosFi); //cache cos/sin values
				//std::cout.precision(50); std::cout<<"cosFi^2+sinFi^2 = "<<cosFi*cosFi+sinFi*sinFi<<std::endl; //test

				for(i=0; i<m_Nf; i++) { m_GenEvt->m_Particles[posF+i].m_Mom.RotateXY(cosFi,sinFi); }
				if(bVecQCreated){ vecQ.RotateXY(cosFi,sinFi); } //rotate also cached vector
			}

			// Check parameters used number
			if(dim!=m_nDim){ std::cout<<"[err]Internal error in FillEvent. Unexpected dimension."<<std::endl; return false; }

			if(Q2>=0.) { //Q-vector is initialized
				m_GenEvt->FillQ2Abs(Q2); if(bVecQCreated)m_GenEvt->FillQVec(vecQ);
				//[alt]m_GenEvt->Q2AbsCache()=Q2; if(bVecQCreated)m_GenEvt->QVecCache()=vecQ;
			} else { // Reconstruct Q-vector as FinFirst-InFirst (FinLast-InLast)
				if(alg.bFillCouplingsCache){
					CVM4Vec qTmp=m_GenEvt->m_Particles[posF+m_Nf-1].m_Mom; qTmp-=m_GenEvt->m_Particles[pos1].m_Mom; m_GenEvt->FillQVec(qTmp);
					m_GenEvt->FillQ2Abs(-qTmp.GetSqr());
					//[alt]m_GenEvt->QVecCache()=m_GenEvt->m_Particles[posF].m_Mom; m_GenEvt->QVecCache()-=m_GenEvt->m_Particles[pos0].m_Mom;
				}
			}

			if(alg.bFillCouplingsCache){
				m_GenEvt->FillCouplingsCache(true,true,align.posIn[0]); //Must be called after 'm_GenEvt->m_vecQ' is set
			}

			if(bLog){
				m_GenEvt->LogEvent();
				std::cout<<"Phase space weight: "<<m_PhSpaceWgt<<std::endl;
				m_GenEvt->LogCouplings();
			}

			//[Old,May not occur:] if(m_PhSpaceWgt!=m_PhSpaceWgt) { std::cout<<"[err] NaN weight at m_nFillEvtCalls="<<m_nFillEvtCalls<<std::endl; m_GenEvt->LogEvent(); exit(0); }
			if ((alg.bCheckGenEvts) && (!bZeroEvt)) {
				CheckEvent();
				// with no consequences if 'failed'?
			}
		}
		if(bRestoreStdOutPrec)std::cout.precision(dbgPrecOld);
	}

	if(bZeroEvt) {
		//Note: vectors may not be NaN here (due to 'breaks');
		m_PhSpaceWgt=0.; m_nEmptyEvents++;
		if(bLog)std::cout<<"[evt] zero-weighted event generated."<<std::endl;
	}

	//post-procedures
	SaveLastParams(params);

	//debugs
	if(bLog)std::cout<<"[evt].................................................... Event #"<<m_nFillEvtCalls<<" created"<<std::endl;
	if (m_nHaltAfter>=0) { if(m_nFillEvtCalls>m_nHaltAfter)exit(0); }
	return true;
}
bool CVMPhaseSpaceGen::CheckEvent() { // self check
	bool bEvtOK=true; int i;
	if (m_GenEvt->m_nParticles != (align.posFin0+m_Nf)) { std::cout<<"CheckEvent: Wrong particles number"<<std::endl; bEvtOK=false; }
	for(i=0; i<m_GenEvt->m_nParticles; i++) {
		if (m_GenEvt->m_Particles[i].m_Mom.m_E<=0.) { std::cout<<"CheckEvent: Wrong particle (#"<<i<<") energy: "<<m_GenEvt->m_Particles[i].m_Mom.m_E<<std::endl; bEvtOK=false; }
		tReal m=m_GenEvt->m_Particles[i].m_Mom.GetMassSigned();
		tReal mReq=-1.;
		if(i>=align.posFin0){ mReq=m_MassFin[i-align.posFin0]; }
		else if(i==align.posBeam[0]) { mReq=m_beam[0].mass; } else if(i==align.posBeam[1]) { mReq=m_beam[1].mass; }
		else if((i==align.posIn[0]) || (i==align.posIn[1])) { mReq=0.; } //parton masses
		if(mReq>0.) {
			tReal mErr=fabs(mReq-m)/mReq;
			if(m_MaxParticleMassError<mErr)m_MaxParticleMassError=mErr;
			if (m<0.) { std::cout<<"CheckEvent: Negative mass for particle (#"<<i<<"): "<<m<<" (Expected="<<mReq<<")"<<std::endl; bEvtOK=false; }
			if (mErr>0.01) { std::cout<<"CheckEvent: Invalid mass for particle (#"<<i<<"): "<<m<<" (Expected="<<mReq<<")"<<std::endl; bEvtOK=false; }
		} else {
			if(m_MaxPhotonMass<m)m_MaxPhotonMass=m;
			if (m < (-1e-4)) { std::cout<<"CheckEvent: Negative mass for particle (#"<<i<<"): "<<m<<" (Expected="<<mReq<<")"<<std::endl; bEvtOK=false; }
			if (m > (1e-4)) { std::cout<<"CheckEvent: Invalid mass for particle (#"<<i<<"): "<<m<<" (Expected="<<mReq<<")"<<std::endl; bEvtOK=false; }
		}
	}
	CVM4Vec pDiff;
	if(m_GenEvt->m_nParticles>3) {
		CVM4Vec pIn=m_GenEvt->m_Particles[align.posIn[0]].m_Mom; pIn+=m_GenEvt->m_Particles[align.posIn[1]].m_Mom;
		CVM4Vec pOut=m_GenEvt->m_Particles[align.posFin0].m_Mom; for(i=align.posFin0+1; i<m_GenEvt->m_nParticles; i++)pOut+=m_GenEvt->m_Particles[i].m_Mom;
		pDiff=pOut; pDiff-=pIn; tReal eps=pIn.m_E/100000000;
		if ((fabs(pDiff.m_E)>eps) || (fabs(pDiff.m_Px)>eps) || (fabs(pDiff.m_Py)>eps) || (fabs(pDiff.m_Pz)>eps)) {
			std::cout<<"CheckEvent: Momenta conservation failed: "<<pDiff<<std::endl; bEvtOK=false;
		}
	}
	if(!bEvtOK) {
		std::cout<<"m_nFillEvtCalls: "<<m_nFillEvtCalls<<std::endl; bEvtOK=false;
		m_GenEvt->LogEvent(); exit(0);
	}
	return bEvtOK;
}
tReal CVMPhaseSpaceGen::EvtFuncUnit(CHEPEvtLite* evt) { return 1.; } // To calc phace space & tests
void CVMPhaseSpaceGen::TestVolume(int nFin) {
	if(nFin<2)nFin=2; if(nFin>20)nFin=20;
	int i;
	tReal E1=100.; tReal E2=10.;

	CVMPhaseSpaceGen gen;
	tReal _FinMasses[20]; for(i=0;i<nFin;i++)_FinMasses[i]=0.;
	CVM4Vec pIn1(E1,0.,0.,E1); CVM4Vec pIn2(E2,0.,0.,-E2);
	CVMPhaseSpaceGen::CBeamInfo b1; b1.p.Set(E1,0.,0.,E1); b1.mass=0.;
	CVMPhaseSpaceGen::CBeamInfo b2; b2.p.Set(E2,0.,0.,-E2); b2.mass=0.;
	gen.Init(b1,b2,nFin,_FinMasses,NULL,
		CVMPhaseSpaceGen::CAlgorithmInfo::Alg_UseLastThetaLimits(-1.,1.),
//		CVMPhaseSpaceGen::CAlgorithmInfo::Alg_UseLastThetaLimits(0.,1.),
//		CVMPhaseSpaceGen::CAlgorithmInfo::Alg_UseLastThetaLimits(0.99996,1.),
		CVMPhaseSpaceGen::EvtFuncUnit);

	((CVMGenLite*)gen.m_Gen)->m_SubEvts=1000;
	((CVMGenLite*)gen.m_Gen)->m_nSubOrderMin=10;
	//((CVMGenLite*)gen.m_Gen)->m_iAxisSelMode=2;

	tReal _I=gen.Integrate(0.01);
	std::cout<<"==================================="<<std::endl;
	std::cout<<"Result: "<<DBLtR(_I)<<" +/- "<<DBLtR(gen.m_Gen->m_Error)<<std::endl;
	//((CVMGenLite*)gen.m_Gen)->ViewGrid(0,1,"bh_grid_01");

	tReal s=(pIn1+pIn2).GetSqr(); tReal _Pi=CVMConst::Pi(); tReal V=2.*_Pi;
	for(i=1;i<=(nFin-1);i++) { V*=(1./(16.*_Pi*_Pi)); if(i>1)V/=i; }
	for(i=1;i<=(nFin-2);i++) { V*=s; if(i>1)V/=i; }
	std::cout<<"Volume: "<<V<<std::endl;
	std::cout<<"s="<<s<<std::endl;
	std::cout<<"pIn="<<(pIn1+pIn2)<<std::endl;

	//CVMHistLite hst_E("Egamma",50,CVM_BHVal::m_Cuts.m_EGammaMin,CVM_BHVal::m_Cuts.m_EGammaMax);
	//CVMHistLite hst_cT("CosTheta",50,0.9,CVM_BHVal::m_Cuts.m_CosThetaMax);
	/*
	long nEvts=100; long i;
	for(i=0; i<nEvts; i++) {
		if (!gen.GenEventWgt()) continue;
		tReal w=gen.m_EvtWgt;
		CHEPEvtLite* evt=gen.GetLastEvent();
		CVM4Vec* pGamma=&evt->m_Particles[4].m_Mom; tReal Egamma=pGamma->m_E; tReal CosTheta=pGamma->m_Pz/sqrt(pGamma->GetPSqr());
		std::cout<<"["<<i<<"] Eg="<<Egamma<<", Th="<<CosTheta<<", w="<<w<<std::endl;
		
		hst_E.PutValue(Egamma,w);
		hst_cT.PutValue(CosTheta,w);
	}
	std::cout<<"Result: "<<gen.m_Gen->m_Integral<<" +/- "<<gen.m_Gen->m_Error<<" (xb)"<<std::endl;
	*/

	// Normalize
	//hst_E.Normalize(1.);
	//CVMHistLiteView view1(&hst_E); view1.SetLogY(); view1.m_bPrintNEvts=true; view1.AddPlotToView(&histCl); view1.SaveAsEPS("hst_E.eps");
	//CVMHistLiteView view2(&hst_cT); view2.m_bPlotBins=true; view2.m_bPrintNEvts=true; view2.SaveAsEPS("hst_cT.eps");
}
#endif //#ifdef Decl_CVMPhaseSpaceGen

//
//	Cuts
//
#ifdef Decl_CHEPEvtLiteCut
///////////////////////////////////////////// CHEPEvtLiteCut: Interface for cuts
// nothing to implement

///////////////////////////////////////////// CHEPEvtLiteCutSet: Container class for several cuts
CHEPEvtLiteCutSet::CHEPEvtLiteCutSet():CHEPEvtLiteCut() {}
CHEPEvtLiteCutSet::~CHEPEvtLiteCutSet() {
	std::vector< CHEPEvtLiteCut* >::iterator i;
	for (i=m_Cuts.begin(); i!=m_Cuts.end();++i){ CHEPEvtLiteCut* cut=*i; if(cut!=NULL)delete cut; }
}
const char* CHEPEvtLiteCutSet::CLASS_NAME = "CHEPEvtLiteCutSet";
void CHEPEvtLiteCutSet::AddCut(const CHEPEvtLiteCut* cut) { if(cut!=NULL){ m_Cuts.push_back(cut->Clone()); } }
int CHEPEvtLiteCutSet::GetNCuts() const { return m_Cuts.size(); }
CHEPEvtLiteCut* CHEPEvtLiteCutSet::GetCut(int i) { return m_Cuts[i]; }
const CHEPEvtLiteCut* CHEPEvtLiteCutSet::GetCutConst(int i) const { return m_Cuts[i]; }
void CHEPEvtLiteCutSet::RemoveAll() {
	std::vector< CHEPEvtLiteCut* >::iterator i;
	for (i=m_Cuts.begin(); i!=m_Cuts.end();++i){ CHEPEvtLiteCut* cut=*i; if(cut!=NULL)delete cut; }
	m_Cuts.clear();
}
void CHEPEvtLiteCutSet::CopyFrom(const CHEPEvtLiteCutSet* cuts) {
	if(this != cuts){
		if(GetNCuts()>0)RemoveAll();
		if(cuts!=NULL){ for(int i=0;i<cuts->GetNCuts();i++){ AddCut(cuts->GetCutConst(i)); } }
	}
}
//cut interface
tReal CHEPEvtLiteCutSet::GetScale(CHEPEvtLite* evt) const { //'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
//bool CHEPEvtLiteCutSet::FitEvent(CHEPEvtLite* evt) const { //true if event is passed, false if cutted
	tReal fScale=(tReal)1.;
	std::vector< CHEPEvtLiteCut* >::const_iterator i;
	for (i=m_Cuts.begin(); i!=m_Cuts.end();++i){
		CHEPEvtLiteCut* cut=*i;
		if(cut!=NULL){ fScale*=cut->GetScale(evt); if(fScale<=0.)break; }
	}
	return fScale;
}
CHEPEvtLiteCut* CHEPEvtLiteCutSet::Clone() const {
	CHEPEvtLiteCutSet* res=new CHEPEvtLiteCutSet();
	std::vector< CHEPEvtLiteCut* >::const_iterator i;
	for (i=m_Cuts.begin(); i!=m_Cuts.end();++i){ CHEPEvtLiteCut* cut=*i; if(cut!=NULL){ res->AddCut(cut); } }
	return res;
}
const char* CHEPEvtLiteCutSet::ClassName() const { return CLASS_NAME; }
bool CHEPEvtLiteCutSet::SetCutSmooth(bool bSmooth) { //'false' if not supported
	std::vector< CHEPEvtLiteCut* >::const_iterator i; bool bOk=true;
	for (i=m_Cuts.begin(); i!=m_Cuts.end();++i){ CHEPEvtLiteCut* cut=*i; if(cut!=NULL){ if(!cut->SetCutSmooth(bSmooth)){ bOk=false; }; } }
	return bOk;
}
void CHEPEvtLiteCutSet::ChangeParticleOrder(const CHEPParticleOrderLite* order) {
	std::vector< CHEPEvtLiteCut* >::const_iterator i;
	for (i=m_Cuts.begin(); i!=m_Cuts.end();++i){ CHEPEvtLiteCut* cut=*i; if(cut!=NULL){ cut->ChangeParticleOrder(order); } }
}

///////////////////////////////////////////// CHEPEvtLiteCutE: Energy cut
CHEPEvtLiteCutE::CHEPEvtLiteCutE(int id, tReal EMin, tReal EMax):CHEPEvtLiteCut() { m_id=id; m_EMin=EMin; m_EMax=EMax; m_bSmooth=false; }
CHEPEvtLiteCutE::~CHEPEvtLiteCutE() {}
const char* CHEPEvtLiteCutE::CLASS_NAME = "CHEPEvtLiteCutE";
CHEPEvtLiteCutE* CHEPEvtLiteCutE::CreateCutEMin(int id, tReal EMin) { return new CHEPEvtLiteCutE(id,EMin,-1.); }
CHEPEvtLiteCutE* CHEPEvtLiteCutE::CreateCutEMax(int id, tReal EMax) { return new CHEPEvtLiteCutE(id,-1.,EMax); }
//cut interface
tReal CHEPEvtLiteCutE::GetScale(CHEPEvtLite* evt) const { //'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
//bool CHEPEvtLiteCutE::FitEvent(CHEPEvtLite* evt) const {	//true if event is passed, false if cutted
	if(evt==NULL)return (tReal)0.; CHEPParticleLite* p=evt->GetParticle(m_id); if(p==NULL)return (tReal)0.;
	if(m_bSmooth) {
		tReal res = 1.; tReal xMin=m_EMin; tReal xMax=m_EMax; tReal x=p->m_Mom.m_E;
		if(xMin>=0.) {
			tReal w=xMin/100; if((xMax>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMin, w, x, true);
		}
		if(xMax>=0.) {
			tReal w=xMax/100; if((xMin>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMax, w, x, false);
		}
		return res;
	}else{
		if(m_EMin>p->m_Mom.m_E)return (tReal)0.;				//if((m_EMin>=0.) && (m_EMin>p->m_Mom.m_E))return (tReal)0.;
		if((m_EMax>=0.) && (m_EMax<p->m_Mom.m_E))return (tReal)0.;
	}
	return (tReal)1.;
}
CHEPEvtLiteCut* CHEPEvtLiteCutE::Clone() const { return new CHEPEvtLiteCutE(m_id,m_EMin,m_EMax); }
const char* CHEPEvtLiteCutE::ClassName() const { return CLASS_NAME; }
bool CHEPEvtLiteCutE::SetCutSmooth(bool bSmooth) { //'false' if not supported
	m_bSmooth=bSmooth; return true;
}
void CHEPEvtLiteCutE::ChangeParticleOrder(const CHEPParticleOrderLite* order) {
	if(order!=NULL){ int idAlt=order->Get(m_id); if(m_id!=idAlt){ m_id=idAlt; } }
}

///////////////////////////////////////////// CHEPEvtLiteCutPt: Pt cut
CHEPEvtLiteCutPt::CHEPEvtLiteCutPt(int id, tReal PtMin, tReal PtMax):CHEPEvtLiteCut() { m_id=id; m_PtMin=PtMin; m_PtMax=PtMax; m_bSmooth=false; }
CHEPEvtLiteCutPt::CHEPEvtLiteCutPt(int id, tReal PtMin) :CHEPEvtLiteCut() { m_id=id; m_PtMin=PtMin; m_PtMax=-1.; m_bSmooth=false; }
CHEPEvtLiteCutPt::~CHEPEvtLiteCutPt() {}
const char* CHEPEvtLiteCutPt::CLASS_NAME = "CHEPEvtLiteCutSet";
CHEPEvtLiteCutPt* CHEPEvtLiteCutPt::CreateCutPtMin(int id, tReal PtMin) { return new CHEPEvtLiteCutPt(id,PtMin,-1.); }
CHEPEvtLiteCutPt* CHEPEvtLiteCutPt::CreateCutPtMax(int id, tReal PtMax) { return new CHEPEvtLiteCutPt(id,-1.,PtMax); }
//cut interface
tReal CHEPEvtLiteCutPt::GetScale(CHEPEvtLite* evt) const { //'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
//bool CHEPEvtLiteCutPt::FitEvent(CHEPEvtLite* evt) const {	//true if event is passed, false if cutted
	if(evt==NULL)return (tReal)0.; CHEPParticleLite* p=evt->GetParticle(m_id); if(p==NULL)return (tReal)0.;
	tReal pt2=p->m_Mom.GetPtSqr();
	if(m_bSmooth) {
		tReal res = 1.; tReal xMin=m_PtMin; tReal xMax=m_PtMax; tReal x=pt2;
		if(xMin>=0.) {
			tReal w=xMin/100; if((xMax>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMin, w, x, true);
		}
		if(xMax>=0.) {
			tReal w=xMax/100; if((xMin>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMax, w, x, false);
		}
		return res;
	}else{
		if((m_PtMin>0.) && (m_PtMin*m_PtMin>pt2))return (tReal)0.;
		if((m_PtMax>0.) && (m_PtMax*m_PtMax<pt2))return (tReal)0.;
	}
	return (tReal)1.;
}
CHEPEvtLiteCut* CHEPEvtLiteCutPt::Clone() const { return new CHEPEvtLiteCutPt(m_id,m_PtMin,m_PtMax); }
const char* CHEPEvtLiteCutPt::ClassName() const { return CLASS_NAME; }
bool CHEPEvtLiteCutPt::SetCutSmooth(bool bSmooth) { //'false' if not supported
	m_bSmooth=bSmooth; return true;
}
void CHEPEvtLiteCutPt::ChangeParticleOrder(const CHEPParticleOrderLite* order) {
	if(order!=NULL){ int idAlt=order->Get(m_id); if(m_id!=idAlt){ m_id=idAlt; } }
}

///////////////////////////////////////////// CHEPEvtLiteCutEta: cut on abs(Eta)
CHEPEvtLiteCutEta::CHEPEvtLiteCutEta(int id, tReal EtaMin, tReal EtaMax):CHEPEvtLiteCut() { m_id=id; m_EtaMin=EtaMin; m_EtaMax=EtaMax; m_bSmooth=false; }
CHEPEvtLiteCutEta::CHEPEvtLiteCutEta(int id, tReal EtaMinMax):CHEPEvtLiteCut() { m_id=id; if(EtaMinMax<0)EtaMinMax=-EtaMinMax; m_EtaMin=-EtaMinMax; m_EtaMax=EtaMinMax; m_bSmooth=false; }
CHEPEvtLiteCutEta::~CHEPEvtLiteCutEta() {}
const char* CHEPEvtLiteCutEta::CLASS_NAME = "CHEPEvtLiteCutEta";
CHEPEvtLiteCutEta* CHEPEvtLiteCutEta::CreateCutPtMin(int id, tReal EtaMin) { return new CHEPEvtLiteCutEta(id,EtaMin,-1.); }
CHEPEvtLiteCutEta* CHEPEvtLiteCutEta::CreateCutPtMax(int id, tReal EtaMax) { return new CHEPEvtLiteCutEta(id,-1.,EtaMax); }
//cut interface
tReal CHEPEvtLiteCutEta::GetScale(CHEPEvtLite* evt) const { //'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
//bool CHEPEvtLiteCutEta::FitEvent(CHEPEvtLite* evt) const { //true if event is passed, false if cutted
	if(evt==NULL)return (tReal)0.; CHEPParticleLite* p=evt->GetParticle(m_id); if(p==NULL)return (tReal)0.;
	tReal etaAbs=p->m_Mom.GetEta(); if(etaAbs<0.)etaAbs=-etaAbs;
	if(m_bSmooth) {
		tReal res = 1.; tReal xMin=m_EtaMin; tReal xMax=m_EtaMax; tReal x=etaAbs;
		if(xMin>=0.) {
			tReal w=xMin/100; if((xMax>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMin, w, x, true);
		}
		if(xMax>=0.) {
			tReal w=xMax/100; if((xMin>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMax, w, x, false);
		}
		return res;
	}else{
		if(m_EtaMin>etaAbs)return (tReal)0.;			//if((m_EtaMin>0.) && (m_EtaMin>etaAbs))return (tReal)0.;
		if((m_EtaMax>0.) && (m_EtaMax<etaAbs))return (tReal)0.;
	}
	return (tReal)1.;
}
CHEPEvtLiteCut* CHEPEvtLiteCutEta::Clone() const { return new CHEPEvtLiteCutEta(m_id,m_EtaMin,m_EtaMax); }
const char* CHEPEvtLiteCutEta::ClassName() const { return CLASS_NAME; }
bool CHEPEvtLiteCutEta::SetCutSmooth(bool bSmooth) { //'false' if not supported
	m_bSmooth=bSmooth; return true;
}
void CHEPEvtLiteCutEta::ChangeParticleOrder(const CHEPParticleOrderLite* order) {
	if(order!=NULL){ int idAlt=order->Get(m_id); if(m_id!=idAlt){ m_id=idAlt; } }
}

///////////////////////////////////////////// CHEPEvtLiteCutDR: DeltaR cut
CHEPEvtLiteCutDR::CHEPEvtLiteCutDR(int id1, int id2, tReal DRMin, tReal DRMax):CHEPEvtLiteCut() { m_id1=id1; m_id2=id2; m_DRMin=DRMin; m_DRMax=DRMax; m_bSmooth=false; }
CHEPEvtLiteCutDR::~CHEPEvtLiteCutDR() {}
const char* CHEPEvtLiteCutDR::CLASS_NAME = "CHEPEvtLiteCutDR";
CHEPEvtLiteCutDR* CHEPEvtLiteCutDR::CreateCutDRMin(int id1, int id2, tReal DRMin) { return new CHEPEvtLiteCutDR(id1,id2,DRMin,-1.); }
CHEPEvtLiteCutDR* CHEPEvtLiteCutDR::CreateCutDRMax(int id1, int id2, tReal DRMax) { return new CHEPEvtLiteCutDR(id1,id2,-1.,DRMax); }
//cut interface
tReal CHEPEvtLiteCutDR::GetScale(CHEPEvtLite* evt) const { //'+1.' if event is passed, '0.' if cutted, '0..1' used for smooth cuts
//bool CHEPEvtLiteCutDR::FitEvent(CHEPEvtLite* evt) const { //true if event is passed, false if cutted
	if(evt==NULL)return (tReal)0.;
	CHEPParticleLite* p1=evt->GetParticle(m_id1); if(p1==NULL)return (tReal)0.;
	CHEPParticleLite* p2=evt->GetParticle(m_id2); if(p2==NULL)return (tReal)0.;
	tReal dr2 = p1->m_Mom.GetDRSqr(p2->m_Mom);
	if(m_bSmooth) {
		tReal res = 1.; tReal xMin=m_DRMin; tReal xMax=m_DRMax; tReal x=dr2;
		if(xMin>=0.) {
			tReal w=xMin/100; if((xMax>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMin, w, x, true);
		}
		if(xMax>=0.) {
			tReal w=xMax/100; if((xMin>=0.)&&((xMax-xMin)<w)){ w=(xMax-xMin)/100; }
			res *= CHEPEvtLiteCut::SmoothCutTest(xMax, w, x, false);
		}
		return res;
	}else{
		if((m_DRMin>0.) && (m_DRMin*m_DRMin>dr2))return (tReal)0.;
		if((m_DRMax>0.) && (m_DRMax*m_DRMax<dr2))return (tReal)0.;
	}
	return (tReal)1.;
}
CHEPEvtLiteCut* CHEPEvtLiteCutDR::Clone() const { return new CHEPEvtLiteCutDR(m_id1,m_id2,m_DRMin,m_DRMax); }
const char* CHEPEvtLiteCutDR::ClassName() const { return CLASS_NAME; }
bool CHEPEvtLiteCutDR::SetCutSmooth(bool bSmooth) { //'false' if not supported
	m_bSmooth=bSmooth; return true;
}
void CHEPEvtLiteCutDR::ChangeParticleOrder(const CHEPParticleOrderLite* order) {
	if(order!=NULL){
		int id1Alt=order->Get(m_id1); if(m_id1!=id1Alt){ m_id1=id1Alt; }
		int id2Alt=order->Get(m_id2); if(m_id2!=id2Alt){ m_id2=id2Alt; }
	}
}
#endif // #ifdef Decl_CHEPEvtLiteCut

//
//	Processes
//
#ifdef Decl_CHEPProcessLite

///////////////////////////////////////////// CHEPProcessLite: Base class for processes
// nothing to implement

///////////////////////////////////////////// CHEPProcessLiteSet: Container class for several processes
CHEPProcessLiteSet::CHEPProcessLiteSet():CHEPProcessLite() {}
CHEPProcessLiteSet::~CHEPProcessLiteSet() {
	std::vector< CHEPProcessLite* >::iterator i;
	for (i=m_subProcs.begin(); i!=m_subProcs.end();++i){ CHEPProcessLite* proc=*i; if(proc!=NULL)delete proc; }
}
const char* CHEPProcessLiteSet::CLASS_NAME = "CHEPProcessLiteSet";
void CHEPProcessLiteSet::AddProcess(CHEPProcessLite* proc) {
	if(proc!=NULL){ m_subProcs.push_back(proc); }
}
const CHEPProcessLite* CHEPProcessLiteSet::GetProcess(int i) const { return m_subProcs[0]; }
int CHEPProcessLiteSet::GetProcessNum() const { return m_subProcs.size(); }
//process interface
void CHEPProcessLiteSet::AddCut(const CHEPEvtLiteCut* cut) {
	std::vector< CHEPProcessLite* >::iterator i;
	for (i=m_subProcs.begin(); i!=m_subProcs.end();++i){ CHEPProcessLite* proc=*i; if(proc!=NULL)proc->AddCut(cut); }
}
bool CHEPProcessLiteSet::Prepare(const CVM4Vec* p1, const CVM4Vec* p2, double fPrecision) {
	if(m_subProcs.size() <= 0)return false;							//no c/s available ==> 'false'
	bool bResOk=true;
	std::vector< CHEPProcessLite* >::iterator i;
	for (i=m_subProcs.begin(); i!=m_subProcs.end(); ++i) {
		CHEPProcessLite* proc=*i; if(proc==NULL){ bResOk=false; break; }
		if(!proc->Prepare(p1, p2, fPrecision)){ bResOk=false; break; }
	}
	return bResOk;
}
bool CHEPProcessLiteSet::GetCrossSection(tReal& cs, tReal* err) const {
	cs=(tReal)0.;
	tReal eSqr=0.; tReal eTmp;
	bool bResOk=true;
	std::vector< CHEPProcessLite* >::const_iterator i;
	for (i=m_subProcs.begin(); i!=m_subProcs.end(); ++i) {
		CHEPProcessLite* proc=*i; if(proc==NULL){ bResOk=false; break; }
		tReal scSub; if(!proc->GetCrossSection(scSub,&eTmp)){ bResOk=false; break; }
		cs+=scSub; eSqr+=eTmp*eTmp;
	}
	if((bResOk) && (err!=NULL)) { *err = sqrt(eSqr); }
	return bResOk;
}
bool CHEPProcessLiteSet::GenEvent(CHEPEvtLite* evt) {
	//special case: 1 process in set
	if(m_subProcs.size()==1) {
		CHEPProcessLite* proc=m_subProcs[0];
		if(proc!=NULL){ return proc->GenEvent(evt); }
	}

	tReal cs; if(!GetCrossSection(cs))return false;
	if(cs<=(tReal)0.)return false;
	CVMRnd* rnd = CVMRnd::GetRndRef(); if(rnd==NULL)return false;

	//select sub-process
	tReal csX=cs*rnd->Rnd(); tReal csPrev=0.;
	bool bResOk=false;
	std::vector< CHEPProcessLite* >::iterator i;
	for (i=m_subProcs.begin(); i!=m_subProcs.end();++i){
		CHEPProcessLite* proc=*i; if(proc==NULL){ break; }
		tReal scSub; if(!proc->GetCrossSection(scSub)){ break; }
		csPrev+=scSub;
		if(csPrev>=csX){ bResOk=proc->GenEvent(evt); break; }
	}
	return bResOk;
}
const char* CHEPProcessLiteSet::ProcName() const { return CLASS_NAME; }
int CHEPProcessLiteSet::ProcID() const { return -1; }

///////////////////////////////////////////// CHEPProcessLiteInst: single process class
CHEPProcessLiteInst::CHEPProcessLiteInstFunc::CHEPProcessLiteInstFunc(CVMPhaseSpaceGen* ph,
																	  CHEPProcessLiteInst* proc,
																	  tReal(vmdecl *func)(CHEPEvtLite*)):CVMGen::CVMGenFunc()
{ m_ph=ph; m_proc=proc; m_pFunc=func; }
tReal CHEPProcessLiteInst::CHEPProcessLiteInstFunc::Call(tReal* params) {
	if ((m_ph!=NULL) && (m_pFunc!=NULL)) {
		if (m_ph->FillEvent(params)) {
			tReal wgt=m_ph->m_PhSpaceWgt; //Note: Wgt==0 means e.g. cuts specified at FillEvent-step
			if (wgt>0.) { // Note: Don't call function for Wgt=0 event
				wgt*=((*m_pFunc)(m_ph->GetLastEvent()));
			}
			return wgt;
		}
	}
	if(m_ph!=NULL)m_ph->IncEvtsFailed();
	return (tReal)0.;
}
CVMGen::CVMGenFunc* CHEPProcessLiteInst::CHEPProcessLiteInstFunc::Clone() const { return new CHEPProcessLiteInstFunc(m_ph,m_proc,m_pFunc); }
//
CHEPProcessLiteInst::CHEPProcessLiteInst():CHEPProcessLite() {}
CHEPProcessLiteInst::~CHEPProcessLiteInst() {}
//process interface
void CHEPProcessLiteInst::AddCut(const CHEPEvtLiteCut* cut) { m_cuts.AddCut(cut); }
bool CHEPProcessLiteInst::Prepare(const CVM4Vec* p1, const CVM4Vec* p2, double fPrecision) {
	return false;
}
bool CHEPProcessLiteInst::GetCrossSection(tReal& cs, tReal* err) const {
	return false;
}
bool CHEPProcessLiteInst::GenEvent(CHEPEvtLite* evt) {
	return false;
}
const char* CHEPProcessLiteInst::ProcName() const { return "CHEPProcessLiteInst"; }
int CHEPProcessLiteInst::ProcID() const { return -1; }

#endif // #ifdef Decl_CHEPProcessLite
